from __future__ import absolute_import
import ldap
import logging
from ldap.filter import escape_filter_chars as escape
from ..config import MDConfigReader

logger = logging.getLogger(__name__)

LDAP_TIMEOUT = 10

class BaseAccountLookup(object):
    """
    AccountLookup provides helper functionality for resolving
    certain user or group account data.
    Currently only LDAP and AD are supported.
    """

    def __init__(self, ldap_config='/etc/zarafa/ldap.bitbone.cfg'):
        """Constructor"""
        self.bitbone_ldap_conf = ldap_config
        self._ldap = None
        self._config = None
        self.needed_user_attributes = [
            self.config['ldap_loginname_attribute'],
            self.config['ldap_fullname_attribute'],
            'givenName',
            'sn',
            'departmentNumber',
            'department',
            'company',
            'telephoneNumber',
            'facsimileTelephoneNumber',
            'physicalDeliveryOfficeName',
            'mail',
            'st',
            'co',
            'postalCode',
            'postOfficeBox',
            'streetAddress',
            'l',
            'maildisclaimerTemplate'
        ]
        self._setupBinaryAttributes()

    def _setupBinaryAttributes(self):
        """
        Scans the config keys for information about binary ldap
        attributes (mostly relevant for AD setups).
        This works by finding any *_type = binary entries and looking
        up the matching _attribute value.
        """
        attrs = set()
        for key, val in self.config.data.items():
            if not key.endswith('_type'):
                continue
            if val.lower() != 'binary':
                continue

            attr_key = key[:-len('_type')] + '_attribute'
            ldap_attr = self.config.get(attr_key)
            if ldap_attr:
                attrs.add(ldap_attr.lower())

        self.binary_attributes = list(attrs)

    def get_template_info_from_attributes(self, values):
        info = {
            'fullname': values[self.config['ldap_fullname_attribute']],
            'givenname': values['givenName'],
            'lastname': values['sn'],
            'department': (values['department'] or
                    values['departmentNumber']),
            'phone': values['telephoneNumber'],
            'fax': values['facsimileTelephoneNumber'],
            'office': values['physicalDeliveryOfficeName'],
            'company': values['company'],
            'email': values['mail'],
            'state': values['st'],
            'country': values['co'],
            'zip': values['postalCode'],
            'pobox': values['postOfficeBox'],
            'street': values['streetAddress'],
            'city': values['l'],
            'template:user': None,
        }

        if values['maildisclaimerTemplate']:
            info['template:user'] = [values['maildisclaimerTemplate']]

        return info

    @property
    def ldap(self):
        if self._ldap:
            return self._ldap

        ldap_uri = self.config.get('ldap_uri', '')
        if not ldap_uri:
            ldap_uri = ('%(ldap_protocol)s://'
                '%(ldap_host)s:%(ldap_port)s' %
                self.config)
        logger.debug('using LDAP server uri=%s' % ldap_uri)
        self._ldap = ldap.ldapobject.ReconnectLDAPObject(
            ldap_uri,
            retry_max=3,
            retry_delay=5)
        self._ldap.set_option(ldap.OPT_NETWORK_TIMEOUT, LDAP_TIMEOUT)
        logger.debug('binding to LDAP as user=%r' %
            self.config['ldap_bind_user'])
        self._ldap.simple_bind_s(
            self.config['ldap_bind_user'],
            self.config['ldap_bind_passwd'])

        # AD sends unnecessary referrals; this leads to hangups and
        # operational errors because python-ldap tries to follow them
        # as anonymous, which doesn't work
        # see: http://www.python-ldap.org/faq.shtml
        self._ldap.set_option(ldap.OPT_REFERRALS, 0)
        return self._ldap

    @property
    def config(self):
        if not self._config:
            self.load_config()
        return self._config

    def load_config(self):
        self._config = MDConfigReader(defaults={
            'ldap_server_charset': 'utf-8',
            'ldap_uri': '',
            'ldap_host': '',
            'ldap_bind_user': '',
            'ldap_bind_passwd': '',
            'ldap_search_base': '',
            'ldap_user_search_filter': '(objectClass=zarafaUser)',
            'ldap_group_search_filter': '(objectClass=zarafaGroup)',
            'ldap_emailaddress_attribute': 'mail',
            'ldap_emailaliases_attribute': 'otherMailbox',
            'ldap_fullname_attribute': 'displayName',
            'ldap_groupmembers_attribute': 'member',
            'ldap_groupname_attribute': 'uid',
            'ldap_group_unique_attribute': 'gidNumber',
            'ldap_group_unique_attribute_type': 'text',
            'ldap_loginname_attribute': 'uid',
            'ldap_user_unique_attribute': 'uidNumber',
            'ldap_user_unique_attribute_type': 'text',
            'ldap_group_unique_attribute': 'gidNumber',
            'ldap_group_unique_attribute_type': 'text',
            'ldap_dynamicgroup_type_attribute_value': 'zarafaDynamicgroup',
        })
        self._config.read(self.bitbone_ldap_conf)

        if not self.config.get('ldap_dynamicgroup_search_filter'):
            # there is no zarafa config option for this case, so we
            # make up that filter ourselves using
            self.config['ldap_dynamicgroup_search_filter'] = (
                '(objectClass=%s)' % (
                self.config['ldap_dynamicgroup_type_attribute_value']))

    def get_objectinfo_by_email(self, email):
        """
        Attempts to look up the user with the given email address
        and returns its uid and gids.

        @param str email address
        @return (uid, [gid1, gid2, ...], {user: info})
        """
        raise NotImplemented("get_objectinfo_by_email() must be implemented "
            "by the sub-class")

    def get_ldap_attributes_list(self, filter, attributes,
            binary_attributes=None):
        if binary_attributes:
            attributes = attributes + binary_attributes
            binary_attributes = [attr.lower() for attr in binary_attributes]
        else:
            binary_attributes = []
        binary_attributes += self.binary_attributes
        ret = []
        results = self.ldap.search_st(
            self.config['ldap_search_base'],
            ldap.SCOPE_SUBTREE,
            filter,
            attributes, timeout=LDAP_TIMEOUT
        )
        if not results:
            return ret

        for result in results:
            if not result[0]:
                # in this case it's no real result, but instead an
                # LDAP referral or something else;
                # these only appear at the end.
                # in other words: there are no more real results.
                return ret

            info = {'dn': result[0]}

            for key in attributes:
                if key.lower() == 'dn':
                    # never overwrite dn
                    continue
                info[key] = None
                for resultkey in result[1]:
                    if key.lower() != resultkey.lower():
                        continue
                    if (result[1][resultkey] and
                        result[1][resultkey][0]):
                        info[key] = result[1][resultkey][0]
                        if key.lower() not in binary_attributes:
                            info[key] = self._decode_ldap_value(info[key])
            ret.append(info)

        return ret

    def _decode_ldap_value(self, val):
        charset = self.config['ldap_server_charset']
        try:
            return val.decode(charset)
        except Exception:
            logger.debug('unable to decode LDAP value %r as %s',
                val, charset)
        return val.decode(charset, 'replace')


    def get_ldap_attributes(self, filter, attributes,
            binary_attributes=None):
        logger.debug("looking for attributes: %s" % attributes)
        logger.debug("using filter: %s" % filter)

        res = self.get_ldap_attributes_list(filter, attributes,
            binary_attributes)

        logger.debug("result: %s" % res)

        if not res:
            return {}
        return res[0]

    def get_user_attributes_by_email(self, *args, **kwargs):
        return self._get_type_attributes_by_email('user',
            *args, **kwargs)

    def get_group_attributes_by_email(self, *args, **kwargs):
        return self._get_type_attributes_by_email('group',
            *args, **kwargs)

    def get_dynamicgroup_attributes_by_email(self, *args, **kwargs):
        return self._get_type_attributes_by_email('dynamicgroup',
            *args, **kwargs)

    def _get_type_attributes_by_email(self, type, email, attributes,
            binary_attributes=None):
        subfilter = self.config['ldap_%s_search_filter' % type]
        filter = ('(&(|'
            '(%(mailattr)s=%(email)s)'
            '(%(mailaliasattr)s=%(email)s)'
            ')%(subfilter)s)' % {
            'mailattr': self.config['ldap_emailaddress_attribute'],
            'mailaliasattr': self.config['ldap_emailaliases_attribute'],
            'email': escape(email),
            'subfilter': subfilter
        })

        return self.get_ldap_attributes(filter, attributes,
            binary_attributes)

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
