from __future__ import absolute_import
import logging
from ldap.filter import escape_filter_chars as escape
from .base import BaseAccountLookup
from .ldap_id_handlers import make_id_handler

logger = logging.getLogger(__name__)

class MissingADCredentials(Exception): pass

class AccountLookup(BaseAccountLookup):
    """
    AccountLookup provides helper functionality for resolving
    certain user or group account data.
    Currently only LDAP and AD are supported.

    This specific implementation uses LDAP queries
    to query the system passwd and group databases.
    """

    def __init__(self, ad_credentials='', *args, **kwargs):
        """
        @param str user%password ActiveDirectory credentials:
            will only be used by the migration helper functions and not
            in the daemon
        """
        BaseAccountLookup.__init__(self, *args, **kwargs)
        self.ad_credentials = ad_credentials
        self.id_handlers = {
            'user': make_id_handler(
                self.config['ldap_user_unique_attribute'],
                self.config['ldap_user_unique_attribute_type']),
            'group': make_id_handler(
                self.config['ldap_group_unique_attribute'],
                self.config['ldap_group_unique_attribute_type']),
        }

    def get_groups_of(self, dn):
        """
        Returns attribute lists for all groups the given DN is a member of.

        @param str dn
        @return [{dn: ..., cn: ...}, {dn: ...}, ...]
        """
        filter = ('(&%(groupsearchfilter)s'
            '(%(member)s=%(dn)s))' % {
            'groupsearchfilter': self.config['ldap_group_search_filter'],
            'member': self.config['ldap_groupmembers_attribute'],
            'dn': escape(dn)
        })
        attributes = [
            self.config['ldap_group_unique_attribute'],
            self.config['ldap_groupname_attribute'],
            # required for distinction between AD distribution lists
            # and security groups:
            'groupType',
            'maildisclaimerTemplate',
        ]
        result = self.get_ldap_attributes_list(
            filter, attributes, attributes)

        # order results by group name
        result.sort(cmp=lambda a, b: cmp(
            a[self.config['ldap_groupname_attribute']].lower(),
            b[self.config['ldap_groupname_attribute']].lower()))

        return result

    def get_domain_template(self, email):
        """
        Returns attribute lists for the domain in the specified email adress

        @param str email
        @return [ domain ]
        """

        if not self.config.get('ldap_domain_attribute') and not self.config.get('ldap_domain_search_filter'):
            return None

        user, domain = email.split('@', 1)
        domain = domain.lower()

        filter = ('(&%(domainsearchfilter)s'
            '(%(domain_attribute)s=%(domain)s))' % {
            'domainsearchfilter': self.config['ldap_domain_search_filter'],
            'domain_attribute': self.config['ldap_domain_attribute'],
            'domain': escape(domain)
        })
        attributes = [
            'maildisclaimerTemplate',
        ]
        result = self.get_ldap_attributes_list(
            filter, attributes, attributes)

        if result:
            return [ result[0]["maildisclaimerTemplate"] ]

        return None

    @property
    def ldap_member_attr(self):
        member_attr = self.config.get(
            'ldap_groupmembers_relation_attribute')
        if member_attr:
            return member_attr
        if self.config.get('ldap_groupmembers_attribute_type') == 'dn':
            return 'dn'
        return self.config.get('ldap_user_unique_attribute')

    def get_objectinfo_by_email(self, email):
        """
        Attempts to look up user or groups with the given email address
        and returns its id (objectGuid or uid/gid) and group ids
        (GUIDs or cn)

        @param str email address
        @return {object-info-as-dict}
        """
        logger.info('retrieving user/group data for email=%s from ldap', email)
        member_attr = self.ldap_member_attr

        # try to find a matching user or group
        for object_type in ('user', 'group', 'dynamicgroup', None):
            if not object_type:
                # still nothing? give up and return defaults
                info = {}

                info['template:domain'] = self.get_domain_template(email)

                return info
            get_func = getattr(self,
                'get_%s_attributes_by_email' % object_type)
            if object_type == 'dynamicgroup':
                # dynamic groups are similar to normal groups, so from
                # now on they will be treated identically
                object_type = 'group'

            fetch_attributes = self.needed_user_attributes[:]
            fetch_attributes.append(member_attr)
            values = get_func(
                email,
                fetch_attributes,
                [self.config['ldap_%s_unique_attribute' % object_type]])
            if values:
                break

        id = values[self.config['ldap_%s_unique_attribute' % object_type]]
        id = self.id_handlers[object_type].to_string(id)
        info = self.get_template_info_from_attributes(values)

        logger.info('retrieving additional group data for email=%s from ldap', email)
        gids = []
        group_templates = []
        # start with the user's dn
        search_for_groups_of = [values[member_attr]]

        # values from:
        # http://adsearch-winzero.blogspot.de/2006/04/object-group-attribute-grouptype.html
        AD_DISTRIBUTION_GROUP_TYPES = ['2', '4', '8']

        while search_for_groups_of:
            dn = search_for_groups_of.pop(0)
            if not dn:
                continue
            for groupdata in self.get_groups_of(dn):
                if (groupdata['groupType'] and
                    groupdata['groupType'] in AD_DISTRIBUTION_GROUP_TYPES):
                    # if a parent group of a user or group is an
                    # AD distribution group, we ignore both the group and
                    # any associated parents. the reason for this is that
                    # in such setups the user intends to set up disclaimers
                    # for the distribution group address, but not the
                    # individual users of this group which should usually
                    # be handled by the global disclaimer.
                    # if other behavior is wanted, this can be implemented
                    # by nesting the groups the other way round
                    # (i.e. the disclaimers-enabled group as a direct
                    # parent of the user)
                    continue
                guid = groupdata[self.config['ldap_group_unique_attribute']]

                # in AD we have to decode the GUID to ASCII:
                guid = self.id_handlers['group'].to_string(guid)

                # only add this gid again (and more importantly) only check
                # this group for parent groups, if it has not been
                # visited before (this avoids loops).
                if guid in gids:
                    continue

                gids.append(guid)

                template = groupdata['maildisclaimerTemplate']
                if template:
                    group_templates.append(template)

                # handle nested groups by testing every group found if it
                # is part of yet another group.
                # nested groups can only be supported if the member
                # attribute is 'dn'.
                if member_attr.lower() == 'dn':
                    search_for_groups_of.append(groupdata[member_attr])

        info['template:domain'] = self.get_domain_template(email)
        info['template:group'] = group_templates

        if object_type == 'group':
            info['system:uid'] = None
            info['system:gids'] = [id] + gids
            return info

        info['system:uid'] = id
        info['system:gids'] = gids
        return info

    def get_group_by_guid(self, guid):
        """
        Returns a list of the group's attributes identified by guid

        This is AD-specific and is currently only used for
        debugging.

        @param str hex guid
        @return {}
        """
        guid = self.id_handlers['group'].to_bin(guid)
        filter = '(%s=%s)' % (
            self.config['ldap_group_unique_attribute'],
            escape(guid)
        )
        attributes = [
            self.config['ldap_group_unique_attribute'],
            self.config['ldap_groupname_attribute']
        ]
        return self.get_ldap_attributes_list(
            filter, attributes, attributes)

    def get_id_by_user_id(self, id):
        """
        Returns the id (guid or username) of the user identified by
        the given numerical id.

        Only used by helper scripts.

        @param int id
        """
        if self.config['ldap_user_unique_attribute_type'] == "binary":
            from subprocess import check_output
            def call(*args):
                try:
                    return check_output(*args, stderr=None)
                except Exception:
                    return None

            sid = call(["wbinfo", "--uid-to-sid=%d" % id]).strip()
            if not sid:
                return None
            if not self.ad_credentials:
                raise MissingADCredentials()
            output = call(["net", "-U", self.ad_credentials,
                "ads", "search", "objectSid=%s" % sid, "objectGUID"])
            guid = None
            if not output:
                return None
            for line in output.split('\n'):
                if not line.lower().startswith('objectguid:'):
                    continue
                guid = line.split(':', 2)[1].strip()
                break
            return guid
        else:
            filter = '(&%s(%s=%s))' % (
                self.config['ldap_user_search_filter'],
                'uidNumber',
                escape(str(id))
            )
            attributes = ['uid']
        result = self.get_ldap_attributes(
            filter, attributes, attributes)
        if not result:
            return None

        return result[attributes[0]]

    def get_id_by_group_id(self, id):
        """
        Returns the id (guid or groupname) of the groupd identified by
        the given numerical id.

        Only used by helper scripts.

        @param int id
        """
        if self.config['ldap_group_unique_attribute_type'] == "binary":
            from subprocess import check_output
            def call(*args):
                try:
                    return check_output(*args, stderr=None)
                except Exception:
                    return None

            sid = call(["wbinfo", "--gid-to-sid=%d" % id]).strip()
            if not sid:
                return None
            output = call(["net", "-U", self.ad_credentials,
                "ads", "search", "objectSid=%s" % sid, "objectGUID"])
            guid = None
            for line in output.split('\n'):
                if not line.lower().startswith('objectguid:'):
                    continue
                guid = line.split(':', 2)[1].strip()
                break
            return guid
            raise NotImplemented()
        else:
            filter = '(&%s(%s=%s))' % (
                self.config['ldap_group_search_filter'],
                'gidNumber',
                escape(str(id))
            )
            attributes = ['cn']
        result = self.get_ldap_attributes(
            filter, attributes, attributes)
        if not result:
            return None

        return result[attributes[0]]




# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
