import logging
from struct import pack, unpack
from uuid import UUID

logger = logging.getLogger(__name__)

class IdHandler(object):
    """
    An IdHandler provides an interface for transparently
    converting several common id types from their string
    representations to binary and the other way round.
    """
    def to_bin(self, string):
        raise NotImplemented

    def to_string(self, binary):
        raise NotImplemented


class PlainIdHandler(IdHandler):
    """
    IdHandler which returns data as-is.
    Useful for OpenLDAP setups.
    """
    def to_bin(self, string):
        return string

    def to_string(self, binary):
        return binary


class GuidHandler(IdHandler):
    """
    IdHandler which can handle AD-style GUIDs
    """
    def to_bin(self, string):
        """
        Converts a hexadecimal GUID from AD into its binary representation

        @param str string
        @return str binary
        """
        uuid = UUID(hex=string)
        return uuid.bytes_le


    def to_string(self, binary):
        """
        Converts the binary GUID from AD to its UUID-style ASCII
        representation

        @param str binary
        @return str ascii
        """
        uuid = UUID(bytes_le=binary)
        return str(uuid)


class SidHandler(IdHandler):
    """
    IdHandler which can handle Samba-/Windows-style SIDs
    """
    def _to_bin(self, string):
        chunks = string.split('-')
        if chunks.pop(0) != 'S':
            raise ValueError("Missing S- prefix")
        revision = int(chunks.pop(0))
        block3 = int(chunks.pop(0))
        num_blocks = len(chunks)
        block3_high = block3 >> 16
        block3_low = block3 - (block3_high << 16)
        result = [
            pack('>BBLH', revision, num_blocks, block3_high, block3_low)]
        for chunk in chunks:
            result.append(pack("<L", int(chunk)))
        return ''.join(result)

    def to_bin(self, string):
        try:
            return self._to_bin(string)
        except Exception, exc:
            logging.warn("giving up on malformed SID: %s", exc)
            return "unknown"

    def _to_string(self, binary):
        chunks = ["S"]
        idx = 8
        revision, num_blocks, block3_high, block3_low = unpack(">BBLH",
            binary[0:idx])
        block3 = (block3_high << 16) + block3_low
        chunks += [revision, block3]
        while True:
            chunk = binary[idx:idx+4]
            idx += 4
            if not chunk:
                break
            chunks.append(unpack("<L", chunk)[0])
        return '-'.join(map(str, chunks))

    def to_string(self, binary):
        try:
            return self._to_string(binary)
        except Exception, exc:
            logging.warn("giving up on malformed SID: %s", exc)
            return "unknown"


def make_id_handler(attr_name, attr_type):
    """
    Returns an appropriate IdHandler for the given
    attribute name/type combination

    @param str attr_name (e.g. "gid", "objectSID", "objectGUID")
    @param str attr_type [text|binary]
    @return IdHandler
    """
    if attr_type.lower() != 'binary':
        return PlainIdHandler()
    elif attr_name.lower() == 'objectsid':
        return SidHandler()
    else:
        return GuidHandler()


# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
