import pwd
import grp
import logging
from .base import BaseAccountLookup
from ..utils import getgrouplist

logger = logging.getLogger(__name__)

class AccountLookup(BaseAccountLookup):
    """
    AccountLookup provides helper functionality for resolving
    certain user or group account data.
    Currently only LDAP and AD are supported.

    This specific implementation uses getent/getgrouplist syscalls
    to query the system passwd and group databases.

    Note: This implementation is older than the LDAP version and lacks
          features such as top-level group lookup.
    """

    def get_objectinfo_by_email(self, email):
        """
        Attempts to look up the user with the given email address
        and returns its uid and gids.
        The gids are sorted according the alphabetical order of their
        names.

        Note: no top-level group lookup is supported (in contrast to the
              ldap backend!)

        @param str email address
        @return (uid, [gid1, gid2, ...], {user: info})
        """
        logger.info('retrieving user data for email=%s from ldap', email)
        values = self.get_user_attributes_by_email(email,
            self.needed_user_attributes)
        if not values:
            return {}
        username = values[self.config['ldap_loginname_attribute']]
        info = self.get_template_info_from_attributes(values)

        if not username:
            return {}

        # using nss (and nss_ldap/nss_winbind) under the hood:
        try:
            pwdentry = pwd.getpwnam(username)
        except KeyError:
            return info
        uid = pwdentry.pw_uid
        groups = []
        # we don't use grp.getgrall() here as it seems to use a different
        # mechanism for getting the groups list (same as getent group),
        # which lists less groups than libc's getgrouplist()
        # (also used by `id`)
        for gid in getgrouplist(pwdentry.pw_name, pwdentry.pw_gid):
            grpentry = grp.getgrgid(gid)
            groups.append((grpentry.gr_name.lower(), gid))

        groups.sort()
        gids = [gid for name, gid in groups]
        # move the primary group the beginning again:
        try:
            gids.remove(pwdentry.pw_gid)
        except ValueError:
            # primary gid should always be in gid list, but who knows...
            pass
        gids.insert(0, pwdentry.pw_gid)
        info['system:uid'] = uid
        info['system:gids'] = gids
        return info


# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
