import sys
import os
import time
import atexit
import logging
from signal import SIGTERM

logger = logging.getLogger(__name__)

# Daemon class
#   written by Sander Marechal
#   improved by lots of authors
#   taken from
# http://www.jejik.com/articles/2007/02/a_simple_unix_linux_daemon_in_python/
# Public domain
#
# Modifications for Bitbone by Christian Hoffmann <hoffmann@bitbone.de>
#   - new object model
#   - privilege dropping
#   - exception handling

class DaemonError(Exception): pass
class ForkError(DaemonError): pass
class PrivilegeDroppingError(DaemonError): pass
class PidfileExists(DaemonError): pass

def pid_is_running(pid):
    """
    Returns whether there is a process with the given PID
    @return bool
    """
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    return True

class Daemon(object):
    """
    A generic daemon class.

    Usage: subclass the Daemon class and override the run() method
    """
    def __init__(self, pidfile, stdin=None, stdout=None,
            stderr=None, uid=None, gid=None):
        self.stdin = stdin or '/dev/null'
        self.stdout = stdout or '/dev/null'
        self.stderr = stderr or '/dev/null'
        self.pidfile = pidfile
        self.uid = uid
        self.gid = gid

    def daemonize(self):
        """
        do the UNIX double-fork magic, see Stevens' "Advanced
        Programming in the UNIX Environment" for details
        (ISBN 0201563177)
        http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
        """
        logger.info("Daemonizing...")
        self._fork()
        self._drop_privileges()

    def _drop_privileges(self):
        if self.gid:
            os.setgroups([])
            os.setgid(self.gid)

        if self.uid:
            os.setuid(self.uid)

        if os.getgid() == 0:
            raise PrivilegeDroppingError("Still running as gid=0, "
                "something is wrong")

        if os.getuid() == 0:
            raise PrivilegeDroppingError("Still running as uid=0, "
                "something is wrong")

        logger.info("Dropped privileges successfully")

    def _fork(self):
        try:
            pid = os.fork()
            if pid > 0:
                # exit first parent
                sys.exit(0)
        except OSError, e:
            raise ForkError("Fork #1 failed: %s (%s)" % (
                e.errno, e.strerror))

        # decouple from parent environment
        os.chdir("/")
        os.setsid()
        os.umask(0)

        # do second fork
        try:
            pid = os.fork()
            if pid > 0:
                # exit from second parent
                sys.exit(0)
        except OSError, e:
            raise ForkError("Fork #2 failed: %s (%s)" % (
                e.errno, e.strerror))

        # redirect standard file descriptors
        sys.stdout.flush()
        sys.stderr.flush()
        si = file(self.stdin, 'r')
        so = file(self.stdout, 'a+')
        if self.stdout == self.stderr:
            # reuse existing fd, if stdout and stderr paths match
            se = so
        else:
            se = file(self.stderr, 'a+', 0)
        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())

        # write pidfile
        atexit.register(self.delpid)
        pid = str(os.getpid())
        try:
            file(self.pidfile, 'w+').write("%s\n" % pid)
        except IOError, exc:
            # catch writes to /var/log/non-existant/some.log
            logger.warn("Could not write PID file (error: %s)", exc)

        logger.info("Forked successfully")

    def delpid(self):
        """
        Removes the pid file if it exists
        """
        try:
            os.remove(self.pidfile)
        except OSError:
            pass

    def start(self):
        """
        Start the daemon
        """
        # Check for a pidfile to see if the daemon already runs
        pid = self.read_from_pidfile()

        if pid:
            if pid_is_running(int(pid)):
                raise PidfileExists("pidfile '%s' already exists. "
                    "Daemon already running?\n" % self.pidfile)
            logger.info("Removing stale PID file...")
            self.delpid()

        # Start the daemon
        self.daemonize()
        self.run()

    def read_from_pidfile(self):
        try:
            pf = file(self.pidfile, 'r')
            pid = int(pf.read().strip())
            pf.close()
            return pid
        except IOError:
            return

    def stop(self):
        """
        Stop the daemon
        """
        # Get the pid from the pidfile
        pid = self.read_from_pidfile()

        if not pid:
            # no valid pid, so the process has not been started properly
            # as such we assume the state is "stopped"
            self.delpid()
            return

        # Try killing the daemon process
        try:
            while 1:
                os.kill(pid, SIGTERM)
                time.sleep(0.1)
        except OSError, err:
            err = str(err)
            if err.find("No such process") > 0:
                self.delpid()
            else:
                raise DaemonError(str(err))

    def restart(self):
        """
        Restart the daemon
        """
        self.stop()
        self.start()

    def status(self):
        """
        LSB-compliant status function -- checks service status and
        returns the proper exit codes:
        0 - service running as expected
        1 - service not running (but pid file exists)
        3 - service not running (and no pid file)
        """
        pid = self.read_from_pidfile()
        if not pid:
            raise SystemExit(3)

        # FIXME: Linux-specific!
        if os.path.exists('/proc/%d' % pid):
            raise SystemExit(0)

        raise SystemExit(1)

    def run(self):
        """
        You should override this method when you subclass Daemon.
        It will be called after the process has been
        daemonized by start() or restart().
        """

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
