from __future__ import absolute_import
import re
import logging
from functools import partial
from StringIO import StringIO
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from ..utils import codec_aliases #implicitly used
from .stdlib_email_generator import Generator
from .html import HTMLResourceInjector, URIHandler, CID_HASH_GENERATOR

logger = logging.getLogger(__name__)

class CannotIntegrateDirectly(Exception): pass

class DisclaimerInjector(object):
    """
    Takes care of appending disclaimer data to emails, handling
    all kinds of different mail formats.
    """

    disclaimer_encoding_charset = 'utf-8'

    def __init__(self, plaintext, html, image_resolver=None,
            always_embed_html=False, prefer_direct_embedding=False,
            embedders=None, request_email_dump=lambda: None,
            avoid_quoted_printable=False):
        """
        Constructor

        @param str plaintext
        @param str html
        @param callable image_resolver
        @param bool always_embed_html: Always (also) inserts the
            HTML disclaimer, even for text/plain messages which would not
            be upconverted to MIME otherwise
        @param bool prefer_direct_embedding: Try hard to retain the
            original format and directly embed the disclaimers into
            text or html parts instead of attaching them as MIME parts
        @param list embedders: list of Embedder instances
            to find disclaimer insertion points in mails
        @param callable request_email_dump:
            callback which will be called if the injector thinks it might
            be good to dump the current raw mail for debugging purposes
        @param avoid_quoted_printable:
            tries to encode all mails as 7bit/8bit as necessary;
            initially introduced to work around Zarafa quoted-printable
            decoding bugs
        """
        self.plaintext = plaintext
        self.html = html
        self.always_embed_html = always_embed_html
        self.prefer_direct_embedding = prefer_direct_embedding
        self.request_email_dump = request_email_dump
        self.avoid_quoted_printable = avoid_quoted_printable
        self.resource_injector = HTMLResourceInjector()
        self.resource_injector.set_uri_handler_class(URIHandler)
        self.resource_injector.set_resolver(
            image_resolver)

        self.embedders = embedders or []
        # fall back to simply appending the disclaimer before </body>
        self.embedders.append(
            Embedder(r"()(</body>)", types=['html'],
                name='default-pattern html (body)'))
        # fallback to simply appending the disclaimer at the end
        self.embedders.append(
            Embedder(r"([\r\n]*$)", types=['html', 'txt'],
                name='default pattern txt|html ($)'))

    def process(self, email):
        """
        Processes the given email (as text) and inserts the
        appropriate disclaimers.
        Returns the modified email as text at the end.

        @param email.message.Message email
        @return str data
        """
        email = self.integrate_disclaimer(email)
        out = StringIO()
        generator = Generator(out, mangle_from_=False)
        generator.flatten(email)
        return out.getvalue()

    def integrate_disclaimer(self, email):
        """Integrates the disclaimer text into the given
        message structure.

        @param email.message.Message email
        @return email.message.Message
        """
        if self.plaintext in (None, '', '\n') and not self.html:
            return email
        type = email.get_content_type()
        if type == 'text/plain' and not self.always_embed_html:
            return self.integrate_disclaimer_directly(email)

        if self.prefer_direct_embedding:
            return self.integrate_disclaimer_directly(email)

        email = self.convert_to_multipart(email)
        return self.integrate_mime_disclaimer(email)

    def integrate_disclaimer_directly(self, email, fullemail=None):
        """
        Integrates the disclaimer text directly into the message

        @param email.message.Message email
        @return email.message.Message
        """
        fullemail = fullemail or email
        try:
            return self._integrate_disclaimer_directly(email,
                fullemail=fullemail)
        except Exception, exc:
            # we did our best, but email or disclaimer charset are either
            # wrong or we cannot convert correctly between the given
            # charsets; we'll have to convert the whole message to
            # MIME instead :(
            logger.warn('direct mail disclaimer embedding failed '
                'falling back to MIME integration')
            logger.debug('error was:\n%s', exc_info=exc)
            self.request_email_dump()
            email = self.convert_to_multipart(email)
            return self.integrate_mime_disclaimer(email)

    def _integrate_disclaimer_directly(self, email, fullemail,
        multipart_related_parent=None):
        """
        Recursively walks down the MIME structure and patches
        - all parts of any multipart/alternative part
        - the first part of multipart/{mixed,related} parts

        Everything else is not supported and returns failure (which is
        caught and triggers the fallback solution [MIME embedding])
        """
        have_pics = URIHandler(self.html).analyze()
        content_type = extract_content_type_from_header(email)
        if not email.is_multipart():
            supported_types = (
                "message/rfc822",
                "text/plain",
                "text/html"
            )
            if content_type not in supported_types:
                return email
            if not have_pics or content_type != 'text/html':
                return self._integrate_disclaimer_directly_into_part(
                    email, fullemail=fullemail)
            else:
                email = multipart_related_parent or \
                    self.convert_to_multipart(email, 'related')
                domain = get_domain_from_header(fullemail)
                html, parts = \
                    self.resource_injector.prepare_html_with_images(
                        self.html,
                        domain,
                        cid_hint=get_cids_and_hashes_from_email(email,
                            content_type='image/'))
                for part in parts:
                    email.attach(part)
                self._integrate_disclaimer_directly_into_part(
                    email.get_payload(0),
                    override_html=html,
                    fullemail=fullemail)
                return email

        if content_type == 'multipart/alternative':
            for part in email.get_payload():
                self.integrate_disclaimer_directly(part, fullemail=email)
        elif content_type in 'multipart/mixed':
            self._integrate_disclaimer_directly(email.get_payload(0),
                fullemail=fullemail)
        elif content_type in 'multipart/related':
            self._integrate_disclaimer_directly(email.get_payload(0),
                fullemail=fullemail,
                multipart_related_parent=email)
            remove_duplicate_mime_parts(email.get_payload())

        else:
            raise CannotIntegrateDirectly("Unsupported content_type=%s" %
                content_type)

        return email

    def _integrate_disclaimer_directly_into_part(self, email,
            override_html=None, fullemail=None):
        """
        Patches the disclaimer into the given text/* MIME part

        @param email.message.Message email
        @param str override_html
        """
        # Why don't let .get_payload() decode the message?
        # Well, because it fails silently. This means for unknown
        # encodings or bad base64/quoted-printable, it returns the raw
        # data and we blindly append plain text to it. This can't work.
        # Because of this, we decode the payload ourselves, catch errors
        # and act appropriately.
        body, encoding = decode_email(email)
        target_charset = sanitized_source_charset = \
            email.get_content_charset()

        # we attempt to simply append our disclaimer, but this will
        # only work if charsets match or we can convert accordingly.
        # in all other cases, we have to fall back using MIME, which we
        # try to avoid for simple text messages for complexity and
        # compatibility reasons.
        if not target_charset or target_charset == 'us-ascii':
            # if no charset is given or the charset is us-ascii,
            # we try to re-encode the email text with our disclaimer
            # default encoding (utf-8)
            target_charset = sanitized_source_charset = \
                self.disclaimer_encoding_charset
        else:
            # body text is in some other charset, so we'll try to
            # make our disclaimer fit to that charset instead.
            # no changes needed here
            pass

        # find and set the correct disclaimer type
        content_type = extract_content_type_from_header(email)
        if content_type == "text/plain":
            disclaimer = self.plaintext
            type = "txt"
        elif content_type == "text/html":
            disclaimer = override_html or self.html
            disclaimer = get_html_body(disclaimer)
            type = "html"
        else:
            raise CannotIntegrateDirectly("Unsupported content_type=%s"
                % content_type)

        # ensure that our disclaimer is encodable to the
        # target charset
        try:
            disclaimer = disclaimer.encode(target_charset)
        except UnicodeEncodeError:
            # if the disclaimer is not encodable in the target charset
            # (e.g. because of unicode characters such das &ndash;
            # which have no latin1 mapping), we attempt to
            # upconvert to utf8:
            target_charset = 'utf-8'
            # if the following call fails as well, we let the error pass
            # through the layers unless our fallback (MIME appending)
            # kicks in
            disclaimer = disclaimer.encode(target_charset)

        # if the target charset had to be changed, we have to
        # convert the original email content as well
        if email.get_content_charset() != target_charset:
            # do we change the body charset? yes...
            body = body.decode(sanitized_source_charset)
            body = body.encode(target_charset)
            # only change the target charset if the previous conversion
            # was successful -- otherwise the fallback mechanism
            # will use the wrong charset (target_charset) although the
            # content has not been converted.
            email.set_charset(target_charset)
        else:
            # body charset will not be changed, so we don't
            # decode the body at all; we don't need it, we can
            # just append encoded + encoded, which avoids
            # lots of trouble if the original mail is bad
            # (wrong encoding for the given text)
            pass

        # append our disclaimer in the correct charset variant
        success = False
        for embedder in self.embedders:
            try:
                body = embedder.insert_into(body, what=disclaimer,
                    type=type, email=fullemail)
                success = True
            except EmbedderFailure:
                continue
            break

        if not success:
            raise CannotIntegrateDirectly("No embedder was able to handle"
                " this mail "
                "(num embedders=%d)" % len(self.embedders))


        if self.avoid_quoted_printable or encoding == '7bit':
            if needs_8bit_encoding(body):
                # encoding was 7bit; we scan the message for 8bit chars
                # and upgrade the encoding to "8bit" if necessary
                encoding = '8bit'
            else:
                encoding = '7bit'
            if has_overlong_lines(body):
                encoding = 'binary'

        # encode the whole body properly:
        # 7bit/8bit don't need any real content transformation;
        # all other encodings are passed to python's str.encode()
        if encoding not in ('7bit', '8bit', 'binary'):
            body = body.encode(encoding)

        if 'Content-Transfer-Encoding' in email:
            # avoid duplicate transfer encoding lines
            del email['Content-Transfer-Encoding']

        email['Content-Transfer-Encoding'] = encoding
        email.set_payload(body)

        logger.info('disclaimer successfully embedded (directly, %s)'
            % embedder.name)
        return email

    def integrate_mime_disclaimer(self, email):
        """Integrates the disclaimer text properly into all supported
        and appropriate multipart/mixed parts.

        @param email.message.Message email
        @return email.message.Message
        """
        disclaimer_root = MIMEMultipart('alternative')

        encoded_plaintext_disclaimer = \
            self.plaintext.encode(self.disclaimer_encoding_charset)
        encoded_html_disclaimer = \
            self.html.encode(self.disclaimer_encoding_charset)

        disclaimer = MIMEText('')
        disclaimer.set_charset(self.disclaimer_encoding_charset)
        disclaimer.set_payload(encoded_plaintext_disclaimer)
        if needs_8bit_encoding(encoded_plaintext_disclaimer):
            # prevent duplicate Content-Transfer-Encoding headers
            # (will not raise an Exception if no header is found)
            del disclaimer['Content-Transfer-Encoding']
            disclaimer['Content-Transfer-Encoding'] = '8bit'
        disclaimer_root.attach(disclaimer)

        disclaimer = MIMEText('')
        disclaimer.set_charset(self.disclaimer_encoding_charset)
        disclaimer.set_payload(encoded_html_disclaimer)
        if needs_8bit_encoding(encoded_html_disclaimer):
            # prevent duplicate Content-Transfer-Encoding headers
            # (will not raise an Exception if no header is found)
            del disclaimer['Content-Transfer-Encoding']
            disclaimer['Content-Transfer-Encoding'] = '8bit'
        disclaimer.set_type("text/html")
        domain = get_domain_from_header(email)

        disclaimer = self.resource_injector.inject_into(disclaimer, domain)
        disclaimer_root.attach(disclaimer)

        email.attach(disclaimer_root)

        logger.info('disclaimer successfully embedded (as MIME part)')

        return email

    def convert_to_multipart(self, email, subtype='mixed'):
        """
        Converts the given email structure to multipart/mixed so
        that we can arbitrarily append new MIME parts with our
        disclaimer elements.

        @param email.message.Message email
        @param str subtype: mixed or alternative
        @return email.message.Message
        """
        if email.get_content_type() == 'multipart/%s' % subtype:
            # nothing to do...
            return email

        if email.get_content_maintype() == 'multipart':
            content = MIMEMultipart(email.get_content_subtype())
            for payload in email.get_payload():
                content.attach(payload)
        else:
            content = MIMEText('')

            copy_mime_headers(email, content)
            content.set_payload(email.get_payload())

        # copy parameters from the original mail
        for key, value in email.get_params([]):
            if key == email.get_content_type():
                continue
            content.set_param(key, value)
            email.del_param(key)

        # convert original mail to multipart
        email.set_type('multipart/%s' % subtype)
        email.set_payload([])

        # multipart parts may not have a Content-Transfer-Encoding,
        # so let's remove it. this is safe to do since the encoding is
        # still there in the actual text part.
        del email['Content-Transfer-Encoding']

        # and attach our original content mime part
        email.attach(content)
        return email


def copy_mime_headers(from_, to, which=('Content-Type',
        'Content-Transfer-Encoding', 'MIME-Version')):
    """
    Copies the given list of headers (which) from the source message
    (from_) to the target message (to).

    @param email.message.Message from_
    @param email.message.Message to_
    @param iterable which: list of header names
    """
    for header_name in which:
        if header_name not in from_:
            continue

        if header_name in to:
            # delete previous occurences to avoid duplicates;
            # yes, Message behaves differently than a dict
            del to[header_name]

        to[header_name] = from_[header_name]


def extract_sender_from_header(mime):
    """
    Analyzes the given MIME email and returns the From address
    without any realnames or other non-address stuff.

    @param email.message.Message mime
    @return str address
    """
    res = extract_addresses_from_header(mime['From'])
    if not res:
        return None
    return res[0]

def extract_addresses_from_header(header):
    """
    Parses comma-separated email addresses from the given header
    (e.g. To or From).

    @return list
    """
    result = []
    cur_str = []
    cur_token = []
    in_lt = in_doublequotes = in_escape = False

    def token_done(final=False):
        if cur_str:
            cur_token.append(''.join(cur_str))
        cur_str[:] = []
        if not final:
            return
        candidates = [token for token in cur_token if '@' in token]
        if candidates:
            result.append(candidates.pop(0))
        cur_token[:] = []

    header = header and header.lower() or ''
    for char in header:
        if in_escape:
            cur_str += char
            in_escape = False
        elif char == '\\':
            in_escape = True
        elif char in '"':
            in_doublequotes = not in_doublequotes
        elif in_doublequotes:
            cur_str += char
        elif char in '>':
            in_lt = False
            token_done()
        elif in_lt:
            cur_str += char
        elif char in '<':
            in_lt = True
            token_done()
        elif char in ',;':
            token_done(final=True)
        elif char in ' \t':
            token_done()
        else:
            cur_str += char

    token_done(final=True)

    return result

def extract_targets_from_header(mime):
    """
    Analyzes the given MIME email and returns all To and Cc addresses
    without any realnames or other non-address stuff.

    @param email.message.Message mime
    @return list-of-str addresses
    """
    return (extract_addresses_from_header(mime['To']) +
        extract_addresses_from_header(mime['Cc']))

def extract_content_type_from_header(mime):
    if 'Content-Type' not in mime:
        return 'text/plain'
    if ';' in mime['Content-Type']:
        type, params = mime['Content-Type'].split(';', 1)
        return type
    return mime['Content-Type']

def get_html_body(text):
    """
    Returns only the body part of the given HTML text

    @param str text (HTML) <html><body>something</body></html>
    @return str text (HTML) something
    """
    # note: re.compile() is used here as python <2.7 does not support
    # re.sub()'s flag parameter
    text = re.sub(
        re.compile(r'\A.*<body(\s+[^>]+)?>', flags=re.DOTALL|re.I),
        '', text)
    text = re.sub(
        re.compile(r'</body(\s+[^>]+)?>.*\Z', flags=re.DOTALL|re.I),
        '', text)
    return text

def get_domain_from_header(email):
    """
    Tries to extract the domain name from the given
    email structure by analyzing Message-ID and From

    @param email.message.Message
    """
    for field in (email.get("Message-ID", ""), email.get("From", "")):
        m = re.search('[^<>@\s]+@([^<>@\s]+)', field)
        if m and m.group(1) and m.group(1) != 'localhost':
            return m.group(1)

    return 'localhost'


def decode_email(email):
    """
    Decodes the given email's body and returns the decoded
    body along with the encoding.

    @param email.message.Message email
    @return str
    """
    body = email.get_payload(decode=False)
    encoding = '7bit'
    if 'Content-Transfer-Encoding' in email:
        encoding = email['Content-Transfer-Encoding'].lower()

    if encoding not in ('7bit', '8bit', 'binary'):
        # 7bit/8bit means for us: take literally, so we only have
        # to decode in all other cases

        # we might get bad/malicious/unspported
        # Content-Transfer-Encoding values or bad encoded data, so
        # .decode() might throw Exceptions here; this is expected
        # behaviour -- if this step fails,
        # .integrate_plaintext_disclaimer() will simply fall back to
        # MIME embedding, which is fine for these cases.
        # @raises Exception

        # we are not using any Content-Transfer-Encoding white list
        # here, so we can transparently support future transfer
        # encodings which are supported by newer python versions;
        # fun fact: this also enables emails with transfer encodings
        # of rot13, zip, bz2, utf-8 (i.e. charsets), ...
        body = body.decode(encoding)

    if encoding not in ('7bit', '8bit', 'quoted-printable',
            'base64', 'binary'):
        logger.info('processing email with uncommon encoding=%s',
            encoding)

    return body, encoding


class EmbedderFailure(Exception):
    """
    Thrown when an Embedder is unable to handle the given parameters.

    This usually happens if pre-conditions, type checks or the pattern
    itself does not match.
    """

class Embedder(object):
    """
    An Embedder supports adding a disclaimer into one specific
    type of email.

    The supported email types can be specified by using the types
    array (txt or html are supported) and the
    {anybody,body,headers}_condition parameters. These parameters accept
    a regular expression, which is matched against all text parts of
    the given email (anybody), against the current body part (body) or
    against each header separately (headers). The pattern describes
    if and where the disclaimer will be inserted.
    Usually it should consist of two regexp groups. It will then be
    run through re.sub() and the resulting string will consist of
    (first group) + disclaimer + (second group).
    """

    anybody_condition = None
    body_condition = None
    headers_condition = None

    def __init__(self, pattern, types=None,
            anybody_condition=None, body_condition=None,
            headers_condition=None,
            name=''):
        """
        Constructor

        @param str pattern: regular expression string
        @param list types: supported types (e.g. ["txt"])
        @param str anybody_condition: regexp
        @param str body_condition: regexp
        @param str headers_condition: regexp
        @param str name: optional (useful for debugging)
        """
        def compile(pattern):
            try:
                return re.compile(pattern, flags=re.I|re.DOTALL)
            except Exception, e:
                raise EmbedderFailure("Invalid regexp: %r (%s)" % (
                    pattern, str(e)))
        self.pattern = compile(pattern)
        self.types = types
        if anybody_condition:
            self.anybody_condition = compile(anybody_condition)
        if body_condition:
            self.body_condition = compile(body_condition)
        if headers_condition:
            self.headers_condition = compile(headers_condition)
        self.name = name

    @classmethod
    def fromstring(cls, string, name=''):
        """
        Parses the given line, extracts type, conditions and pattern
        from it and returns a new Embedder instance

        @param str string (one line, no trailing line break)
        """
        state = 'type'
        type = ''
        target = ''
        pattern = ''
        body_condition = ''
        anybody_condition = ''
        headers_condition = ''
        escape = False
        for char in string:
            if state == 'type':
                if char in ' \t':
                    continue
                elif char in '/':
                    state = 'expect-pattern'
                    continue
                type += char
            elif state == 'expect-pattern':
                if char in '\\':
                    escape = True
                    continue
                if char in '/' and not escape:
                    state = 'have-pattern'
                    continue
                if escape:
                    escape = False
                    pattern += '\\'
                pattern += char
            elif state == 'have-pattern':
                if char in ' \t':
                    if pattern:
                        if target == 'any-body':
                            anybody_condition = pattern
                        elif target == 'body':
                            body_condition = pattern
                        elif target == 'headers':
                            headers_condition = pattern
                        else:
                            logger.warn("unknown condition type %r",
                                target)
                    pattern = target = ''
                    state = 'target-done'
                    continue
                target += char
            elif state == 'target-done':
                if char in ' \t':
                    continue
                if char in '/':
                    state = 'expect-pattern'
                    continue
                # parse error...

        types = re.split('[\|,]+', type)
        return cls(
            pattern=pattern,
            types=types,
            anybody_condition=anybody_condition,
            body_condition=body_condition,
            headers_condition=headers_condition,
            name=name)

    def _replace_callback(self, what, match):
        """
        Internal function, used as re.sub callback wrapped in partial()
        """
        num_matches = len(match.groups())
        idx = 1
        for idx in xrange(1, num_matches, 2):
            if match.group(idx) != None:
                break
        prefix = match.group(idx)
        try:
            postfix = match.group(idx + 1)
        except IndexError:
            postfix = ''
        return prefix + what + postfix

    def verify_headers_condition(self, email):
        """
        Checks that the headers condition (if set) is fulfilled.

        Throws EmbedderFailure otherwise.

        @param email.message.Message email
        """
        if self.headers_condition:
            matched = False
            for header in [': '.join(kv) for kv in email.items()]:
                if self.headers_condition.search(header):
                    matched = True
                    break
            if not matched:
                raise EmbedderFailure("Header condition not fulfilled")

    def verify_anybody_condition(self, email):
        """
        Checks that the any-body condition (if set) is fulfilled.

        Throws EmbedderFailure otherwise.

        @param email.message.Message email
        """
        if self.anybody_condition:
            matched = False
            for part in email.walk():
                if part.get_content_type() in ('text/plain', 'text/html'):
                    part_body, encoding = decode_email(part)
                    if self.anybody_condition.search(part_body):
                        matched = True
                        break
                if part.get_content_type() not in (
                        'multipart/mixed', 'multipart/alternative'):
                    continue
            if not matched:
                raise EmbedderFailure("Any-body condition not fulfilled")

    def verify_body_condition(self, body):
        """
        Checks that the body condition (if set) is fulfilled.

        Throws EmbedderFailure otherwise.

        @param str body
        """
        if self.body_condition and not self.body_condition.search(body):
            raise EmbedderFailure("Body condition not fulfilled")

    def verify_type_condition(self, type):
        """
        Checks that the type condition is fulfilled.

        Throws EmbedderFailure otherwise.

        @param str type [txt|html]
        """
        if not type in self.types:
            raise EmbedderFailure("Type does not match")

    def insert_into(self, body, what, type, email):
        """
        Inserts the given text (what) into the mail text (body)

        @param str body: mail body
        @param str what: disclaimer text (what to insert)
        @param str type: [txt|html]
        @param email.message.Message email
        @return str modified body
        """

        self.verify_type_condition(type)
        self.verify_headers_condition(email)
        self.verify_body_condition(body)
        self.verify_anybody_condition(email)

        body, replacements = re.subn(
            self.pattern,
            partial(self._replace_callback, what),
            body,
            count=1)
        if replacements:
            return body
        else:
            raise EmbedderFailure("Didn't match")


def remove_duplicate_mime_parts(payloads, cid_prefix='md.'):
    # ensures that disclaimers images are not uselessly
    # appended twice as MIME parts
    idx = 0
    while idx < len(payloads):
        payload = payloads[idx]
        idx += 1
        if not payload['Content-Id']:
            continue
        if not payload['Content-Id'].startswith('<%s' % cid_prefix):
            # only touch MIME parts generated by us
            continue
        for other in payloads[idx:]:
            if other is payload:
                # should never happen...
                continue
            if other['Content-Id'] == payload['Content-Id']:
                payloads.remove(other)

def get_cids_and_hashes_from_email(email, content_type):
    """
    Scans all MIME parts of the given multipart structure and returns
    {hash_of_content: content_id} for all parts which match the given
    content_type prefix.

    @param email.message email
    @param str content_type
    @return {hash_of_content: content_id}
    """
    ret = {}
    for part in email.get_payload():
        if part.is_multipart():
            continue
        if not part['Content-Id']:
            continue
        if content_type:
            if not part['Content-Type']:
                continue
            if not part['Content-Type'].startswith(content_type):
                continue
        hash = CID_HASH_GENERATOR()
        hash.update(part.get_payload(decode=True))
        ret[hash.hexdigest()] = part['Content-Id'].strip('<> ')

    return ret

def needs_8bit_encoding(text):
    """
    Returns True if the given text contains bytes which need
    8bit encoding.

    @param str text
    @return bool
    """
    for char in text:
        if ord(char) > 0x7f:
            return True
    return False


def has_overlong_lines(text):
    """
    Returns True if the given text contains lines longer than 997.
    (rfc2045 suggests "998 octets or less" but some implementations
    seem to have problems with these edge cases).

    @param str text
    @return bool
    """
    for line in text.split('\n'):
        if len(line) > 997:
            return True
    return False



# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
