from __future__ import absolute_import
import re
import urllib2
from HTMLParser import HTMLParser
from functools import partial
from hashlib import sha1
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
from email.mime.image import MIMEImage

CID_HASH_GENERATOR = sha1

class HTMLResourceInjector(object):
    """
    Takes a MIME object and replaces all referenced file resources with
    cid URIs (Content-ID) and generates the appropriate MIME parts for it.
    This includes wrapping the whole MIME object in multipart/related,
    if necessary.
    """
    def set_uri_handler_class(self, uri_handler_class):
        """Configures the HTML URI handler class to extract external
        resource identifiers from the given HTML.

        @param class URIHandler
        """
        self.uri_handler_class = uri_handler_class

    def set_resolver(self, resolver):
        """Configures the callback which will be called when a file name
        has to be resolved to an absolute path.
        The callback receives one argument, the filename, and should return
        the full path.

        @param callable resolver
        """
        self.resolver = resolver

    def inject_into(self, mime, domain):
        """Takes a text/html MIME object, scans it, replaces all
        virtual URIs with cid URIs, generates the appropriate MIME
        parts and returns the result.

        @param email.message.Message mime: The MIME part which contains
            the HTML to be analyzed and modified.
        @param str domain (used for generating content ids)
        @return email.message.Message result
        """
        if mime.is_multipart():
            raise NotImplementedError(
                    "Processing multipart objects is not supported")
        html = mime.get_payload()
        patched_html, parts = self.prepare_html_with_images(html, domain)
        if html == patched_html and not parts:
            # if nothing is to be done, just return the original
            return mime

        table = {}
        wrapper = MIMEMultipart(_subtype='related')
        wrapper.attach(mime)
        for part in parts:
            wrapper.attach(part)

        mime.set_payload(patched_html)
        return wrapper

    def prepare_html_with_images(self, html, domain, cid_hint=None):
        """
        Scans the given HTML, replaces all
        virtual URIs with cid URIs, generates the appropriate MIME
        parts and returns the result.

        @param str html: use this string as HTML source for
            scanning
        @param str domain (used for generating content ids)
        @return (str html, email.message.Message result)
        """
        handler = self.uri_handler_class(html)
        parts = []
        if not handler.analyze():
            # if nothing is to be done, just return the original
            return html, parts

        table = {}
        for virtpath, realpath in handler.get_paths():
            fd = open(self.resolver(realpath), 'rb')
            data = ""
            hash = CID_HASH_GENERATOR()
            read = True
            while read:
                read = fd.read(4096)
                data += read
                hash.update(read)
            fd.close()
            uuid = hash.hexdigest()
            if cid_hint and uuid in cid_hint:
                table[virtpath] = 'cid:%s' % cid_hint[uuid]
                continue
            cid = 'md.%s@%s' % (
                uuid,
                domain)
            table[virtpath] = 'cid:%s' % cid
            part_already_there = False
            for otherpart in parts:
                if otherpart['Content-ID'] == '<%s>' % cid:
                    part_already_there = True
                    break

            if part_already_there:
                continue
            try:
                part = MIMEImage(data)
            except TypeError:
                part = MIMEApplication(data, 'octet-stream')
            part.add_header('Content-Disposition', 'inline')
            part.add_header('Content-ID', '<%s>' % cid)
            parts.append(part)
        html = handler.translate(table)
        return html, parts


class URIHandler(object):
    """
    Manages URIs in HTML texts:
      - can extract a list of URIs
      - manages replacement of these with substitution URIs
    """

    def __init__(self, html):
        """Constructor

        @param str html
        """
        self.html = html
        self.paths = []
        self.pattern = re.compile(
                r'((?:'
                    '/maildisclaimers/images\.cgi\?show=|'
                    '/umcp/command/maildisclaimer/getimage\?name=|'
                    '/univention-management-console/command/maildisclaimer/getimage\?name=|'
                    '/univention/command/maildisclaimer/getimage\?name=|'
                    'embed\+cid:'
                ')([^\'"\)]+))')

    def analyze(self):
        """
        Analyzes the html content which was passed to the constructor.
        Returns True, if URIs (e.g. <img src="URI" />) have been found,
        False otherwise.
        Also populates the list which can later be looked up using
        UriHandler.get_paths()

        @return bool
        """
        matches = self.pattern.findall(self.html)
        if not matches:
            return False

        html = HTMLParser()
        # Warning: always doing html.unescape() is technically not correct
        # as it depends on the context. It's correct for
        # <img src="some/path.gif" />, but e.g. not for
        # <style> body { background: url(some/path.gif); } </style>
        # Considering that it does not break anything (entities start with
        # &something; in HTML, and & must be urlencoded in any case) and
        # doing real parsing while maintaining all the functionality
        # would be really hard, we keep this as-is for now.
        for virtpath, realpath in matches:
            realpath = html.unescape(realpath)
            if not virtpath.startswith('embed+cid:'):
                # umlauts are encoded as utf-8 and then urlencoded
                # by the editor. therefore, we urldecode and
                # utf-8-decode afterwards.
                realpath = urllib2.unquote(realpath)
                # the following step is necessary because we have
                # utf-8-encoded data in a unicode string.
                # without this step, python would try to encode with
                # the system encoding first, which would fail as it
                # defaults to ascii.
                realpath = realpath.encode("raw_unicode_escape")
                realpath = realpath.decode("utf-8", "replace")
            value = (virtpath, realpath)
            if value not in self.paths:
                self.paths.append(value)
        return True

    def get_paths(self):
        """
        Returns a list of all detected "virtual" paths.

        @return list paths
        """
        return self.paths

    def translate(self, table):
        """
        Translates the given virtual paths to their substitution values
        using the given table.

        @param dict table
        @return str modified html
        """
        return re.sub(
            self.pattern,
            partial(self._translate_callback, table),
            self.html)

    def _translate_callback(self, table, match):
        try:
            return table[match.group(1)]
        except KeyError:
            # don't break the whole template because of a single missing
            # file; instead return the empty string which is tolerated
            # by HTML parsers;
            # returning the original URI would possibly expose private
            # paths, so we better don't do that
            return ""

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
