from __future__ import absolute_import
import logging
from tempfile import NamedTemporaryFile
from email import message_from_string
from ..smtp.esmtpd_proxy import ESMTPProxy
from ..smtp.client import Client as SMTPClient
from ..accounts.passwd import AccountLookup
from ..postfix import PostfixConfigReader
from .storage import Store
from .email import (DisclaimerInjector, Embedder,
    extract_sender_from_header,
    extract_targets_from_header, extract_content_type_from_header)
from ..utils.license import License

logger = logging.getLogger(__name__)

class ContentFilter(ESMTPProxy):
    """
    The ContentFilter implements the necessary interfaces for integrating
    into our disclaimer injection code into Postfix using Postfix's
    content_filter interface.

    This works by setting up an (E)SMTP proxy, where Postfix forwards
    all incoming mails to, processing the mail and re-injecting it back
    to postfix using SMTP on a different Postfix SMTP listener
    (to avoid mail filter loops).
    """

    server_software_name = 'bitbone.disclaimers'

    # per-message state which controls whether the data of the current
    # mail should be dumped to disk
    email_dump_requested = False

    def __init__(self, bindaddress, gateway, local_domain_config=None,
            ldap_config=None, disclaimer_dir=None,
            always_embed_html=False, prefer_direct_embedding=False,
            enable_template_ldap=False, account_lookup=AccountLookup,
            email_dump_path=None, avoid_quoted_printable=False,
            custom_rules_backend=None):
        """
        Constructor

        @param (str, int) bindaddress: address of the local listener
        @param (str, int) gateway: address of the upstream SMTP interface
        @param str local_domain_config: usually path to
            /etc/postfix/main.cf
        @param str ldap_config: usually path to
            /etc/zarafa/ldap.bitbone.cfg
        @param str disclaimer_dir: usually path to
            /opt/bitbone/config/maildisclaimers
        @param bool always_embed_html:
            @see bitbone.disclaimers.email.DisclaimerInjector
        @param bool prefer_direct_embedding:
            @see bitbone.disclaimers.email.DisclaimerInjector
        @param bool enable_template_ldap:
            whether to allow ${LDAP:someAttribute} syntax
        @param str email_dump_path:
            controls whether raw email data should be written to disk
            in case of embedding errors; if specified, it must point
            to a writable directory for the dumps
        @param bool avoid_quoted_printable:
            @see bitbone.disclaimers.email.DisclaimerInjector
        @param disclaimers.storage.CustomRulesBackend custom_rules_backend
        """
        def smtp_factory():
            # test runs in subprocess
            logger.info('connecting to upstream gateway "%s:%s"',
                gateway[0], gateway[1]) #pragma: no cover
            c = SMTPClient() #pragma: no cover
            c.connect(host=gateway[0], port=gateway[1]) #pragma: no cover
            return c #pragma: no cover

        ESMTPProxy.__init__(self,
            smtp_handler=smtp_factory)
        self.bindaddress = bindaddress
        self.always_embed_html = always_embed_html
        self.prefer_direct_embedding = prefer_direct_embedding
        self.enable_template_ldap = enable_template_ldap
        self.avoid_quoted_printable = avoid_quoted_printable
        self.disclaimer_store = Store(disclaimer_dir)
        if custom_rules_backend:
            self.disclaimer_store.prepend_backend(custom_rules_backend)
        self.account_lookup = account_lookup(ldap_config=ldap_config)
        self.account_lookup.load_config()
        self.postfix = PostfixConfigReader()
        self.postfix.read(local_domain_config)
        self.license = License()
        if not self.postfix.get_local_domains():
            logger.error("no local mail domains detected, this is "
                "probably wrong")
            logger.error("all mail will be forwarded as-is")
            logger.error("check your virtual_mailbox_domains in '%s'"
                % local_domain_config)

        self.postfix.prefill_config()
        self.email_dump_path = email_dump_path

    def enable_email_dump(self):
        self.email_dump_requested = True

    def make_listener(self):
        """
        Creates a listener using the address passed in in the constructor
        """
        ESMTPProxy.make_listener(self,
            self.bindaddress) #pragma: no cover [test runs in subprocess]
        logger.info("Listening on %r", self.bindaddress)

    def rewrite_message(self, data):
        """
        Rewrites the body of a message.

        This method is called by the superclass on every message which
        is received. It gives us the chance to inject our disclaimer text.

        @param str data message body
        """

        if not self.license.is_valid():
            logger.info('passing through; license is invalid')
            return data

        self.email_dump_requested = False
        email = message_from_string(data)
        sender_addr = extract_sender_from_header(email)
        logger.info('handling mail from=%s', sender_addr)

        if not self.is_local_address(sender_addr):
            logger.info('passing through non-local mail from=%s',
                sender_addr)
            return data

        target_addresses = extract_targets_from_header(email)
        if self.disclaimer_store.disclaimers_disabled_for_targets(
                target_addresses):
            logger.info('disclaimers disabled for all targets %r',
                target_addresses)
            return data

        content_type = extract_content_type_from_header(email)
        if self.disclaimer_store.disclaimers_disabled_for_type(
                content_type):
            logger.info('disclaimers disabled for content type %r',
                content_type)
            return data

        if self.disclaimer_store.disclaimers_disabled_for_sender(
                sender_addr):
            logger.info('disclaimers disabled for sender address %r',
                sender_addr)
            return data

        headers = [': '.join(kv) for kv in email.items()]
        if self.disclaimer_store.disclaimers_disabled_for_headers(
                headers):
            logger.info('disclaimers disabled because of positive '
                'header blacklist match')
            return data

        # only domains from @some-locally-handled domain are
        # processed at all, all other mails are passed through as they
        # are. this avoids LDAP and getent calls for each and every
        # incoming mail which would introduce a serious performance
        # hit.

        # handle dynamic ${ldap:...} template syntax
        if self.enable_template_ldap:
            def get_dynamic_attributes(keys):
                # str() is needed as it might be unicode() which libldap
                # will not handle
                lookup = [str(key[5:]) for key in keys
                    if key.lower().startswith('ldap:')]
                results = self.account_lookup.get_user_attributes_by_email(
                    sender_addr, lookup)
                ret = {}
                for key, value in results.items():
                    ret['ldap:%s' % key.lower()] = value
                return ret
        else:
            get_dynamic_attributes = False

        info = self.account_lookup.get_objectinfo_by_email(
                sender_addr)
        info['from'] = sender_addr
        txt_disclaimer = self.disclaimer_store.get(
            'txt', email, info,
            get_dynamic_attributes=get_dynamic_attributes)
        html_disclaimer = self.disclaimer_store.get(
            'html', email, info,
            get_dynamic_attributes=get_dynamic_attributes)

        embedders = []
        patterns = self.disclaimer_store.get_embedder_patterns()
        for idx, pattern in enumerate(patterns):
            try:
                embedder = Embedder.fromstring(pattern,
                    name='user-pattern #%d' % idx)
                embedders.append(embedder)
            except Exception:
                logger.error("loading special embedder %d failed, "
                    "continuing anyway" % idx)
                import traceback
                logger.debug('error was:\n%s', traceback.format_exc())

        injector = DisclaimerInjector(
            txt_disclaimer,
            html_disclaimer,
            image_resolver=self.disclaimer_store.get_image_path,
            always_embed_html=self.always_embed_html,
            prefer_direct_embedding=self.prefer_direct_embedding,
            embedders=embedders,
            request_email_dump=self.enable_email_dump,
            avoid_quoted_printable=self.avoid_quoted_printable)
        email.add_header("X-Maildisclaimers", "processed")
        modified = injector.process(email)

        if self.email_dump_requested and self.email_dump_path:
            # only if the current mail shall be dumped
            # (email_dump_requested will indirectly be set by the
            # Injector) and a dump path has been configured by the user
            # (email_dump_path)
            self.dump_mail(data)

        if modified != data:
            logger.info('appended disclaimer to mail from=%s',
                sender_addr)
        else:
            logger.info('passing through local mail from=%s', sender_addr)
        return modified

    def dump_mail(self, data):
        """
        Writes the given raw email data to a disk under a
        randomly chosen name in the configured directory. Logs the file
        name.

        @param str data
        """
        try:
            file = NamedTemporaryFile(
                mode='w+b',
                suffix='.eml',
                prefix='mail',
                dir=self.email_dump_path,
                delete=False)
            file.write(data)
            file.close()
            logger.info('raw mail has been written to %s',
                file.name)
        except Exception, exc:
            logger.warn('failed to write raw email to disk '
                '(for debugging); wrong dump mail path? ("%s")',
                self.email_dump_path, exc_info=exc)

    def is_local_address(self, address):
        """Returns True if the given address is a local address
        of this mail server.

        We use this to determine if the mail in question needs the
        disclaimers appended.

        @param str address
        @return bool
        """
        if not address:
            # no valid address? just assume it was external
            return False

        if '@' not in address:
            return True
        user, domain = address.split('@', 1)
        domain = domain.lower()
        return self.postfix.is_local_domain(domain)

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
