import os
import re
import codecs
import logging
import fnmatch
import datetime
from .template import Template

logger = logging.getLogger(__name__)

def domain_of_address(addr):
    """
    Returns the domain part of an email address.
    """
    m = re.search('[^<>@\s]+@([^<>@\s]+)', addr)
    if m and m.group(1):
        return m.group(1)
    return 'localhost'


def try_names(base_path, names, type, not_found=None):
    """
    Attempts to find one of the given disclaimer names and returns
    its contents if found.
    Returns None if no disclaimer can be found.
    """
    logger.info("will try the following names for "
            "type=%s: %r" % (type, names))

    f = False
    for name in names:
        try:
            path = os.path.join(base_path, name + '.' + type)
            path = path.encode('utf-8')
            f = codecs.open(path, "r", "utf-8", "replace")
            break
        except (IOError, TypeError):
            continue

    logger.info("using disclaimer file '%s'" % path)
    if f:
        return f.read()
    return not_found


class FilesystemBackend(object):
    """
    FilesystemBackend is our legacy backend which searches for
    user-*, group-* etc. disclaimers in the data directory.
    """

    def __init__(self, data_dir):
        self.data_dir = data_dir

    def find_disclaimer(self, type, msg, sender_info):
        """
        Returns either the content of the matching disclaimer, or None,
        indicating that other backends should now be consulted.
        """
        company = sender_info['company'] if 'company' in sender_info else None
        from_ = sender_info.get('from')
        names = []
        if 'system:uid' in sender_info:
            names.append('user-%s' % sender_info['system:uid'])
        for gid in sender_info.get('system:gids', ()):
            names.append('group-%s' % gid)
        if company and '..' not in company and '/' not in company:
            names.append('company-%s' % company)
        if from_ and '/' not in from_:
            domain = domain_of_address(from_)
            names.append('sdomain-%s' % domain)
        names.append('global')

        return try_names(self.data_dir, names, type)

class LDAPBackend(object):
    """
    LDAPBackend checks for maildisclaimerTemplate in the LDAP attributes
    of the user
    """

    def __init__(self, data_dir):
        self.data_dir = data_dir

    def find_disclaimer(self, type, msg, sender_info):
        """
        Returns either the content of the matching disclaimer, or None,
        indicating that other backends should now be consulted.
        """

        names = []

        for t in ("user", "group", "domain"):
            name = "template:%s" % t

            if name in sender_info and sender_info[name]:
                names.extend(sender_info[name])

        names.append("global")

        return try_names(self.data_dir+"/templates/", names, type)

class CustomRulesError(RuntimeError): pass

class CustomRulesBackend(object):
    """
    This backend allows specifying an external python file which is
    consulted for disclaimer selection.
    The function being looked for is find_disclaimer with the same arguments
    as the function in this class.
    """
    def __init__(self, path):
        self.path = path
        self.default_env = {}
        self.env = {}
        self.load()

    def load(self):
        try:
            self._load()
        except Exception as exc:
            raise CustomRulesError("unable to open %s (%s)" %
                (self.path, exc))

        if 'find_disclaimer' not in self.env:
            raise CustomRulesError("missing find_disclaimer function")

    def _load(self):
        name = '<custom rule %s %s>' % (self.path, datetime.datetime.now())
        with open(self.path, 'rb') as f:
            data = f.read()
            code = compile(data, name, 'exec')
            self.env = {}
            self.env.update(self.default_env)
            exec(code, self.env, self.env)

    def find_disclaimer(self, type, email, sender_info):
        try:
            return self.env['find_disclaimer'](type, email, sender_info)
        except Exception as exc:
            raise CustomRulesError("processing error in %s (%s)" %
                (self.path, exc))

    def add_helper(self, name, func):
        self.default_env[name] = func


class Store(object):
    def __init__(self, data_dir):
        """Constructor"""
        self.data_dir = data_dir
        self.backends = [LDAPBackend(data_dir), FilesystemBackend(data_dir)]
        self.image_dir = os.path.join(self.data_dir, 'images')
        self.no_disclaimer_targets_file = os.path.join(self.data_dir,
            'no-disclaimer-targets.db')
        self.no_disclaimer_senders_file = os.path.join(self.data_dir,
            'no-disclaimer-senders.db')
        self.no_disclaimer_types_file = os.path.join(self.data_dir,
            'no-disclaimer-types.db')
        self.no_disclaimer_headers_file = os.path.join(self.data_dir,
            'no-disclaimer-headers.db')
        self.special_embedders_file = os.path.join(self.data_dir,
            'special-embedders.db')

    def prepend_backend(self, backend):
        """
        Adds a new backend for the disclaimer selection process at the
        beginning of the list of backends (i.e. it is preferred to all
        previously added backends).
        """
        self.backends.insert(0, backend)

    def get(self, type, msg, sender_info, get_dynamic_attributes=False):
        """
        Retrieves the best-matching disclaimer text of the given type for
        the given scenario from the file system, prepares the text by
        substituting necessary template markers, handles
        the automatically prepended line break and makes sure that a
        proper charset is used.

        @param str type txt|html
        @param email.Message msg
        @param dict sender_info: optional user info which can be used in
                    templates and/or additional context
        @param callable get_dynamic_attributes: if given, this will be used
            to resolve any non-translated attributes (e.g. using LDAP)
        @return unicode readily-rendered disclaimer text
        """
        text = self.find_disclaimer(type, msg, sender_info)
        if not text:
            return ''

        template = Template()
        if get_dynamic_attributes:
            resolve = set()
            for key in template.get_necessary_keys_for(type, text):
                if key.lower() not in sender_info:
                    resolve.add(key)
            if resolve:
                sender_info.update(get_dynamic_attributes(resolve))
        template.use_data(sender_info)
        text = template.process(type, text)
        if text and type != 'html':
            text = u'\n' + text
        return text

    def find_disclaimer(self, type, msg, sender_info):
        """
        Finds the best-matching disclaimer text file and returns its
        content

        @param str type
        @param str company
        @param int uid
        @param list-of-int gids
        @return unicode
        """
        if type not in ('txt', 'html'):
            logger.warn("Unsupported type '%s' requested" % type)
            return ''

        logger.info("using sender_info: %s" % repr(sender_info))

        for backend in self.backends:
            res = backend.find_disclaimer(type, msg, sender_info)
            if isinstance(res, basestring):
                return res
            # backend did not find anything, try next

        logger.info("no disclaimer configured for this user (uid=%s)"
                % sender_info.get('system:uid', '!!!unknown!!!'))
        return ''


    def count(self):
        """
        Returns the number of configured disclaimers.

        @return int
        """
        try:
            disclaimers = os.listdir(self.data_dir)
        except OSError:
            return 0
        return len(disclaimers)

    def get_image_path(self, filename=""):
        """
        Returns the (sanitized) path to the given filename,
        or, if no filename given, to the image base directory.

        @param str filename
        @return str path
        """
        path = self.image_dir
        if filename:
            filename = filename.strip('/')
            filename = filename.replace('../', './').replace('/..', '/.')
            path = os.path.join(path, filename)
        return os.path.abspath(path)

    def disclaimers_disabled_for_target(self, target_email):
        """
        Returns whether the given target email address is excluded
        from the disclaimers list.

        @param str target_email mail addresses
        """
        return self._value_matched_by_list(target_email,
            self.no_disclaimer_targets_file)

    def disclaimers_disabled_for_sender(self, sender):
        """
        Returns whether the given sender email address is excluded
        from the disclaimers list.

        @param str sender mail address
        """
        return self._value_matched_by_list(sender,
            self.no_disclaimer_senders_file)

    def disclaimers_disabled_for_headers(self, headers):
        """
        Returns whether the email being processed should not
        receive an disclaimer because of its headers.

        @param list headers
        @return bool
        """
        for header in headers:
            if self._value_matched_by_list(header,
                    self.no_disclaimer_headers_file):
                return True
        return False

    def _value_matched_by_list(self, value, list_file):
        """
        Returns True if the given value is matched by any
        entry in the given list file.

        Globbing (and as such "plain" values as well) and Regexps are
        supported.
        E.g.

        some*value (globbing)
        /some.*value/ (regexp)

        @param str value
        @param str list_file (path)
        @return bool
        """
        value = value.lower()
        try:
            db = open(list_file, 'rb')
        except IOError:
            return False
        comment_re = re.compile('#.*$')
        for pattern in db:
            pattern = comment_re.sub('', pattern)
            pattern = pattern.strip().lower()
            if not pattern:
                continue

            if pattern[0] == '+':
                # negate return value (= whitelisting)
                ret = False
                strip = 1
                if len(pattern) > 1 and pattern[1] == ' ':
                    strip = 2
                pattern = pattern[strip:]
            else:
                # normal mode (= blacklisting)
                ret = True

            if not pattern:
                continue
            if pattern[0] == '/':
                pattern = pattern[1:]
                if pattern[-1] == '/':
                    pattern = pattern[:-1]
                try:
                    if re.search(pattern, value, re.I):
                        return ret
                except re.error, e:
                    logger.warn('bad regexp "%s" in "%s": %s',
                        pattern, list_file, str(e))
                    pass #ignore

                continue

            if fnmatch.fnmatch(value, pattern):
                return ret

        return False

    def disclaimers_disabled_for_targets(self, target_addresses):
        """
        Returns True, if all given target addresses have disclaimers
        disabled.
        If a single address is not set as disclaimer-disabled, the
        whole result will be False.

        @param list target_addresses mail addresses
        """
        if not target_addresses:
            # empty list of target address; might happen if mail is
            # sent to symbolic addresses (distribution groups) such as
            # "all:;" or something;
            # in this case, we assume that disclaimers are NOT
            # disabled
            return False

        for target_email in target_addresses:
            if not self.disclaimers_disabled_for_target(target_email):
                return False

        return True

    def disclaimers_disabled_for_type(self, type):
        """
        Returns whether the given Content-Type is excluded
        from the disclaimers list.

        @param str type: Content-Type
        @return bool
        """
        return self._value_matched_by_list(type,
            self.no_disclaimer_types_file)

    def get_embedder_patterns(self):
        """
        Returns a list of all configured special embedders in string
        form (suitable for passing them into Embedder.fromstring)

        @return list("", "", "", ...)
        """
        ret = []
        try:
            fd = open(self.special_embedders_file, "rb")
            for line in fd:
                line = line.strip()
                if not line or line[0] == '#':
                    continue
                ret.append(line)
            fd.close()
        except (OSError, IOError):
            pass
        return ret

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
