import ldap
import re
import logging
from .config import MDConfigReader

LDAP_TIMEOUT = 10

logger = logging.getLogger(__name__)

class PostfixConfigReader(MDConfigReader):
    """
    Provides an interface for reading Postfix's main.cf

    It really does not only what is said above -- it reads the values,
    it does not interpret them. Postfix supports some rather sophisticated
    ways for configuration which will appear in raw format here, e.g.
    something = /etc/postfix/some-list, hash:/etc/postfix/some.db, ldap:,
    mysql: etc.
    Most of them will just be returned as-is. Only ldap: will be interpreted
    right now.

    Also, there is currently no support for line continuations as
    supported by postfix. In these cases only the data from the first
    line will be visible.

    These limitations exist, because we currently don't need any higher
    level interpretation as we can rely on bitbone writing simple
    config files.
    """

    def __init__(self):
        super(PostfixConfigReader, self).__init__()
        self.ldap_cfg = {}

    def read(self, path=None):
        """
        @see MDConfigReader.read()
        """
        path = path or '/etc/postfix/main.cf'
        return MDConfigReader.read(self, path)

    def get_local_domains(self):
        """
        Returns a list of domains which the local mail server handles
        mail for.

        @return list
        """
        try:
            return self['virtual_mailbox_domains'].split(', ')
        except KeyError:
            return []

    def is_local_domain(self, domain):
        """
        Checks if given domain is local. This is the case if
         - domain value is in virtual_mailbox_domains
         - domain can be found in ldap specified by virtual_mailbox_domains

        @return boolean
        """
        local_domains = self.get_local_domains()

        for local_domain in local_domains:
            match = re.match("^ldap:(.*)", local_domain)
            if match:
                cfg_file = match.group(1)
                if self.ask_ldap(cfg_file, domain) == True:
                    return True
            else:
                if local_domain == domain:
                    return True

        return False

    def ask_ldap(self, cfg_file, domain):

        if cfg_file in self.ldap_cfg:
            config = self.ldap_cfg[cfg_file]
        else:
            config = MDConfigReader()
            config.read(cfg_file)
            self.ldap_cfg[cfg_file] = config

        ldapcon = ldap.ldapobject.ReconnectLDAPObject(
            config["server_host"],
            retry_max=3,
            retry_delay=5)
        ldapcon.set_option(ldap.OPT_NETWORK_TIMEOUT, LDAP_TIMEOUT)
        logger.debug('binding to LDAP as user=%r' %
            config['bind_dn'])
        ldapcon.simple_bind_s(
            config['bind_dn'],
            config['bind_pw'])

        # AD sends unnecessary referrals; this leads to hangups and
        # operational errors because python-ldap tries to follow them
        # as anonymous, which doesn't work
        # see: http://www.python-ldap.org/faq.shtml
        ldapcon.set_option(ldap.OPT_REFERRALS, 0)

        results = ldapcon.search_st(
            config['search_base'],
            ldap.SCOPE_SUBTREE,
            config["query_filter"] % (domain),
            [config['result_attribute']], timeout=LDAP_TIMEOUT
        )

        logger.debug('results: %s' % repr(results))

        return len(results) >= 1

    def prefill_config(self):
        """
        Reads all domains and prefill ldap_cfg to overcome permission
        problems when they are dropped
        """
        self.is_local_domain("foo.bar")


# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
