import asynchat
import socket
from ..defer import Deferred
from .helpers import FeatureList

class UnexpectedServerReply(Exception):
    def __init__(self, code, message):
        Exception.__init__(self, "Unexpected server reply: %s, %r" % (
            code, message))

class CommandNotSentError(Exception):
    def __init__(self, command):
        Exception.__init__(self, "Requested command %s has not been sent "
            "yet" % command)

class Client(asynchat.async_chat):
    """
    A simple, really dumb, asynchronous (E)SMTP client.

    This is mainly useful for simply relaying readily prepared SMTP
    commands, as no extensive SMTP command wrapping is implemented.

    For now, this is enough for us, as we get our SMTP commands directly
    from our SMTP server and pass them through in most cases.
    """
    def __init__(self):
        """Constructor"""
        asynchat.async_chat.__init__(self)
        self.ibuffer = []
        self.esmtp_features = FeatureList()
        self._next_result_callbacks = []
        self._multiline_reply_lines = []
        self._outgoing_queue = []
        self.myname = socket.getfqdn()
        self.esmtp = False
        self.helo_ehlo_done = False
        self.set_terminator('\r\n')
        self._expecting_reply = False
        self._close_handler = Deferred()

    def connect(self, host, port):
        """
        Connects to the target host.

        This only wraps asynchat.async_chat.connect() so that a callback
        can be specified.
        """
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        asynchat.async_chat.connect(self, (host, port))
        return self._make_next_result_callback(id=':CONNECT')

    def handle_connect(self):
        """
        Called upon successful connect by asynchat.async_chat.

        Used for setting up internal state.
        """
        asynchat.async_chat.handle_connect(self)
        self._expecting_reply = True

    def _make_next_result_callback(self, id):
        """
        Internal helper method which creates a Deferred object and
        ensures that it is called when the next reply from the server
        is received.

        @param id: ESMTP command or other identifier, used for
            .want_reply_for
        @return Deferred
        """
        d = Deferred()
        self._next_result_callbacks.append((id.upper(), d))
        return d

    def collect_incoming_data(self, data):
        """
        This method buffers all incoming data until a line terminator
        is found. Called by asynchat.async_chat.
        """
        self.ibuffer.append(data)

    def found_terminator(self):
        """
        This method parses the line from the server and calls the
        associated callbacks.
        It is called by asynchat.async_chat whenever a line terminator
        is found.
        """
        data = ''.join(self.ibuffer)
        reply = self._parse_reply(data)
        if reply:
            if not self._next_result_callbacks:
                raise UnexpectedServerReply(*reply)
            self._expecting_reply = False
            name, callback = self._next_result_callbacks.pop(0)
            callback(*reply)
            self._flush_outgoing_queue()
        self.ibuffer = []

    def _flush_outgoing_queue(self):
        """
        Checks if any outgoing data can be sent to the server.

        This class uses SMTP-level buffering. If the server does not
        support command pipelining (ESMTP PIPELINING), this class buffers
        all outgoing commands until the next reply from the server arrived.
        Otherwise, commands are sent out as soon as possible.

        This ensures that no server is flooded with commands in an
        unexpected way.

        This method manages exactly this logic, it is called both
        by .push() (the only place where new to-be-sent data can be
        added) and .found_terminator() (the only place where server
        replies are processed).
        """
        pipelining = 'PIPELINING' in self.esmtp_features
        while self._outgoing_queue and (
                pipelining or not self._expecting_reply):
            command = self._outgoing_queue.pop(0)
            asynchat.async_chat.push(self, command)
            self._expecting_reply = True
            if pipelining:
                break

    def _parse_reply(self, data):
        """
        Internal helper method which parses a line from the server.

        In case of single-line responses, the parsed result is returned
        immediately.
        For multi-line responses, the parsed result is saved and returned
        once the last line of the response is received.

        @param str data
        @return None/(code, message)/(code, messages)
        """
        code = data[:3]
        message = data[4:]
        if data[3] == '-':
            self._multiline_reply_lines.append(message)
            return

        if self._multiline_reply_lines:
            self._multiline_reply_lines.append(message)
            ret = code, self._multiline_reply_lines
        else:
            ret = code, message
        self._multiline_reply_lines = []
        return ret

    def docmd(self, cmd, arg=''):
        """
        Sends a command to the server

        Callback will be called, when the reply is received.

        @param str cmd
        @param str arg
        @return Deferred
        """
        line = cmd
        if arg:
            line += " " + arg
        d = self.push(line + "\r\n", command_hint=cmd)
        if line.lower().startswith('ehlo '):
            d.callback(self._record_esmtp_features)
        return d

    def ehlo_or_helo_if_needed(self):
        """
        Ensures that EHLO or HELO has been sent to the server.

        If it hasn't been, it will first try EHLO, then HELO if that fails.

        @return Deferred will be called once a valid EHLO or HELO reply
            has been received.
        """
        if self.helo_ehlo_done:
            return
        d = self.docmd('ehlo', self.myname)
        d.callback(self._ehlo_or_helo_stage_2)
        d = self._ehlo_or_helo_deferred = Deferred()
        return d

    def _ehlo_or_helo_stage_2(self, code, reply):
        """
        Internal helper method for .ehlo_or_helo_if_needed() which
        interprets the EHLO result.
        Depending on it, it either sends a HELO request or sets
        state accordingly and executes the associated Deferred object.

        @param str code
        @param str reply
        """
        if code[0] == '2':
            self.esmtp = True
            self.helo_ehlo_done = True
            self._ehlo_or_helo_deferred(code, reply)
            del self._ehlo_or_helo_deferred
            return
        d = self.docmd('helo', self.myname)
        d.callback(self._ehlo_or_helo_stage_3)

    def _ehlo_or_helo_stage_3(self, code, reply):
        """
        Internal helper method for .ehlo_or_helo_if_needed() which
        interprets the HELO result.
        Depending on it, it either sends a HELO request or silently
        drops the call.

        @param str code
        @param str reply
        """
        if code[0] == '2':
            self.helo_ehlo_done = True
            self._ehlo_or_helo_deferred(code, reply)
            del self._ehlo_or_helo_deferred

    def _record_esmtp_features(self, code, replies):
        """
        Internal helper which is called for any EHLO replies and records
        the transmitted ESMTP features for later use.

        @param str code
        @param list replies
        """
        if isinstance(replies, list):
            self.esmtp_features = FeatureList(replies[1:])

    def push(self, data, command_hint='.'):
        """
        Sends raw data to the server.

        Callback will be called once the server replies.

        @param str data
        @return Deferred
        """
        d = self._make_next_result_callback(id=command_hint)
        self._outgoing_queue.append(data)
        self._flush_outgoing_queue()
        return d

    def want_reply_for(self, command):
        """
        This method allows for delayed callback registration.

        Use case:
          - Client sends EHLO foobar
          - Client user discovery, that it requires the EHLO reply,
            which has not yet been processed; also, the client user has
            no access to the docmd('EHLO', 'foobar') return value
          - Client user does: client.want_reply_for('EHLO').callback(...)
          - Client is happy

        @param str command: (E)SMTP command in uppercase or
            ':CONNECT' or ':CLOSE'
        """
        if command == ':CLOSE':
            return self._close_handler

        for name, callback in self._next_result_callbacks:
            if name == command:
                return callback

        raise CommandNotSentError(command)

    def handle_close(self):
        asynchat.async_chat.handle_close(self)
        self._close_handler()

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
