# Inspired by Python-2.7's smtpd lib and
# http://cr.yp.to/smtp.html
import socket
import asyncore
import asynchat
import logging
from functools import partial
from email.utils import formatdate as rfc2822_date
from .helpers import (InvalidAddressPrefixError, parse_address,
    decode_message, FeatureList)
from ..timer import CallLater

logger = logging.getLogger(__name__)

class ESMTPChannel(asynchat.async_chat):
    """
    A simple, but extensible RFC1869 ESMTP implementation

    Based on asyncore, this means you must have some main loop running
    asyncore.loop() somewhere.
    Based on bitbone.timer, this means, bitbone.timer.timer_manager.run()
    must be called regularly as well.
    """
    COMMAND = 0
    DATA = 1
    server_software_name = 'bitbone.esmtpd'

    def __init__(self, sock, address, message_handler=None, timeout=300):
        """
        Constructor

        @param socket.socket sock
        @param (host, port) address
        @param callable message_handler(sender, recipients, message)
        @param int timeout
        """
        asynchat.async_chat.__init__(self, sock)
        self.remote_addr = address
        if isinstance(address, tuple):
            logger.info('accepted new client, from=%s', address[0])
        self.message_handler = message_handler
        self.esmtp_features = FeatureList([
            '8BITMIME',
            'PIPELINING',
            'ENHANCEDSTATUSCODES'
        ])
        self.esmtp = None
        self.ibuffer = []
        self.client_greeting = None
        self.disconnected = False
        self.reset_state()
        self.fqdn = socket.getfqdn()
        self.timeout_call = CallLater(self.handle_timeout, delay=timeout)
        self.connection_started()

    def push(self, code, data):
        """
        Sends a reply to the connected client.

        @param str code
        @param str/list data
        """
        delim = '-'
        if isinstance(data, basestring):
            # this allows direct passing of return values from
            # smtlib.SMTP.getreply() to this function
            data = data.split('\n')

        for line in data:
            if line is data[-1]:
                delim = ' '
            asynchat.async_chat.push(
                self, str(code) + delim + line + '\r\n')

    def connection_started(self):
        """
        Called, when a new client connected.
        Useful for sending the banner.
        """
        self.push('220', '%s ESMTP Python/%s' % (self.fqdn,
            'bitbone.esmtpd'))

    def reset_state(self):
        """
        Resets the state of this channel (includes sender, recipients
        and command state).

        This is used for initialization, for implementing the SMTP
        RSET command and after a successful DATA request.
        """
        self.mail_from = None
        self.mail_params = []
        self.recipients = []
        self.set_state(self.COMMAND)

    def set_state(self, state):
        """
        Sets the internal parsing state and therefore implicitly the
        terminator.

        This effectiveley enables us to parse both commands and "arbitrary"
        data in DATA requests.

        @param int state [COMMAND, DATA]
        """
        if state == self.COMMAND:
            self.set_terminator('\r\n')
        elif state == self.DATA:
            self.set_terminator('\r\n.\r\n')
        else:
            raise RuntimeError("Unknown state") #pragma: no cover
        self.state = state

    def collect_incoming_data(self, chunk):
        """
        Appends the given chunk to the internal data buffer.

        This method is called by asynchat whenever new data is available.

        @param str chunk
        """
        self.ibuffer.append(chunk)
        self.timeout_call.reset()

    def found_terminator(self):
        """
        This method is called by asynchat whenever our currently set
        terminator is found.
        For us, this either indicates a command call or the end of a
        DATA command.
        This method handles both cases and dispatches work according
        to the current internal state.
        """
        data = ''.join(self.ibuffer)
        self.ibuffer = []

        if self.state == self.COMMAND:
            ret = self.command_received(data)
        elif self.state == self.DATA:
            ret = self.data_received(data)
        else:
            raise RuntimeError("Unknown state") #pragma: no cover
        if ret:
            self.push(*ret)

    def command_received(self, data):
        """
        Command dispatcher

        This method is called whenever a new command call is received.
        Parsing and dispatching happens here.

        @param str data (command string)
        """
        if not data:
            return '500', '5.5.2 Error: bad syntax'

        if ' ' in data:
            command, arg = data.split(' ', 1)
        else:
            command = data
            arg = None

        try:
            func = getattr(self, 'smtp_' + command.upper())
        except AttributeError:
            try:
                # we support a catchall method
                func = getattr(self, 'smtp_')
            except AttributeError:
                func = partial(self.unknown_command_called, command)
        return func(arg)

    def unknown_command_called(self, cmd, arg):
        """
        Called whenever an unknown command is requested.

        @param str cmd
        @param str arg
        """
        logger.warn('Rejecting unsupported command %r %r' % (cmd, arg))
        return ('502', '5.5.2 Error: command "%s" not recognized' % cmd)

    def data_received(self, data):
        """
        DATA handler

        This method is called when a full message body has been
        received after a DATA call.

        @param str data (mail body)
        """
        data = self.prepend_received_from_line(data)
        data = decode_message(data)
        try:
            ret = self.message_handler(
                self.mail_from, self.recipients, data)
        except Exception:
            return '451', '4.0.0 Unknown message handler error'
        if not ret or len(ret) != 2:
            ret = '250', '2.0.0 Ok'
        self.reset_state()
        return ret

    def smtp_HELO(self, arg):
        """
        Handles SMTP HELO, i.e. the SMTP handshake

        @param str arg hostname
        """
        if not arg:
            return '501', 'Syntax: HELO hostname'
        self.client_greeting = arg
        self.esmtp = False
        return '250', self.fqdn

    def smtp_EHLO(self, arg):
        """
        Handles ESMTP HELO, i.e. the ESMTP handshake

        @param str arg hostname
        """
        if not arg:
            return '501', 'Syntax: EHLO hostname'
        self.client_greeting = arg
        self.esmtp = True
        return '250', [self.fqdn] + list(self.esmtp_features)

    def smtp_RSET(self, arg):
        """
        Handles SMTP RSET, which resets connection state.

        @param str arg '' (must be empty)
        """
        if arg:
            return "501", "5.5.4 Syntax: RSET"
        self.reset_state()
        return "250", "2.0.0 Ok"

    def smtp_MAIL(self, arg):
        """
        Handles SMTP MAIL, which creates a new message/envelope.

        @param arg e.g. FROM:<address>
        """
        if not self.client_greeting:
            return '503', '5.5.1 Error: send HELO/EHLO first'

        if self.mail_from is not None:
            return '203', '5.5.1 Error: nested MAIL command'

        try:
            self.mail_from, self.mail_params = parse_address('from:', arg)
        except InvalidAddressPrefixError:
            return '501', '5.5.4 Syntax: MAIL FROM:<address>'

        return '250', '2.1.0 Ok'

    def smtp_QUIT(self, arg):
        """
        Handles SMTP QUIT, which terminates the connection

        @param arg (anything)
        """
        # we also accept args here
        self.push('221', '2.0.0 Bye')
        # we cannot use 'return' for replying to the client, because
        # we have to close the connection somewhere...
        self.close_when_done()

    def smtp_RCPT(self, arg):
        """
        Handles SMTP RCPT, which adds recipients for the next mail.

        @param arg e.g. TO:<address>
        """
        if self.mail_from is None:
            return '503', '5.5.1 Error: need MAIL command'
        try:
            recipient, params = parse_address('to:', arg)
        except InvalidAddressPrefixError:
            return '501', '5.5.4 Syntax: RCPT TO:<address>'
        self.recipients.append(recipient)
        return '250', '2.1.5 Ok'

    def smtp_DATA(self, arg):
        """
        Handles SMTP DATA, which indicates the beginning of a email body

        @param str arg '' (must be empty)
        """
        if not self.recipients:
            return '503', '5.5.1 Error: need RCPT command'
        if arg:
            return '501', '5.5.4 Syntax: DATA'

        self.set_state(self.DATA)
        return '354', 'End data with <CR><LF>.<CR><LF>'

    def smtp_NOOP(self, arg):
        """
        Handles SMTP NOOP, a do-nothing command (might be used to keep
        the connection from timing out.

        @param str arg anything
        """
        # any arg is accepted
        return '250', '2.0.0 Ok'

    def handle_timeout(self):
        """
        Called when our timeout timer fires, i.e. no data has been
        sent by the client in the configured timeout.
        In this case we gracefully disconnect the client.
        """
        self.push('421', '4.4.2 %s Error: timeout exceeded' % self.fqdn)
        self.handle_close()
        if isinstance(self.remote_addr, tuple):
            logger.info('forcefully disconnecting client (from=%s), '
                'timeout' % self.remote_addr[0]) #pragma: no cover

    def handle_close(self):
        """
        Called by asyncore when the connection is closed.
        """
        self.disconnect()

    def disconnect(self):
        """
        Clean up all network connections
        """
        if self.disconnected:
            return
        self.disconnected = True
        asynchat.async_chat.handle_close(self)
        self.timeout_call.destroy()
        if isinstance(self.remote_addr, tuple):
            logger.info('client (from=%s) disconnecting',
                self.remote_addr[0]) #pragma: no cover

    def prepend_received_from_line(self, message):
        """
        As we act as a mail server, we should also prepend a Received:
        header to each mail to enable simple tracing of this mail's
        routing.

        This method does exactly that.

        @param str message as text
        @return str message as text
        """
        client_ip = (isinstance(self.remote_addr, tuple) and
            self.remote_addr[0])
        header = (
            'Received: from [%s] (helo=%s)\r\n\t'
            'by %s (%s) with %s id PASSTHROUGH;\r\n\t'
            '%s\r\n'
        ) % (
            client_ip, # from
            self.client_greeting, # helo=
            self.fqdn, # by myhostname
            self.server_software_name,
            self.esmtp and 'ESMTP' or 'SMTP', #esmtp
           rfc2822_date(localtime=True)
        )
        return header + message


class ESMTPServer(asyncore.dispatcher):
    """
    A simple, yet extensible (E)SMTP server

    It's probably a good advice to use this server in trusted environments
    for now only (i.e. behind a real SMTP server, e.g. when using it
    as a content filter).
    """

    server_software_name = 'bitbone.esmtpd'

    def __init__(self, handler=ESMTPChannel):
        """Constructor

        @param callable handler: usually a class which handles (E)SMTP
        """
        asyncore.dispatcher.__init__(self)
        self.protocol_handler = handler

    def make_listener(self, address, backlog=5):
        """Creates the server listener by binding to the given address.

        @param (str host, int port) address
        @param int backlog: listen backlog
        """
        try:
            self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
            self.set_reuse_addr()
            self.bind(address)
            asyncore.dispatcher.listen(self, backlog)
        except Exception:
            self.close()
            raise

    def handle_accept(self):
        """Called by asyncore when a new client arrives which
        we should accept."""
        sock, address = self.accept()
        client = self.protocol_handler(sock, address,
                message_handler=self.process_message)
        client.server_software_name = self.server_software_name

    def process_message(self):
        """Callback which is supposed to be overriden and should do
        something with the given message."""
        raise NotImplementedError(
            "No message handler implemented") # pragma: no cover

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
