import logging
from functools import partial
from .esmtpd import ESMTPChannel, ESMTPServer
from .helpers import decode_message, encode_message, FeatureList
from .client import CommandNotSentError

logger = logging.getLogger(__name__)

class ESMTPProxyChannel(ESMTPChannel):
    """
    Implements the protocol-level part of ESMTPProxy.

    For the common operations MAIL, RCPT and DATA this means just
    forwarding the requests and updating internal state accordingly.
    """

    def __init__(self, sock, address, smtp_handler,
            message_handler=None, rewriter=None):
        ESMTPChannel.__init__(self, sock, address,
            message_handler=lambda mail_from, rcpt_tos, message: None)
        """
        Constructor

        @param socket.socket sock
        @param (str host, int port) address
        @param callable smtp_handler
        @param callable message_handler (ignored, just for compatibility)
        @param callable rewriter optional rewriter object
            @see ESMTPProxy.__init__
        """
        self.smtp_handler = smtp_handler
        smtp_handler.want_reply_for(':CLOSE').callback(
            self.handle_upstream_disconnect)
        self.rewriter = rewriter
        smtp_handler.ehlo_or_helo_if_needed()
        self.esmtp_features.add('XFORWARD')

    def smtp_EHLO(self, arg):
        """
        Our client greets us using ESMTP and we are supposed to reply.
        Sadly it's not that easy for us as a proxy.
        There are three cases to consider:

        (1) our handshake (EHLO/HELO) with our upstream gateway is done
            already; in this case we're lucky and can return our
            reply right away
        (2) in the handshake with our upstream gateway, EHLO has been
            sent, but no reply has been received yet;
            in this case, we register a callback for the reply, so that
            we can reply to our client at this time
        (3) in the handshake with our upstream gateway, EHLO has been
            sent and an error code has been received;
            this means our gateway does not support ESMTP, as such we
            can happily send out our reply to our client

        Why that? Because for constructing our reply to our client
        we need to know about the upstream's ESMTP capabilities first.

        @param str arg: client hostname, passed through to ESMTP.smtp_EHLO
        """
        if self.smtp_handler.helo_ehlo_done:
            # case (1)
            return self._send_EHLO_reply(arg)
        try:
            # case (2)
            return self.smtp_handler.want_reply_for('EHLO').callback(
                partial(self._send_EHLO_reply, arg))
        except CommandNotSentError:
            # case (3)
            return self._send_EHLO_reply(arg)

    def _send_EHLO_reply(self, arg, code=None, messages=None):
        # we know about the upstream server's features;
        # and we know our features;
        # the client is shown only the features which are supported both
        # by our implementation and the upstream gateway
        self.esmtp_features = \
            FeatureList(self.esmtp_features).intersection(
            FeatureList(self.smtp_handler.esmtp_features))

        # we also advertise 8BITMIME (we pass data through as-is, and this
        # is 8bit-safe)
        self.esmtp_features.add('8BITMIME')
        self.push(*ESMTPChannel.smtp_EHLO(self, arg))

    def smtp_XFORWARD(self, arg):
        """
        Forwards XFORWARD commands
        """
        self.smtp_handler.docmd('xforward', arg).callback(self.push)

    def smtp_MAIL(self, arg):
        """
        Forwards MAIL commands

        @see ESMTPChannel.smtp_MAIL
        """
        ESMTPChannel.smtp_MAIL(self, arg)
        self.smtp_handler.docmd('mail', arg).callback(self.push)

    def smtp_RCPT(self, arg):
        """
        Forwards RCPT commands

        @see ESMTPChannel.smtp_RCPT
        """
        ESMTPChannel.smtp_RCPT(self, arg)
        self.smtp_handler.docmd('rcpt', arg).callback(self.push)

    def smtp_DATA(self, arg):
        """
        Forwards DATA commands

        @see ESMTPChannel.smtp_DATA
        """
        ESMTPChannel.smtp_DATA(self, arg)
        self.smtp_handler.docmd('data').callback(self.push)

    def data_received(self, data):
        """
        Forwards DATA contents, passing them through the rewriter
        first, if possible.

        @see ESMTPChannel.data_received
        """
        try:
            ESMTPChannel.data_received(self, data)
            return self._process_message(data)
        except Exception:
            # fail gracefully
            logger.error(
                '%s.rewrite_message failed, temporarily rejecting mail',
                self.rewriter.__class__)
            import traceback
            logger.debug('error was:\n%s', traceback.format_exc())
            # reject the mail temporarily
            return ('451', '4.4.0 Requested action aborted: local '
                'error in processing')

    def _process_message(self, data):
        data = self.prepend_received_from_line(data)
        if self.rewriter and hasattr(self.rewriter, 'rewrite_message'):
            data = decode_message(data)
            data = self.rewriter.rewrite_message(data)
            data = encode_message(data)

        # note regearding escaping (line[0]=="." escaping):
        # either, data is passed as we get it, which means that it is
        # escaped by the client; or, data is decoded and encoded once
        # again, where the encoding stage is responsible for encoding
        # exactly the leading dot.
        self.smtp_handler.push(data + "\r\n.\r\n").callback(
            self.push)

    def handle_close(self):
        """
        Terminates both our gateway connection and the local connection
        with our client.

        @see ESMTPChannel.smtp_QUIT
        """
        ESMTPChannel.handle_close(self)
        self.smtp_handler.docmd('quit')
        self.smtp_handler.close_when_done()

    def handle_upstream_disconnect(self, command=None):
        self.push('421', '4.4.0 Service not available, closing '
            'transmission channel (upstream failure)')
        self.handle_close()


class ESMTPProxy(ESMTPServer):
    """
    ESMTPProxy is a special ESMTPServer which forwards all
    MAIL FROM, RCPT TO and DATA commands to an upstream server. It
    also passes back the relevant replies.

    If used unmodified, this server should basically behave like the
    given upstream server.

    An optional rewriter object can be passed in, whose methods will
    be called on different message parts so that these can be manipulated
    before being forwarded.
    This allows for implementation of mail filtering systems.

    The real work is done in the ESMTPProxyChannel, which is used
    as the protocol handler here.
    """
    def __init__(self, smtp_handler, rewriter=None):
        """
        Constructor

        @param callable smtp_handler:
            must support the following methods:
              - putcmd(cmd, arg='')
              - docmd(cmd, arg='')
              - send(data)
              - close()
        @param callable rewriter
            can implement the following method:
              - rewrite_message(data): should return the rewritten message
        """
        self.smtp_handler = smtp_handler
        ESMTPServer.__init__(self, handler=self._make_smtp_handler)

    def _make_smtp_handler(self, *args, **kwargs):
        """
        Internal helper method which acts as an SMTP handler factory.

        @return callable
        """
        handler = ESMTPProxyChannel(
            *args,
            smtp_handler=self.smtp_handler(),
            rewriter=self,
            **kwargs)
        return handler

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
