import smtplib

class InvalidAddressPrefixError(Exception):
    """Thrown by parse_address() if the given value does not start
    with the expected prefix."""

def parse_address(prefix, data):
    """
    Internal helper method for parsing addresses in the format
    PREFIX:<address> with some variations.

    Useful for RCPT TO: and MAIL FROM:

    @param str prefix
    @param str data
    @return (address, params)
    """
    if not data.lower().startswith(prefix):
        raise InvalidAddressPrefixError('Expected prefix "%s"' % prefix)

    data = data[len(prefix):]

    params = []
    address = ""
    value = ""
    quoted = False
    escaped = False

    ST_ADDRESS = 0
    ST_ROUTE = 1
    ST_PARAM = 2
    state = ST_ADDRESS

    data_len = len(data)
    idx = 0
    while idx < data_len:
        char = data[idx]
        idx += 1
        if char == '"':
            quoted = not quoted
            continue

        if char == '\\':
            if not escaped:
                value += data[idx]
                idx += 1
            escaped = not escaped
            continue

        if quoted:
            value += char
            continue

        if escaped:
            value += char
            escaped = False
            continue

        if char == ' ' and not value:
            # leading space, will just be stripped
            continue

        if state == ST_ADDRESS:
            if char == '<':
                # strip all < in mail addresses
                continue
            if char == '@' and not value:
                value = char
                state = ST_ROUTE
                continue
            if char in '> ':
                # mail address done
                address = value
                state = ST_PARAM
                value = ""
                continue

        elif state == ST_ROUTE:
            if char == ':':
                state = ST_ADDRESS
                # we could save the route here, but nobody uses
                # this anyway...
                value = ""
                continue

        elif state == ST_PARAM:
            if char == ' ':
                params.append(value)
                value = ""
                continue

        value += char

    if value:
        if state == ST_ADDRESS:
            address = value
        elif state == ST_PARAM:
            params.append(value)

    return address, params


def decode_message(data):
    """
    Decodes a message body according to RFC 5321 4.5.2.

    This effectiveley strips a leading dot from any lines containing
    leading dot(s).

    @param str data
    @return str data
    """
    lines = data.split('\r\n')

    for idx, line in enumerate(lines):
        if line and line[0] == '.':
            lines[idx] = line[1:]
    return '\r\n'.join(lines)


"""
Encodes a message body according to RFC 5321 4.5.2.

This effectiveley adds a leading dot to any leading containing
leading dot(s) at the beginning.

@param str data
@return str data
"""
encode_message = smtplib.quotedata


class FeatureList(set):
    """
    This is a set-derived model which is used to store (E)SMTP features.
    The difference to a normal set is, that it overloads the "in"
    operator so that feature comparison happens case-insensitively and
    in a way where "XFORWARD" in FeatureList(["XFORWARD NAME"]) also
    returns True.
    The .intersection() method is also overriden to behave the same
    as described above.
    """
    def __contains__(self, key):
        """
        Returns True, if the given (E)SMTP feature is in this
        FeatureList.

        'feature' in FeatureList(['FEATURE']) -> True
        'FEATURE' in FeatureList(['feature']) -> True
        'FEAture' in FeatureList(['feature xyz']) -> True

        @param str key
        @return bool
        """
        if set.__contains__(self, key):
            return True
        for my_feature in self:
            if (self._normalize_feature(my_feature) ==
                    self._normalize_feature(key)):
                return True
        return False

    def _normalize_feature(self, name):
        """
        Internal helper which normalizes a feature name in a way
        useful for comparisons.

        @param str name
        """
        space = name.find(' ')
        if space != -1:
            name = name[:space]
        return name.upper()

    def intersection(self, otherset):
        """
        Returns all features which are both in the otherset and in ours.
        Values are always taken from the otherset.

        @see set.intersection
        @param set otherset
        @return FeatureList
        """
        newset = self.__class__()
        for otheritem in otherset:
            if otheritem in self:
                newset.add(otheritem)

        return newset

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
