import os
import shutil
import unittest
from tempfile import mkdtemp
os.environ['TEST'] = '1'
from ..mock import Mock
from ..utils import replace_global_symbols

def getgrouplist_mock(username, gid):
    return [2002, 2000, gid]

class LdapMock(object):
    uri = ''
    binddn = ''
    bindpw = ''
    lastsearch = None

    def __init__(self, uri=uri, **kwargs):
        self.uri = uri

    def simple_bind_s(self, dn, pw):
        self.binddn = dn
        self.bindpw = pw

    def search_st(self, *args, **kwargs):
        if 'fail@' in args[2]:
            return []
        self.lastsearch = args
        uid = 'x'
        if 'failuser' in args[2]:
            uid = 'failuser'
        if 'referral' in args[2]:
            return [[None, {'foo': 'bar'}]]
        if 'nouid' in args[2]:
            return [['something', {'uid': []}]]

        if 'testAttr' in args[3]:
            return [['uid=%s,dc=local' % uid, {'testAttr': ['123']}]]

        if 'return-bad-utf8' in args[2]:
            return [['uid=%s' % uid, {'sn': ['\xcd']}]]

        if 'template.org' in args[2]:
            return [['something', {'maildisclaimerTemplate': ["sometemplate"]}]]

        ret = [
            ['uid=%s,dc=local' % uid, {
                'uid': [uid],
                'telephoneNumber': ['1234'],
                'facsimileTelephoneNumber': ['5678'],
                'displayName': ['Max Mustermann'],
                'givenName': ['Max'],
                'sn': ['Mustermann'],
                'departmentNumber': ['IT'],
                'physicalDeliveryOfficeName': ['3. OG']
            }]
        ]
        if 'emptygivennamelist' in args[2]:
            ret[0][1]['givenName'] = []
        if 'emptygivennamefield' in args[2]:
            ret[0][1]['givenName'] = ['']

        return ret

    def set_option(self, key, value):
        pass

LdapMock.ReconnectLDAPObject = LdapMock

def getpwnam_mock(username):
    if username == 'failuser':
        raise KeyError
    ret = Mock()
    ret.pw_name = 'x'
    ret.pw_uid = 1001
    ret.pw_gid = 2001
    return ret

def getgrgid_mock(gid):
    ret = Mock()
    ret.gr_gid = gid
    ret.gr_mem = ['unknown']
    ret.gr_name = '3unrelated'
    if gid == 2001:
        ret.gr_mem = []
        ret.gr_name = '2primarygroup'
    elif gid == 2002:
        ret.gr_mem = ['x']
        ret.gr_name = '1secondarygroup'
    return ret


SCOPE_SUBTREE = 999
with replace_global_symbols({
    'ldap.OPT_REFERRALS': None,
    'ldap.OPT_NETWORK_TIMEOUT': 3,
    'ldap.SCOPE_SUBTREE': SCOPE_SUBTREE,
    'ldap.ldapobject': LdapMock,
    'pwd.getpwnam': getpwnam_mock,
    'grp.getgrgid': getgrgid_mock,
    'bitbone.utils.getgrouplist': getgrouplist_mock}):
    from ...accounts.base import BaseAccountLookup
    from ...accounts.passwd import AccountLookup
    from ...accounts.ldap import AccountLookup as LdapAccountLookup


class BaseAccountLookupTest(unittest.TestCase):
    def setUp(self):
        self.tmpdir = mkdtemp()
        os.makedirs(os.path.join(self.tmpdir, 'etc', 'zarafa'))
        self.ldap_conf = os.path.join(self.tmpdir, 'etc', 'zarafa',
                'ldap.bitbone.cfg')
        f = open(self.ldap_conf, 'wb')
        f.write('ldap_protocol = ldaps\n')
        f.write('ldap_host = localhost\n')
        f.write('ldap_port = 639\n')
        f.write('ldap_server_charset = utf-8\n')
        f.write('ldap_bind_user = uid=nobody,dc=local\n')
        f.write('ldap_bind_passwd = s3cr3t\n')
        f.write('ldap_search_base = dc=local\n')
        f.write('ldap_emailaddress_attribute = mail\n')
        f.write('ldap_emailaliases_attribute = zarafaAliases\n')
        f.write('ldap_user_search_filter = (objectClass=user)\n')
        f.write('ldap_loginname_attribute = uid\n')
        f.write('ldap_fullname_attribute = displayName\n')
        f.write('ldap_test_binattr_attribute = objectGuid\n')
        f.write('ldap_test_binattr_type = binary\n')
        f.write('ldap_user_unique_attribute = objectGuid\n')
        f.write('ldap_user_unique_type = binary\n')
        f.close()

    def tearDown(self):
        if self.tmpdir != '/' and '..' not in self.tmpdir:
            shutil.rmtree(self.tmpdir)

    def make_account_lookup(self):
        return AccountLookup(ldap_config=self.ldap_conf)

class TestAccountLookupBinaryAttrs(BaseAccountLookupTest):
    def test_binary_attrs(self):
        l = BaseAccountLookup(ldap_config=self.ldap_conf)
        self.assertEquals(l.binary_attributes, ['objectguid'])


class TestLdapAccountLookupMemberAttr(BaseAccountLookupTest):
    def test_member_attr_explicit(self):
        f = open(self.ldap_conf, 'ab')
        f.write('ldap_groupmembers_relation_attribute = cn\n')
        f.close()
        l = LdapAccountLookup(ldap_config=self.ldap_conf)
        # from ldap_user_unique_attribute as per Zarafa docs/code
        self.assertEquals(l.ldap_member_attr, 'cn')

    def test_member_attr_implicit(self):
        l = LdapAccountLookup(ldap_config=self.ldap_conf)
        # from ldap_user_unique_attribute as per Zarafa docs/code
        self.assertEquals(l.ldap_member_attr, 'objectGuid')

    def test_member_attr_type_dn(self):
        f = open(self.ldap_conf, 'ab')
        f.write('ldap_groupmembers_attribute_type = dn\n')
        f.close()
        l = LdapAccountLookup(ldap_config=self.ldap_conf)
        # from ldap_user_unique_attribute as per Zarafa docs/code
        self.assertEquals(l.ldap_member_attr, 'dn')

    def test_domain_template_without_config(self):
        l = LdapAccountLookup(ldap_config=self.ldap_conf)
        res = l.get_domain_template("test@example.org")
        self.assertEquals(res, None)

    def test_domain_template_with_config(self):
        f = open(self.ldap_conf, 'ab')
        f.write('ldap_domain_attribute = domain\n')
        f.write('ldap_domain_search_filter = (domain=%s)\n')
        f.close()
        l = LdapAccountLookup(ldap_config=self.ldap_conf)
        res = l.get_domain_template("test@example.org")
        self.assertEquals(res, [None])

    def test_domain_template(self):
        f = open(self.ldap_conf, 'ab')
        f.write('ldap_domain_attribute = domain\n')
        f.write('ldap_domain_search_filter = (domain=%s)\n')
        f.close()
        l = LdapAccountLookup(ldap_config=self.ldap_conf)
        res = l.get_domain_template("test@template.org")
        self.assertEquals(res, ['sometemplate'])



class TestAccountLookupAttrs(BaseAccountLookupTest):
    def test(self):
        ac = self.make_account_lookup()
        res = ac.get_user_attributes_by_email("test@example.org", ["testAttr"])
        self.assertEquals(ac.ldap.uri, 'ldaps://localhost:639')
        self.assertEquals(ac.ldap.binddn, 'uid=nobody,dc=local')
        self.assertEquals(ac.ldap.bindpw, 's3cr3t')
        self.assertEquals(ac.ldap.lastsearch, (
            'dc=local', SCOPE_SUBTREE,
            '(&(|(mail=test@example.org)(zarafaAliases=test@example.org))'
            '(objectClass=user))',
            ['testAttr']))
        self.assertEquals(res, {'dn': 'uid=x,dc=local', 'testAttr': '123'})

class TestAccountLookup(BaseAccountLookupTest):
    def test(self):
        ac = self.make_account_lookup()
        res = ac.get_objectinfo_by_email("test@example.org")
        self.assertEquals(ac.ldap.uri, 'ldaps://localhost:639')
        self.assertEquals(ac.ldap.binddn, 'uid=nobody,dc=local')
        self.assertEquals(ac.ldap.bindpw, 's3cr3t')
        self.assertEquals(ac.ldap.lastsearch, (
            'dc=local', SCOPE_SUBTREE,
            '(&(|(mail=test@example.org)(zarafaAliases=test@example.org))'
            '(objectClass=user))',
            ['uid', 'displayName', 'givenName', 'sn', 'departmentNumber',
            'department', 'company',
            'telephoneNumber', 'facsimileTelephoneNumber',
            'physicalDeliveryOfficeName', 'mail', 'st', 'co',
            'postalCode', 'postOfficeBox', 'streetAddress', 'l', 'maildisclaimerTemplate']))
        self.assertEquals(res,
            {
                'system:uid': 1001,
                'system:gids': [2001, 2002, 2000],
                'department': u'IT',
                'fax': u'5678',
                'fullname': u'Max Mustermann',
                'givenname': u'Max',
                'lastname': u'Mustermann',
                'office': u'3. OG',
                'phone': u'1234',
                'pobox': None,
                'state': None,
                'street': None,
                'zip': None,
                'city': None,
                'company': None,
                'country': None,
                'email': None,
                'template:user': None,
            }
        )

    def test_userinfo_fail(self):
        ac = self.make_account_lookup()
        res = ac.get_objectinfo_by_email("fail@example.org")
        self.assertEquals(res, {})

    def test_fail_getpwd(self):
        ac = self.make_account_lookup()
        res = ac.get_objectinfo_by_email("failuser@example.org")
        self.assertEquals(res,
            {
                'department': u'IT',
                'fax': u'5678',
                'fullname': u'Max Mustermann',
                'givenname': u'Max',
                'lastname': u'Mustermann',
                'office': u'3. OG',
                'phone': u'1234',
                'pobox': None,
                'state': None,
                'street': None,
                'zip': None,
                'city': None,
                'company': None,
                'country': None,
                'email': None,
                'template:user': None,
            }
        )

    def test_fail_referral(self):
        ac = self.make_account_lookup()
        res = ac.get_objectinfo_by_email("referral@example.org")
        self.assertEquals(res, {})

    def test_no_uid(self):
        ac = self.make_account_lookup()
        res = ac.get_objectinfo_by_email("nouid@example.org")
        self.assertEquals(res, {})

    def test_bad_utf8_attribute(self):
        ac = self.make_account_lookup()
        res = ac.get_objectinfo_by_email("return-bad-utf8@example.org")
        self.assertEquals(res, {})

    def test_empty_givenname_list(self):
        ac = self.make_account_lookup()
        res = ac.get_objectinfo_by_email("emptygivennamelist@example.org")
        self.assertEquals(res,
            {
                'system:uid': 1001,
                'system:gids': [2001, 2002, 2000],
                'department': u'IT',
                'fax': u'5678',
                'fullname': u'Max Mustermann',
                'givenname': None,
                'lastname': u'Mustermann',
                'office': u'3. OG',
                'phone': u'1234',
                'pobox': None,
                'state': None,
                'street': None,
                'zip': None,
                'city': None,
                'company': None,
                'country': None,
                'email': None,
                'template:user': None,
            }
        )

    def test_empty_givenname_field(self):
        ac = self.make_account_lookup()
        res = ac.get_objectinfo_by_email("emptygivennamefield@example.org")
        self.assertEquals(res,
            {
                'system:uid': 1001,
                'system:gids': [2001, 2002, 2000],
                'department': u'IT',
                'fax': u'5678',
                'fullname': u'Max Mustermann',
                'givenname': None,
                'lastname': u'Mustermann',
                'office': u'3. OG',
                'phone': u'1234',
                'pobox': None,
                'state': None,
                'street': None,
                'zip': None,
                'city': None,
                'company': None,
                'country': None,
                'email': None,
                'template:user': None,
            }
        )

class TestAccountLookupWithEmptyLDAPUri(TestAccountLookup):
    def setUp(self):
        TestAccountLookup.setUp(self)
        f = open(os.path.join(self.tmpdir, 'etc', 'zarafa',
                'ldap.bitbone.cfg'), 'ab')
        f.write("""
ldap_uri = \n
ldap_host = localhost
""")
        f.close()

    def test(self):
        ac = self.make_account_lookup()
        res = ac.get_objectinfo_by_email("test@example.org")
        self.assertEquals(ac.ldap.uri, 'ldaps://localhost:639')


class TestAccountLookupWithSingleLDAPUri(TestAccountLookup):
    def setUp(self):
        TestAccountLookup.setUp(self)
        f = open(os.path.join(self.tmpdir, 'etc', 'zarafa',
                'ldap.bitbone.cfg'), 'ab')
        f.write('ldap_uri = ldaps://127.0.0.1\n')
        f.close()

    def test(self):
        ac = self.make_account_lookup()
        res = ac.get_objectinfo_by_email("test@example.org")
        self.assertEquals(ac.ldap.uri, 'ldaps://127.0.0.1')


class TestAccountLookupWithMultiLDAPUri(TestAccountLookup):
    def setUp(self):
        TestAccountLookup.setUp(self)
        f = open(os.path.join(self.tmpdir, 'etc', 'zarafa',
                'ldap.bitbone.cfg'), 'ab')
        f.write('ldap_uri = ldaps://127.0.0.1 ldap://127.0.0.2\n')
        f.close()

    def test(self):
        # untested if this makes sense, i.e. if python-ldap supports this
        # syntax
        ac = self.make_account_lookup()
        res = ac.get_objectinfo_by_email("test@example.org")
        self.assertEquals(ac.ldap.uri, 'ldaps://127.0.0.1 '
            'ldap://127.0.0.2')

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
