# -*- coding: utf-8 -*-
from __future__ import absolute_import
import os
import sys
import re
import unittest
from email.message import Message
from email.mime.text import MIMEText
from email import message_from_string
from ....disclaimers.email import (DisclaimerInjector, Embedder,
        EmbedderFailure,
        extract_sender_from_header, extract_targets_from_header,
        extract_content_type_from_header, get_domain_from_header)

def get_sample(name):
    input = os.path.join(os.path.dirname(__file__), 'samples',
            '%s.in.eml' % name)
    output = os.path.join(os.path.dirname(__file__), 'samples',
            '%s.out.eml' % name)

    input_text = open(input, 'rb').read()
    input_object = message_from_string(input_text)
    output_text = open(output, 'rb').read().rstrip('\n')
    output_text = fix_boundary_lf(output_text)
    return input_object, output_text

def dump_eml(email, id):
    """Dumps all expected data to $DUMPEML, if given, so that it can be
    manually checked and moved to samples/ if it is ok."""
    if 'DUMPEML' not in os.environ:
        return email
    path = os.path.join(os.environ['DUMPEML'], '%s.eml' % id)
    fd = open(path, 'wb')
    fd.write(email)
    fd.close()
    return email

def normalize(email):
    """Heuristic which replaces generated boundaries and Content-IDs by a
    static string so that we can test against it.

    IOW: makes the test deterministic"""
    boundaries = []
    def replace_boundaries(match):
        b = match.group(2)
        if b not in boundaries:
            boundaries.append(b)
        return '%sBOUNDARY%s%s' % (
            match.group(1), boundaries.index(b), match.group(3))

    email = re.sub(r'("|--)(===============\d+==)(\r|\n|"|--)',
        replace_boundaries, email)

    # compatibility adjustments (py-2.6/2.7)
    # unfold Content-Type etc.
    email = re.sub(r'; ?\r?\n\s+(protocol|boundary)', r'; \1', email)

    # >=python-2.7 uses a single space for line continuations,
    # earlier versions use a single tab. this breaks our test cases,
    # so we convert the tab to a single space, if it appears at the
    # beginning of a header line
    email = re.sub('(?<!\n\n)(\n)\t(\S)', r'\1 \2', email, re.MULTILINE)
    return email.rstrip('\n')

FIX_BOUNDARY_LF_RE = re.compile(r"(--BOUNDARY\d+--\r?\n)\r?\n(--BOUNDARY\d+--)")
FIX_BOUNDARY_LF_RE = re.compile(r"(--BOUNDARY\d+--\r?\n)(?:\r?\n)")

def fix_boundary_lf(email):
    """
    Current python versions (>=2.7.7) add new lines between
    MIME boundaries while previous versions don't.

    See http://bugs.python.org/issue14983

    This heuristic adjusts the expected test output so that depending
    on the python version the correct output is chosen.
    """
    if sys.version_info >= (2, 7, 7):
        # Test cases have been updated to work with >=2.7.7, so
        # don't do anything here.
        return email

    return FIX_BOUNDARY_LF_RE.sub(r"\1", email)

dummy_txt = "-- \nDisclaimer"
dummy_html = "<i>-- <br>\n<b>Disclaimer</b></i>"

class TestDisclaimerInjector(unittest.TestCase):
    def build_test(id):
        def test(self):
            i = DisclaimerInjector(dummy_txt, dummy_html)
            input, output = get_sample(id)
            self.assertMultiLineEqual(dump_eml(
                    normalize(i.process(input)), id), fix_boundary_lf(output))
        return test

    test_plaintext = build_test("plaintext")
    test_plaintext_very_long_line = build_test("plaintext.very-long-line")
    test_plaintext_keep_binary = build_test("plaintext.keep-binary")
    test_html = build_test("html")
    test_multipart_alternative_html_zarafa = \
        build_test("multipart-alternative.html.zarafa")
    test_multipart_alternative_html_thunderbird = \
        build_test("multipart-alternative.html.thunderbird")
    test_multipart_mixed_plaintext_attachment_gif_thunderbird = \
        build_test("multipart-mixed.plaintext.attachment-gif.thunderbird")
    test_multipart_mixed_plaintext_attachment_txthtml_thunderbird = \
        build_test("multipart-mixed.plaintext.attachment-txt-"
                "html.thunderbird")
    test_multipart_mixed_plaintext_forwarded_plaintext_thunderbird = \
        build_test("multipart-mixed.plaintext.forwarded-plaintext."
                "thunderbird")
    test_plaintext_pgp_signed_inline_thunderbird = \
        build_test("plaintext.pgp-signed-inline.thunderbird")

    test_plaintext_pgp_signed_mime_thunderbird = \
        build_test("plaintext.pgp-signed-mime.thunderbird")

    test_plaintext_smime_signed_thunderbird = \
        build_test("plaintext.smime-signed.thunderbird")

    test_plaintext_pgp_encrypted_mime_thunderbird = \
        build_test("plaintext.pgp-encrypted-mime.thunderbird")

    test_multipart_mixed_plaintext_gpg_signed_thunderbird = \
        build_test("multipart-mixed.plaintext.gpg-signed.thunderbird")

    test_basic_plaintext = \
        build_test("basic-plaintext")

    test_bad_base64 = \
        build_test("bad-base64")

    test_encoding_rot13 = \
        build_test("encoding-rot13")

    test_py_issue8769_subjects = \
        build_test("py-issue8769-subjects")

    test_py_issue8769_subjects_ascii = \
        build_test("py-issue8769-subjects-ascii")

    test_no_content_type_invalid_utf8 = \
        build_test("no-content-type+invalid-utf8")

    test_encoding_binary = \
        build_test("encoding-binary")

    test_subject_body_encoding_py26 = \
        build_test("subject-body-encoding-py26")

    def test_html_disclaimer_image(self):
        html = """
<html>
  <body>
    <img src="/maildisclaimers/images.cgi?show=somepic.gif" />
  </body>
</html>
"""
        def get_image_path(path):
            return gif

        id = "html.disclaimer-image"

        i = DisclaimerInjector(dummy_txt, html,
            image_resolver=get_image_path)
        input, output = get_sample(id)
        gif = os.path.join(os.path.dirname(__file__), 'samples',
                '%s.gif' % id)

        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_html_disclaimer_image_special_char(self):
        html = u"""
<html>
  <body>
    <img src="/maildisclaimers/images.cgi?show=html.disclaimer-image-special-%3F-%C3%A4-%C3%B6-%C3%BC.gif" />
  </body>
</html>
"""
        def get_image_path(path):
            return os.path.join(os.path.dirname(__file__), 'samples',
                path)


        id = "html.disclaimer-image-special-char"

        i = DisclaimerInjector(dummy_txt, html,
            image_resolver=get_image_path,
            prefer_direct_embedding=True)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_plaintext_always_embed_html(self):
        id = "plaintext.always-embed-html"
        i = DisclaimerInjector("Text-Disclaimer", "<b>HTML</b>-Disclaimer",
            always_embed_html=True)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_plaintext_prefer_direct_embedding(self):
        id = "plaintext.prefer-direct-embedding"
        i = DisclaimerInjector("Text-Disclaimer", "<b>HTML</b>-Disclaimer",
            prefer_direct_embedding=True)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_html_prefer_direct_embedding(self):
        id = "html.prefer-direct-embedding"
        i = DisclaimerInjector("Text-Disclaimer", "<b>HTML</b>-Disclaimer",
            prefer_direct_embedding=True)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_avoid_quoted_printable(self):
        id = "avoid-quoted-printable"
        i = DisclaimerInjector("Text-Disclaimer", "<b>HTML</b>-Disclaimer",
            prefer_direct_embedding=True, avoid_quoted_printable=True)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_html_without_body_prefer_direct_embedding(self):
        id = "html.without-body.prefer-direct-embedding"
        i = DisclaimerInjector("Text-Disclaimer", "<b>HTML</b>-Disclaimer",
            prefer_direct_embedding=True)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_html_with_prefer_direct_embedding_and_image(self):
        id = "html.prefer-direct-embedding.image"
        html = """<b>HTML</b>-Disclaimer
<img src="/maildisclaimers/images.cgi?show=somepic.gif" />
"""
        def get_image_path(path):
            return os.path.join(os.path.dirname(__file__), 'samples',
                '%s.gif' % "html.disclaimer-image")

        i = DisclaimerInjector("Text-Disclaimer", html,
            image_resolver=get_image_path,
            prefer_direct_embedding=True)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_direct_embedding_latin1_upconvert(self):
        """
        This test checks if non-UTF-8 mails get their
        disclaimer properly embedded if the disclaimer contains
        UTF-8 characters which are not existing in the target
        encoding (here: latin1)
        """
        id = "direct-embedding.latin1-upconvert"
        txt = u"Signature with UTF-8 char: \u2013"
        i = DisclaimerInjector(txt, txt,
            prefer_direct_embedding=True)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_multipart_alternative_prefer_direct_embedding_image(self):
        id = "multipart-alternative.prefer-direct-embedding.image"
        html = """<b>HTML</b>-Disclaimer
<img src="/maildisclaimers/images.cgi?show=somepic.gif" />
"""
        def get_image_path(path):
            return os.path.join(os.path.dirname(__file__), 'samples',
                '%s.gif' % "html.disclaimer-image")

        i = DisclaimerInjector("Text-Disclaimer", html,
            image_resolver=get_image_path,
            prefer_direct_embedding=True)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_multipart_mixed_prefer_direct_embedding_image(self):
        id = "multipart-mixed.prefer-direct-embedding.txt+html.image"
        html = """<br/><b>HTML</b>-Disclaimer
<img src="/maildisclaimers/images.cgi?show=somepic.gif" />
"""
        def get_image_path(path):
            return os.path.join(os.path.dirname(__file__), 'samples',
                '%s.gif' % "html.disclaimer-image")

        i = DisclaimerInjector("Text-Disclaimer", html,
            image_resolver=get_image_path,
            prefer_direct_embedding=True)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_multipart_alternative_prefer_direct_embedding(self):
        id = "multipart-alternative.prefer-direct-embedding"
        i = DisclaimerInjector("Text-Disclaimer", "<b>HTML</b>-Disclaimer",
            prefer_direct_embedding=True)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_multipart_alternative_calendar_prefer_direct_embedding(self):
        id = "multipart-alternative-calendar.prefer-direct-embedding"
        i = DisclaimerInjector("Text-Disclaimer", "<b>HTML</b>-Disclaimer",
            prefer_direct_embedding=True)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_multipart_mixed_prefer_direct_embedding_txt_html(self):
        id = "multipart-mixed.prefer-direct-embedding.txt+html"
        i = DisclaimerInjector("Text-Disclaimer", "<b>HTML</b>-Disclaimer",
            prefer_direct_embedding=True)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_multipart_alternative_related_prefer_direct_embedding(self):
        id = "multipart-alternative-related.prefer-direct-embedding"
        i = DisclaimerInjector("Text-Disclaimer", "<b>HTML</b>-Disclaimer",
            prefer_direct_embedding=True)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_html_prefer_direct_embedding_strip_body(self):
        id = "html.prefer-direct-embedding.strip-body"
        i = DisclaimerInjector("Text-Disclaimer",
            "<html>\n<body><b>HTML</b>-Disclaimer</body>\n</html>",
            prefer_direct_embedding=True)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_charset_disclaimer_plaintext_upconvert(self):
        id = "plaintext.disclaimer-charsets"
        txt = u"Mäx Müstermäñńß"
        i = DisclaimerInjector(txt, txt)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_charset_disclaimer_plaintext_upconvert_uppercase(self):
        id = "plaintext.disclaimer-charsets.uppercase"
        txt = u"Mäx Müstermäñńß"
        i = DisclaimerInjector(txt, txt)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_multipart_alternative_latin1(self):
        id = "multipart-alternative.latin1"
        txt = u"Mäx Müstermäñńß"
        i = DisclaimerInjector(txt, txt)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_win874_content(self):
        id = "win874"
        txt = u"Mäx Müstermäñńß"
        i = DisclaimerInjector(txt, txt, prefer_direct_embedding=True)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_plaintext_alternative_latin1(self):
        id = "plaintext.latin1"
        txt = u"Mäx Müstermäñńß"
        i = DisclaimerInjector(txt, txt)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_charset_disclaimer_mime(self):
        id = "multipart-alternative.disclaimer-charsets.thunderbird"
        txt = u"Mäx Müstermäñńß"
        i = DisclaimerInjector(txt, txt)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_charset_body(self):
        id = "plaintext.charsets.kmail"
        txt = u"Mäx Müstermäñńß"
        i = DisclaimerInjector(txt, txt)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_no_body_charset(self):
        id = "plaintext.no-charset"
        txt = u"Mäx Müstermäñńß"
        i = DisclaimerInjector(txt, txt)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_wrong_body_charset(self):
        id = "plaintext.wrong-charset"
        txt = u"Mäx Müstermäñńß"
        i = DisclaimerInjector(txt, txt)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_unknown_body_charset(self):
        id = "plaintext.unknown-charset"
        txt = u"Mäx Müstermäñńß"
        i = DisclaimerInjector(txt, txt)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_undecodable_body_charset(self):
        # yes, this test produces a non-readable EML file
        # (as is the input), at least in thunderbird
        id = "plaintext.undecodable-charset"
        txt = u"Mäx Müstermäñńß"
        i = DisclaimerInjector(txt, txt)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_unencodable_body_charset(self):
        id = "plaintext.unencodable-charset"
        txt = u"Mäx Müstermäñńß"
        i = DisclaimerInjector(txt, txt)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_prefer_direct_embedding_fail_and_dump(self):
        id = "bad-base64.direct-embedding"
        dumps = []
        def test():
            dumps.append(1)
        i = DisclaimerInjector(
            u"-- \nDisclaimer 8 \u2013 9 Uhr",
            u"<i>-- <br>\n<b>Disclaimer</b></i> 8 \u2013 9 Uhr",
            prefer_direct_embedding=True,
            request_email_dump=test)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)
        self.assertEquals(len(dumps), 1)

    def test_unknown_content_type(self):
        # at least thunderbird doesn't do anything useful with this,
        # but let's at least make sure that we handle this unlikeley
        # case rationally
        id = "unknown-type"
        i = DisclaimerInjector(dummy_txt, dummy_html)
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_html_mail_no_disclaimers_1(self):
        id = "empty-txt+html-disclaimer"
        i = DisclaimerInjector('', '')
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_html_mail_no_disclaimers_2(self):
        id = "empty-txt+html-disclaimer"
        i = DisclaimerInjector(None, '')
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_html_mail_no_disclaimers_3(self):
        id = "empty-txt+html-disclaimer"
        i = DisclaimerInjector('\n', '')
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_html_mail_only_plain_disclaimer(self):
        id = "empty-html-disclaimer"
        i = DisclaimerInjector(dummy_txt, '')
        input, output = get_sample(id)
        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_html_with_sig_reprocess(self):
        html = """
<html>
  <body>
    <img src="/maildisclaimers/images.cgi?show=somepic.gif" />
  </body>
</html>
"""
        def get_image_path(path):
            return gif

        id = "html.with-sig.reprocess"
        i = DisclaimerInjector(dummy_txt, html,
            image_resolver=get_image_path,
            prefer_direct_embedding=True)
        input, output = get_sample(id)
        gif = os.path.join(os.path.dirname(__file__), 'samples',
                "html.disclaimer-image.gif")

        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_html_with_same_sig_reprocess(self):
        html = """
<html>
  <body>
    <img src="/maildisclaimers/images.cgi?show=somepic.gif" />
  </body>
</html>
"""
        def get_image_path(path):
            return gif

        id = "html.with-same-sig.reprocess"
        i = DisclaimerInjector(dummy_txt, html,
            image_resolver=get_image_path,
            prefer_direct_embedding=True)
        input, output = get_sample(id)
        gif = os.path.join(os.path.dirname(__file__), 'samples',
                "html.disclaimer-image.gif")

        self.assertMultiLineEqual(dump_eml(
                normalize(i.process(input)), id), output)

    def test_html_with_same_sig_reprocess_cid(self):
        html = """
<html>
  <body>
    <img src="/maildisclaimers/images.cgi?show=somepic.gif" />
  </body>
</html>
"""
        def get_image_path(path):
            return gif

        id = "html.with-same-sig.reprocess.cid"
        i = DisclaimerInjector(dummy_txt, html,
            image_resolver=get_image_path,
            prefer_direct_embedding=True)
        input, output = get_sample(id)
        gif = os.path.join(os.path.dirname(__file__), 'samples',
                "html.disclaimer-image.gif")

        self.assertMultiLineEqual(dump_eml(
                fix_boundary_lf(normalize(i.process(input))), id), output)


class TestExtractSenderFromHeader(unittest.TestCase):
    def test_simple(self):
        email = Message()
        email.add_header('From', 'foo@example.org')
        self.assertEquals(extract_sender_from_header(email),
            'foo@example.org')

    def test_realname(self):
        email = Message()
        email.add_header('From',
            'Someone with Some name <someone@example.org>')
        self.assertEquals(extract_sender_from_header(email),
            'someone@example.org')

    def test_realname_postfixed(self):
        email = Message()
        email.add_header('From',
            'strangeguy@example.org (Strange guy)')
        self.assertEquals(extract_sender_from_header(email),
            'strangeguy@example.org')

    def test_from_with_quotes(self):
        email = Message()
        email.add_header('From',
            '"s.britt1124."@mocha.ocn.ne.jp')
        self.assertEquals(extract_sender_from_header(email),
            's.britt1124.@mocha.ocn.ne.jp')

    def test_from_lowercase(self):
        email = Message()
        email.add_header('From',
            'uPPEr@cASe.example')
        self.assertEquals(extract_sender_from_header(email),
            'upper@case.example')

    def test_from_with_included_quote(self):
        email = Message()
        email.add_header('From',
            r'"test\"test"@test')
        self.assertEquals(extract_sender_from_header(email),
            'test"test@test')

    def test_no_from(self):
        email = Message()
        self.assertEquals(extract_sender_from_header(email),
            None)

    def test_outlook_quotes(self):
        email = Message()
        email.add_header('From',
            '"Max Mustermann" <mustermann@example.com>')
        self.assertEquals(extract_sender_from_header(email),
            'mustermann@example.com')


class TestExtractTargetsFromHeader(unittest.TestCase):
    def test_simple(self):
        email = Message()
        email.add_header('To', 'foo@example.org')
        self.assertEquals(extract_targets_from_header(email),
            ['foo@example.org'])

    def test_realname(self):
        email = Message()
        email.add_header('To',
            'Someone with Some name <someone@example.org>')
        self.assertEquals(extract_targets_from_header(email),
            ['someone@example.org'])

    def test_realname_postfixed(self):
        email = Message()
        email.add_header('To',
            'strangeguy@example.org (Strange guy)')
        self.assertEquals(extract_targets_from_header(email),
            ['strangeguy@example.org'])

    def test_no_to(self):
        email = Message()
        self.assertEquals(extract_targets_from_header(email),
            [])

    def test_multiple(self):
        email = Message()
        email['To'] = 'user1@example.org, user2@example.org'
        self.assertEquals(extract_targets_from_header(email),
            ['user1@example.org', 'user2@example.org'])

    def test_multiple_with_realname_semikolon(self):
        email = Message()
        email['To'] = ('User1 :) <user1@example.org>; '
            'User2 :( <user2@example.org>')
        self.assertEquals(extract_targets_from_header(email),
            ['user1@example.org', 'user2@example.org'])

    def test_cc(self):
        email = Message()
        email['To'] = 'to@example.org'
        email['Cc'] = 'cc@example.org'
        self.assertEquals(extract_targets_from_header(email),
            ['to@example.org', 'cc@example.org'])

    def test_distribution_list_1(self):
        email = Message()
        email['To'] = 'alle:'
        self.assertEquals(extract_targets_from_header(email),
            [])

    def test_distribution_list_2(self):
        email = Message()
        email['To'] = 'alle:;'
        self.assertEquals(extract_targets_from_header(email),
            [])

class TestExtractContentTypeFromHeader(unittest.TestCase):
    def test(self):
        email = Message()
        email['Content-Type'] = 'foo/bar'
        self.assertEquals(extract_content_type_from_header(email),
            'foo/bar')

    def test_no(self):
        email = Message()
        self.assertEquals(extract_content_type_from_header(email),
            'text/plain')

    def test_params(self):
        email = Message()
        email['Content-Type'] = 'foo/bar;\n param1=foo boundary=foobar'
        self.assertEquals(extract_content_type_from_header(email),
            'foo/bar')


class TestGetDomainFromHeader(unittest.TestCase):
    def test_get_domain_from_header_mid1(self):
        email = MIMEText("")
        email.add_header("Message-ID", "<123456.7890@example.org>")
        self.assertEquals(get_domain_from_header(email),
                "example.org")

    def test_get_domain_from_header_mid2(self):
        email = MIMEText("")
        email.add_header("Message-ID", "123456.7890@example.org")
        self.assertEquals(get_domain_from_header(email),
                "example.org")

    def test_get_domain_from_header_mid3(self):
        email = MIMEText("")
        email.add_header("Message-ID", "<123456.7890@example.org")
        self.assertEquals(get_domain_from_header(email),
                "example.org")

    def test_get_domain_from_header_mid4(self):
        email = MIMEText("")
        email.add_header("Message-ID", "123456.7890@example.org>")
        self.assertEquals(get_domain_from_header(email),
                "example.org")

    def test_get_domain_from_header_mid5(self):
        email = MIMEText("")
        email.add_header("Message-Id",
            "<zarafa.12345678.9abc.defghZZZZ@example.ex.local>")
        email.add_header("From", "foo@example.org")
        self.assertEquals(get_domain_from_header(email),
                "example.ex.local")

    def test_get_domain_from_header_from1(self):
        email = MIMEText("")
        email.add_header("Message-ID", "123456.7890@")
        email.add_header("From", "foo@example.org")
        self.assertEquals(get_domain_from_header(email),
                "example.org")

    def test_get_domain_from_header_from2(self):
        email = MIMEText("")
        email.add_header("Message-ID", "123456.7890@localhost")
        email.add_header("From", "foo@example.org")
        self.assertEquals(get_domain_from_header(email),
                "example.org")

    def test_get_domain_from_header_from3(self):
        email = MIMEText("")
        email.add_header("From", "Someone <foo@example.org>")
        self.assertEquals(get_domain_from_header(email),
                "example.org")

    def test_get_domain_from_header_from4(self):
        email = MIMEText("")
        email.add_header("From", "foo@example.org (Someone)")
        self.assertEquals(get_domain_from_header(email),
                "example.org")

    def test_get_domain_from_header_lastresort(self):
        email = MIMEText("")
        self.assertEquals(get_domain_from_header(email),
                "localhost")

class TestEmbedder(unittest.TestCase):
    def test_insert_into(self):
        e = Embedder("()pattern()", ["txt"])
        self.assertEquals(
            e.insert_into("Textpatterntest",
                what="passes",
                type="txt",
                email=None), "Textpassestest")

    def test_multiple_pattern_occurences(self):
        e = Embedder("()pattern()", ["txt"])
        self.assertEquals(
            e.insert_into("Textpatterntestpattern",
                what="passes",
                type="txt",
                email=None), "Textpassestestpattern")

    def test_headers_condition(self):
        email = Message()
        email.add_header('X-Test', '0')
        email.set_payload('TextpatternText')
        e = Embedder("()pattern()", ["txt"],
            headers_condition='^X-Test: 1$')
        self.assertRaises(EmbedderFailure,
            e.insert_into, "TextpatternText",
                what="result",
                type="txt",
                email=email)

        email = Message()
        email.add_header('X-Test', '1')
        email.set_payload('TextpatternText')
        self.assertEquals(e.insert_into("TextpatternText",
            what="result",
            type="txt",
            email=email), "TextresultText")

    def test_body_condition(self):
        email = Message()
        email.set_payload('Gener |')
        e = Embedder("()\|()", ["txt"],
            body_condition='Generator')
        self.assertRaises(EmbedderFailure,
            e.insert_into, "Gener |",
                what="result",
                type="txt",
                email=email)

        email.set_payload('Generator |')
        self.assertEquals(e.insert_into("Generator |",
            what="result",
            type="txt",
            email=email), "Generator result")

    def test_anybody_condition(self):
        email = Message()
        email.set_type("multipart/mixed")
        a = Message()
        a.set_payload('Not a match')
        email.attach(a)
        e = Embedder("()\|()", ["txt"],
            anybody_condition='<meta [^>]*Outlook[^>]*>')
        self.assertRaises(EmbedderFailure,
            e.insert_into, "TextpatternText",
                what="result",
                type="txt",
                email=email)

        b = Message()
        b.set_payload('<html><head><meta ...Outlook ...>This will | match')
        email.attach(b)
        self.assertEquals(e.insert_into("|",
            what="result",
            type="txt",
            email=email), "result")

    def test_anybody_condition_non_multipart(self):
        email = Message()
        email.set_payload('<meta Outlook>Matching |')
        e = Embedder("()\|()", ["txt"],
            anybody_condition='<meta [^>]*Outlook[^>]*>')
        self.assertRaises(EmbedderFailure,
            e.insert_into, "<meta Outlook>Matching result",
                what="result",
                type="txt",
                email=email)

    def test_type_mismatch(self):
        e = Embedder("()pattern()", ["txt"])
        self.assertRaises(EmbedderFailure,
            e.insert_into, "Textpatterntest",
                what="fails",
                type="html",
                email=None)

    def test_pattern_multi_match(self):
        e = Embedder("(A)()|()(D)", ["txt"])
        self.assertEquals(e.insert_into(
            "AB", what="pass", type="txt", email=None),
            "ApassB")

        self.assertEquals(e.insert_into(
            "CD", what="pass", type="txt", email=None),
            "CpassD")

    def test_pattern_mismatch(self):
        e = Embedder("()pat!tern()", ["txt"])
        self.assertRaises(EmbedderFailure,
            e.insert_into, "Textpatterntest",
                what="fails",
                type="txt",
                email=None)

    def test_fromstring(self):
        e = Embedder.fromstring('txt|unknown   /()T E S T()/')
        self.assertEquals(e.types, ["txt", "unknown"])
        self.assertEquals(
            e.insert_into("Text T E S T Text",
                what="blubb",
                type="txt",
                email=None),
            "Text blubb Text")

    def test_fromstring_name(self):
        e = Embedder.fromstring('txt /()test()/', name='test line 8')
        self.assertEquals(e.name, 'test line 8')

    def test_fromstring_escape(self):
        e = Embedder.fromstring('txt|unknown   /()T E S T()\//')
        self.assertEquals(
            e.insert_into("Text T E S T/ Text",
                what="blubb",
                type="txt",
                email=None),
            "Text blubb Text")

    def test_fromstring_escape_keep_regexp_escapes(self):
        e = Embedder.fromstring('txt|unknown   /()\|()/')
        self.assertEquals(
            e.insert_into("Text | Text",
                what="blubb",
                type="txt",
                email=None),
            "Text blubb Text")

    def test_fromstring_multi_space(self):
        e = Embedder.fromstring('txt|unknown  /Text/body  /()\|()/')
        self.assertEquals(
            e.insert_into("Text | Text",
                what="blubb",
                type="txt",
                email=None),
            "Text blubb Text")

    def test_fromstring_anybody_condition(self):
        e = Embedder.fromstring(
            'txt|unknown   /mso:microsoft-/any-body     /()TEST()/')
        self.assertEquals(e.anybody_condition.match("test"), None)
        self.assertNotEquals(
            e.anybody_condition.search("xyzmso:microsoft-asd"),
            None)

    def test_fromstring_body_condition(self):
        e = Embedder.fromstring(
            'txt|unknown   /mso:microsoft-/body     /()TEST()/')
        self.assertEquals(e.body_condition.match("test"), None)
        self.assertNotEquals(
            e.body_condition.search("xyzmso:microsoft-asd"),
            None)

    def test_fromstring_headers_condition(self):
        e = Embedder.fromstring(
            'txt|unknown   /^X-Test: 1$/headers     /()TEST()/')
        self.assertEquals(e.headers_condition.match("test"), None)
        self.assertNotEquals(
            e.headers_condition.search("X-Test: 1"),
            None)


# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
