from __future__ import absolute_import
import os
import unittest
import shutil
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from tempfile import mkdtemp
from ....disclaimers.html import HTMLResourceInjector, URIHandler
from .email import normalize, dump_eml, fix_boundary_lf

class URIHandlerMock(object):
    def __init__(self, text):
        self.text = text

    def analyze(self):
        if 'Just some' in self.text:
            return False
        if 'strange' in self.text:
            self.paths = [
                ("images.cgi?show=strange", "strange")
            ]
        elif 'png' in self.text:
            self.paths = [
                ("images.cgi?show=test.gif", "test.gif"),
                ("images.cgi?show=test.GIF", "test.GIF"),
                ("images.cgi?show=test.png", "test.png"),
                ("images.cgi?show=test.PNG", "test.PNG"),
                ("images.cgi?show=test.jpg", "test.jpg"),
                ("images.cgi?show=test.JPG", "test.JPG"),
                ("images.cgi?show=test.jpeg", "test.jpeg"),
                ("images.cgi?show=test.JPEG", "test.JPEG"),
                ("images.cgi?show=test.invalid", "test.invalid"),
            ]
        else:
            self.paths = [
                ("images.cgi?show=test.gif", "test2.gif"),
            ]
        return True

    def translate(self, table):
        return self.text

    def get_paths(self):
        return self.paths


class TestHTMLResourceInjector(unittest.TestCase):
    def setUp(self):
        self.path = mkdtemp()
        self.oldcwd = os.getcwd()
        self.testpath = os.path.join(self.path, 'test2.gif')
        fd = open(self.testpath, "wb")
        fd.write("GIF89aas")
        fd.close()
        self.testpath2 = os.path.join(self.path, 'strange')
        fd = open(self.testpath2, "wb")
        fd.write("GIF89aas")
        fd.close()
        self.hri = HTMLResourceInjector()
        self.hri.set_uri_handler_class(URIHandlerMock)
        self.hri.set_resolver(lambda path: os.path.join(self.path, path))
        self.root = MIMEText('html')
        self.root.set_type("text/html")

    def tearDown(self):
        if self.path != '/' and '..' not in self.path:
            shutil.rmtree(self.path)
        os.chdir(self.oldcwd)

    def test_multipart(self):
        mime = MIMEMultipart('mixed')
        self.assertRaises(NotImplementedError, self.hri.inject_into,
                mime, 'example.org')

    def test_no_url(self):
        self.root.set_payload('<html><body>'
                '<p>Just some text/<p></body></html>')
        result = self.hri.inject_into(self.root, 'example.org')
        self.assertEquals(result.as_string(False),
            'Content-Transfer-Encoding: 7bit\n'
            'MIME-Version: 1.0\n'
            'Content-Type: text/html; charset="us-ascii"\n\n'
            '<html><body><p>Just some text/<p>'
            '</body></html>')

    def test_img_url(self):
        self.root.set_payload('<html><body>'
                '<img src="images.cgi?show=test.gif" />'
                '</body></html>')
        result = self.hri.inject_into(self.root, 'example.org')
        self.assertMultiLineEqual(normalize(result.as_string(False)),
            fix_boundary_lf(
            'Content-Type: multipart/related; boundary="BOUNDARY0"\n'
            'MIME-Version: 1.0\n\n'
            '--BOUNDARY0\n'
            'Content-Transfer-Encoding: 7bit\n'
            'MIME-Version: 1.0\n'
            'Content-Type: text/html; charset="us-ascii"\n\n'
            '<html><body><img src="images.cgi?show=test.gif" />'
            '</body></html>\n'
            '--BOUNDARY0\n'
            'Content-Type: image/gif\n'
            'MIME-Version: 1.0\n'
            'Content-Transfer-Encoding: base64\n'
            'Content-Disposition: inline\n'
            'Content-ID: <md.49cd60e1969a0a3855853e2bd5e71592a545518e@example.org>\n\n'
            'R0lGODlhYXM=\n'
            '--BOUNDARY0--'))

    def test_img_url_image_types(self):
        self.root.set_payload('<html><body>\n'
                '<img src="images.cgi?show=test.gif" />\n'
                '<img src="images.cgi?show=test.GIF" />\n'
                '<img src="images.cgi?show=test.png" />\n'
                '<img src="images.cgi?show=test.PNG" />\n'
                '<img src="images.cgi?show=test.jpg" />\n'
                '<img src="images.cgi?show=test.JPG" />\n'
                '<img src="images.cgi?show=test.jpeg" />\n'
                '<img src="images.cgi?show=test.JPEG" />\n'
                '<img src="images.cgi?show=test.invalid" />\n'
                '</body></html>')
        gif1 = 'GIF87a'
        gif2 = 'GIF89a'
        png = '\x89PNG\r\n\x1a\n'
        jpeg = '\xff\xd8\xff\xe0\x00\x10JFIF\0'
        invalid = "z"
        def make_img(ext, content):
            path = os.path.join(self.path, 'test.%s' % ext)
            fd = open(path, "wb")
            fd.write(content)
            fd.close()
        make_img("gif", gif1)
        make_img("GIF", gif2)
        make_img("png", png)
        make_img("PNG", png)
        make_img("jpg", jpeg)
        make_img("JPG", jpeg)
        make_img("jpeg", jpeg)
        make_img("JPEG", jpeg)
        make_img("invalid", invalid)

        result = self.hri.inject_into(self.root, 'example.org')
        self.assertMultiLineEqual(dump_eml(normalize(result.as_string(False)),
            "html.image-test"), fix_boundary_lf(
"""Content-Type: multipart/related; boundary="BOUNDARY0"
MIME-Version: 1.0

--BOUNDARY0
Content-Transfer-Encoding: 7bit
MIME-Version: 1.0
Content-Type: text/html; charset="us-ascii"

<html><body>
<img src="images.cgi?show=test.gif" />
<img src="images.cgi?show=test.GIF" />
<img src="images.cgi?show=test.png" />
<img src="images.cgi?show=test.PNG" />
<img src="images.cgi?show=test.jpg" />
<img src="images.cgi?show=test.JPG" />
<img src="images.cgi?show=test.jpeg" />
<img src="images.cgi?show=test.JPEG" />
<img src="images.cgi?show=test.invalid" />
</body></html>
--BOUNDARY0
Content-Type: image/gif
MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Disposition: inline
Content-ID: <md.e166b4a8d2182ba8711b37db8367c73e08226cdc@example.org>

R0lGODdh
--BOUNDARY0
Content-Type: image/gif
MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Disposition: inline
Content-ID: <md.25c9b37ae36a0a08318d4dca7ca57ea98d776821@example.org>

R0lGODlh
--BOUNDARY0
Content-Type: image/png
MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Disposition: inline
Content-ID: <md.4caece539b039b16e16206ea2478f8c5ffb2ca05@example.org>

iVBORw0KGgo=

--BOUNDARY0
Content-Type: image/jpeg
MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Disposition: inline
Content-ID: <md.d84fd0dec6cdbb3ec64be38d7f200f37487a7e0a@example.org>

/9j/4AAQSkZJRgA=
--BOUNDARY0
Content-Type: application/octet-stream
MIME-Version: 1.0
Content-Transfer-Encoding: base64
Content-Disposition: inline
Content-ID: <md.395df8f7c51f007019cb30201c49e884b46b92fa@example.org>

eg==
--BOUNDARY0--"""))

    def test_img_url_strange(self):
        self.root.set_payload('<html><body>'
                '<img src="images.cgi?show=strange" />'
                '</body></html>')
        result = self.hri.inject_into(self.root, 'example.org')
        self.assertMultiLineEqual(normalize(result.as_string(False)),
            fix_boundary_lf(
            'Content-Type: multipart/related; boundary="BOUNDARY0"\n'
            'MIME-Version: 1.0\n\n'
            '--BOUNDARY0\n'
            'Content-Transfer-Encoding: 7bit\n'
            'MIME-Version: 1.0\n'
            'Content-Type: text/html; charset="us-ascii"\n\n'
            '<html><body><img src="images.cgi?show=strange" />'
            '</body></html>\n'
            '--BOUNDARY0\n'
            'Content-Type: image/gif\n'
            'MIME-Version: 1.0\n'
            'Content-Transfer-Encoding: base64\n'
            'Content-Disposition: inline\n'
            'Content-ID: <md.49cd60e1969a0a3855853e2bd5e71592a545518e@example.org>\n\n'
            'R0lGODlhYXM=\n'
            '--BOUNDARY0--'))


class TestURIHandler(unittest.TestCase):
    def test_noop(self):
        uh = URIHandler("<html><body><p>Just some text</p></body></html>")
        self.assertEquals(uh.analyze(), False)

    def test_simple(self):
        uh = URIHandler('<html><body>'
                '<img src="/maildisclaimers/images.cgi?show=abc.gif" />'
                '</body></html>')
        self.assertEquals(uh.analyze(), True)
        self.assertEquals(uh.get_paths(),
            [('/maildisclaimers/images.cgi?show=abc.gif', 'abc.gif')])
        table = {'/maildisclaimers/images.cgi?show=abc.gif':
            'cid:123@example.org'}
        self.assertEquals(uh.translate(table), '<html><body>'
                '<img src="cid:123@example.org" />'
                '</body></html>')

    def test_reusing_url(self):
        uh = URIHandler('<html><body>'
                '<img src="/maildisclaimers/images.cgi?show=abc.gif" />'
                '<img src="/maildisclaimers/images.cgi?show=abc.gif" />'
                '</body></html>')
        self.assertEquals(uh.analyze(), True)
        self.assertEquals(uh.get_paths(),
            [('/maildisclaimers/images.cgi?show=abc.gif', 'abc.gif')])
        table = {'/maildisclaimers/images.cgi?show=abc.gif':
            'cid:123@example.org'}
        self.assertEquals(uh.translate(table), '<html><body>'
                '<img src="cid:123@example.org" />'
                '<img src="cid:123@example.org" />'
                '</body></html>')

    def test_htmlquoted(self):
        uh = URIHandler('<html><body>'
                '<img src="/maildisclaimers/images.cgi?'
                'show=a&quot;b.gif" />'
                '</body></html>')
        self.assertEquals(uh.analyze(), True)
        self.assertEquals(uh.get_paths(),
            [('/maildisclaimers/images.cgi?show=a&quot;b.gif',
                'a"b.gif')])
        table = {'/maildisclaimers/images.cgi?show=a&quot;b.gif':
            'cid:123@example.org'}
        self.assertEquals(uh.translate(table), '<html><body>'
                '<img src="cid:123@example.org" />'
                '</body></html>')

    def test_urlencoded(self):
        uh = URIHandler('<html><body>'
                '<img src="/maildisclaimers/images.cgi?'
                'show=a%26b.gif" />'
                '</body></html>')
        self.assertEquals(uh.analyze(), True)
        self.assertEquals(uh.get_paths(),
            [('/maildisclaimers/images.cgi?show=a%26b.gif',
                'a&b.gif')])
        table = {'/maildisclaimers/images.cgi?show=a%26b.gif':
            'cid:123@example.org'}
        self.assertEquals(uh.translate(table), '<html><body>'
                '<img src="cid:123@example.org" />'
                '</body></html>')

    def test_standalone(self):
        uh = URIHandler('<html><body>'
                '<img src="embed+cid:b.gif" />'
                '</body></html>')
        self.assertEquals(uh.analyze(), True)
        self.assertEquals(uh.get_paths(),
            [('embed+cid:b.gif',
                'b.gif')])
        table = {'embed+cid:b.gif':
            'cid:123@example.org'}
        self.assertEquals(uh.translate(table), '<html><body>'
                '<img src="cid:123@example.org" />'
                '</body></html>')

    def test_standalone_proper_quoting(self):
        uh = URIHandler('<html><body>'
                '<img src="embed+cid:b%26c.gif" />'
                '<img src="embed+cid:d&amp;e.gif" />'
                '</body></html>')
        self.assertEquals(uh.analyze(), True)
        self.assertEquals(uh.get_paths(),
            [('embed+cid:b%26c.gif', 'b%26c.gif'),
            ('embed+cid:d&amp;e.gif', 'd&e.gif')])
        table = {
            'embed+cid:b%26c.gif': 'cid:456@example.org',
            'embed+cid:d&amp;e.gif': 'cid:789@example.org'
        }
        self.assertEquals(uh.translate(table), '<html><body>'
                '<img src="cid:456@example.org" />'
                '<img src="cid:789@example.org" />'
                '</body></html>')

    def test_missing_translate(self):
        uh = URIHandler('<html><body>'
                '<img src="/maildisclaimers/images.cgi?'
                'show=abc.gif" />'
                '</body></html>')
        table = {}
        self.assertEquals(uh.translate(table), '<html><body>'
                '<img src="" />'
                '</body></html>')

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
