# -*- coding: utf-8 -*-
import os
import sys
import re
import time
import unittest
import uuid
import socket
import select
from shutil import rmtree
from subprocess import Popen, PIPE, STDOUT
from tempfile import mkdtemp, mkstemp

def find_free_port(base=10000, host='127.0.0.1'):
    for port in xrange(base, 65000):
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.bind((host, port))
            sock.close()
        except socket.error:
            continue
        return port

class BaseTestIntegration(unittest.TestCase):
    def setUp(self):
        self.path = mkdtemp()
        cfgpath = os.path.join(self.path, 'opt', 'bitbone', 'etc',
            'maildisclaimers')
        os.makedirs(cfgpath)
        os.makedirs(os.path.join(self.path, 'etc', 'postfix'))
        zpath = os.path.join(self.path, 'etc', 'zarafa')
        os.makedirs(zpath)
        self.path_ldap_cfg = os.path.join(zpath, 'ldap.bitbone.cfg')
        fd = open(self.path_ldap_cfg, 'wb')
        fd.write("""
ldap_protocol = ldap
ldap_host = localhost
ldap_port = 637
ldap_bind_user = uid=anon
ldap_bind_passwd = none
ldap_search_base = cn=dummy
ldap_emailaddress_attribute = mail
ldap_emailaliases_attribute = zarafaAlias
ldap_user_search_filter = (uid=*)
ldap_loginname_attribute = uid
ldap_fullname_attribute = cn
ldap_server_charset = utf-8
""")
        fd.close()

        fd = open(
            os.path.join(self.path, 'etc', 'postfix', 'main.cf'), 'wb')
        fd.write("""
virtual_mailbox_domains = local.example.org
""")
        fd.close()

    def tearDown(self):
        if self.path != '/' and '..' not in self.path:
            rmtree(self.path)


class TestIntegration(BaseTestIntegration):
    @classmethod
    def setUpClass(cls):
        f = open("systemid", "w")
        f.write(str(uuid.uuid4()))
        f.close()

    @classmethod
    def tearDownClass(cls):
        os.unlink("systemid")

    def test(self):
        # this is not race condition-safe, but better than hardcoding
        # ports
        gateway_port = find_free_port()
        filter_port = find_free_port(gateway_port + 1)
        gateway_args = [sys.executable, '-uc',
"""
import asyncore
from smtpd import SMTPServer

class TestGateway(SMTPServer):
    def __init__(self, *args, **kwargs):
        SMTPServer.__init__(self, *args, **kwargs)

    def process_message(self, *args):
        print repr(args[1:])
        return None

gw = TestGateway(('127.0.0.1', %d), (None, None))

asyncore.loop()
""" % gateway_port
        ]

        client_args = [sys.executable,
            "-uc",
            """
import socket
from smtplib import SMTP
from time import sleep

data = \"\"\"From: sender@local.example.org
To: target1@example.org, target2@example.org
Subject: Test
Content-Type: text/plain

Eine Test-Mail, die hoffentlich einen Disclaimer erhält.
\"\"\"
for x in xrange(5):
    try:
        smtp = SMTP('127.0.0.1', %d)
        smtp.sendmail('sender@local.example.org',
            ['target1@example.org', 'target2@example.org'], data)
        smtp.quit()
        raise SystemExit(0)
    except socket.error:
        sleep(0.3*x)
        continue

print "5 connect attempts failed"

""" % filter_port]

        filter_args = [sys.executable, '-uc',
"""
import sys
import os
sys.path.insert(0, "%s")
from bitbone import accounts
from bitbone import postfix
from bitbone.utils.license import License
TESTDIR = postfix.TESTDIR = accounts.TESTDIR = "%s"
from postfix_disclaimers_cf import Main

class LdapMock(object):
    def search_st(self, *args, **kwargs):
        return [
            ["uid=someone", {'uid': ['someone']}]
        ]

class StoreMock(object):
    def get(self, *args, **kwargs):
        return '\\n\\n-- \\nTest-Disclaimer'

    def disclaimers_disabled_for_targets(self, targets):
        return False

    def disclaimers_disabled_for_type(self, type):
        return False

    def disclaimers_disabled_for_sender(self, sender):
        return False

    def disclaimers_disabled_for_headers(self, headers):
        pass

    def get_image_path(self, name):
        return name

    def get_embedder_patterns(self):
        return []

tmpfile = os.path.join(TESTDIR, "empty")
fd = open(os.path.join(TESTDIR, "empty"), "w")
fd.write('''virtual_mailbox_domains = local.example.org
ldap_search_base = dc=local
ldap_emailaddress_attribute = mail
ldap_emailaliases_attribute = zarafaAlias
ldap_user_search_filter = (uid=*)
ldap_loginname_attribute = uid
ldap_fullname_attribute = cn
ldap_server_charset = utf-8
''')
fd.close()
m = Main()
m.make_optparser()
m.parse_arguments(['-b', '127.0.0.1:%d',
    '-g', '127.0.0.1:%d', '-d',
    '--ldap-config', tmpfile,
    '--local-domain-config', tmpfile,
])
m.make_content_filter()
m.cf.account_lookup._ldap = LdapMock()
m.cf.disclaimer_store = StoreMock()
m.cf.license = License(key_file="bitbone/tests/bitbone/utils/samples/valid.key", gpg_home="bitbone/tests/bitbone/utils/samples/gnupg")
print "about to run()"
m.run()
""" % (os.path.join(os.path.dirname(__file__), '..', '..', '..', '..'),
    self.path, filter_port, gateway_port)]

        gateway = Popen(gateway_args, stdout=PIPE, stderr=STDOUT)
        filter = Popen(filter_args, stdout=PIPE, stderr=STDOUT)
        # now we're simulating the main SMTP server connecting to the
        # filter
        client = Popen(client_args, stdout=PIPE, stderr=STDOUT)
        client.wait()

        # hack to make a race condition unlikelier...
        filter_output = ''
        timeout = 5
        endtime = time.time() + timeout
        while (endtime - time.time()) > 0:
            rs, ws, xs = select.select([filter.stdout], [], [], 1)
            if rs: # not timeout-related
                read = filter.stdout.read(1)
                filter_output += read
            else:
                read = True
            if not read or 'disconnecting\n' in filter_output:
                break
        filter.terminate()

        expected = """postfix_disclaimers_cf.make_content_filter INFO: using passwd account backend
about to run()
bitbone.disclaimers.smtp.make_listener INFO: Listening on ('127.0.0.1', 10001)
postfix_disclaimers_cf.loop INFO: Entering mainloop
bitbone.disclaimers.smtp.smtp_factory INFO: connecting to upstream gateway "127.0.0.1:10000"
bitbone.smtp.esmtpd.__init__ INFO: accepted new client, from=127.0.0.1
bitbone.disclaimers.smtp.rewrite_message INFO: handling mail from=sender@local.example.org
bitbone.accounts.passwd.get_objectinfo_by_email INFO: retrieving user data for email=sender@local.example.org from ldap
bitbone.accounts.base.get_ldap_attributes DEBUG: looking for attributes: ['uid', 'cn', 'givenName', 'sn', 'departmentNumber', 'department', 'company', 'telephoneNumber', 'facsimileTelephoneNumber', 'physicalDeliveryOfficeName', 'mail', 'st', 'co', 'postalCode', 'postOfficeBox', 'streetAddress', 'l', 'maildisclaimerTemplate']
bitbone.accounts.base.get_ldap_attributes DEBUG: using filter: (&(|(mail=sender@local.example.org)(zarafaAlias=sender@local.example.org))(uid=*))
bitbone.accounts.base.get_ldap_attributes DEBUG: result: [{'dn': 'uid=someone', 'departmentNumber': None, 'physicalDeliveryOfficeName': None, 'uid': u'someone', 'co': None, 'facsimileTelephoneNumber': None, 'postOfficeBox': None, 'company': None, 'l': None, 'st': None, 'maildisclaimerTemplate': None, 'postalCode': None, 'streetAddress': None, 'sn': None, 'mail': None, 'department': None, 'telephoneNumber': None, 'givenName': None, 'cn': None}]
bitbone.disclaimers.email._integrate_disclaimer_directly_into_part INFO: disclaimer successfully embedded (directly, default pattern txt|html ($))
bitbone.disclaimers.smtp.rewrite_message INFO: appended disclaimer to mail from=sender@local.example.org
bitbone.smtp.esmtpd.disconnect INFO: client (from=127.0.0.1) disconnecting
"""

        filter_output = re.sub("^\[.*?\] ", "", filter_output, flags=re.MULTILINE)
        filter_output = re.sub('gateway "127.0.0.1:\d+', 'gateway "127.0.0.1:10000', filter_output, flags=re.MULTILINE)
        filter_output = re.sub("Listening on \('127.0.0.1', \d+\)", "Listening on ('127.0.0.1', 10001)", filter_output, flags=re.MULTILINE)

        self.assertMultiLineEqual(filter_output, expected)

        gateway.terminate()

        self.assertEquals(client.stdout.read(), '')
        gateway_out = gateway.stdout.read()
        gateway_out = re.sub(
            r'\S{3}, \d{2} \S{3} \d{4,} \d{2}:\d{2}:\d{2} [\-\+]\d{4}',
            'VALID-DATE', gateway_out)
        gateway_out = re.sub(
            r'\(helo=\S+\)',
            '(helo=VALID-HELO)', gateway_out)
        gateway_out = re.sub(
            r'by \S+',
            'by VALID-BY', gateway_out)
        gateway_out = re.sub(
            r'\\n(?: |\\t)(\S)',
            r'\\n \1', gateway_out)
        gateway_out = re.sub(
            r'with ESMTP id\\n',
            r'with ESMTP id ', gateway_out)
        gateway_out = re.sub(
            r'PASSTHROUGH; \\n',
            r'PASSTHROUGH;\\n', gateway_out)
        self.assertEquals(gateway_out,
            '(\'sender@local.example.org\', '
            '[\'target1@example.org\', \'target2@example.org\'], '
            '\'Received: from [127.0.0.1] (helo=VALID-HELO)\\n'
            ' by VALID-BY (bitbone.disclaimers) with ESMTP id PASSTHROUGH;\\n'
            ' VALID-DATE\\n'
            'From: sender@local.example.org\\n'
            'To: target1@example.org, target2@example.org\\n'
            'Subject: Test\\n'
            'X-Maildisclaimers: processed\\n'
            'MIME-Version: 1.0\\n'
            'Content-Type: text/plain; charset="utf-8"\\n'
            'Content-Transfer-Encoding: 8bit\\n\\n'
            'Eine Test-Mail, die hoffentlich einen Disclaimer '
            'erh\\xc3\\xa4lt.\\n\\n'
            '-- \\nTest-Disclaimer\')\n')

class TestArgPassing(BaseTestIntegration):
    def __init__(self, *args, **kwargs):
        BaseTestIntegration.__init__(self, *args, **kwargs)
        fd, tmpfile = mkstemp()
        self.base = ["--gateway", "127.0.0.1:10025", "--bind",
            "127.0.0.1:10024",
            "--local-domain-config", tmpfile]

    def test_always_embed_html(self):
        from postfix_disclaimers_cf import Main
        m = Main()
        m.make_optparser()
        m.parse_arguments(self.base +
            ['--ldap-config', self.path_ldap_cfg])
        m.make_content_filter()
        self.assertEquals(m.cf.always_embed_html, False)
        m = Main()
        m.make_optparser()
        m.parse_arguments(self.base +
            ['--ldap-config', self.path_ldap_cfg] +
            ["--always-embed-html"])
        m.make_content_filter()
        self.assertEquals(m.cf.always_embed_html, True)

    def test_prefer_direct_embedding(self):
        from postfix_disclaimers_cf import Main
        m = Main()
        m.make_optparser()
        m.parse_arguments(self.base +
            ['--ldap-config', self.path_ldap_cfg])
        m.make_content_filter()
        self.assertEquals(m.cf.prefer_direct_embedding, False)
        m = Main()
        m.make_optparser()
        m.parse_arguments(self.base +
            ["--prefer-direct-embedding"] +
            ['--ldap-config', self.path_ldap_cfg])
        m.make_content_filter()
        self.assertEquals(m.cf.prefer_direct_embedding, True)

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
