import unittest
import uuid
import os
from functools import partial
from ...utils import replace_global_symbols
from ....utils.license import License

class ESMTPProxyMock(object):
    def __init__(self, *args, **kwargs):
        if not callable(kwargs['smtp_handler']):
            raise RuntimeError("smtp_handler not callable")

with replace_global_symbols({
    'bitbone.smtp.esmtpd_proxy.ESMTPProxy': ESMTPProxyMock},
    ['bitbone.disclaimers.smtp']):
    from ....disclaimers import smtp


class StoreMock(object):
    def __init__(self, data_dir):
        pass

    def get(self, type, msg, info, get_dynamic_attributes=False):
        if info.get('system:uid') == 2:
            return ''
        debug = ''
        if get_dynamic_attributes:
            debug = repr(get_dynamic_attributes("ldap:comment"))
        return '\n-- \nEin Disclaimer' + debug

    def disclaimers_disabled_for_targets(self, list):
        return list == ['no-disclaimers@example.org']

    def disclaimers_disabled_for_type(self, type):
        return type == 'no/disclaimers'

    def disclaimers_disabled_for_sender(self, sender):
        return sender == "no-disclaimer-source-address@example.org"

    def disclaimers_disabled_for_headers(self, headers):
        return 'X-Zarafa-Rule-Action: redirect' in headers

    def get_image_path(self, name):
        return 'NOTIMPLEMENTED'

    def get_embedder_patterns(self):
        return ["txt    /()\|()/"]

class DisclaimerInjectorMock(object):
    pass

class AccountLookupMock(object):
    def __init__(self, ldap_config=None):
        pass

    def load_config(self):
        pass

    def get_objectinfo_by_email(self, address):
        if 'nodisclaimer' in address:
            return {'system:uid': 2, 'system:gids': [2]}
        return {'system:uid': 1, 'system:gids': [2]}

    def get_user_attributes_by_email(self, address, attributes):
        return {'comment': 'xxx'}

ContentFilter = partial(smtp.ContentFilter,
    account_lookup=AccountLookupMock)


class PostfixConfigReaderMock(object):
    def read(self, path=None):
        pass

    def get_local_domains(object):
        return ['localhost', 'example.org']

    def is_local_domain(self, domain):
        return domain in ['localhost', 'example.org']

    def prefill_config(self):
        pass

class TestContentFilter(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        f = open("systemid", "w")
        f.write(str(uuid.uuid4()))
        f.close()

    @classmethod
    def tearDownClass(cls):
        os.unlink("systemid")

    def setUp(self):
        self.store_orig = smtp.Store
        smtp.Store = StoreMock
        self.pcr_orig = smtp.PostfixConfigReader
        smtp.PostfixConfigReader = PostfixConfigReaderMock

    def tearDown(self):
        smtp.Store = self.store_orig
        smtp.PostfixConfigReader = self.pcr_orig

    def test_process_local_nochange(self):
        cf = ContentFilter(None, None)
        cf.license = License(key_file="bitbone/tests/bitbone/utils/samples/valid.key", gpg_home="bitbone/tests/bitbone/utils/samples/gnupg")
        results = []
        def deliver(*args):
            results.append(args)

        cf.deliver = deliver

        # now simulate an incoming message;
        # normally this is handled by the real PureProxy class
        data = """From: Tester <test-nodisclaimer@example.org>
To: dontcare@example.org, Name ist wichtig <important@example.org>
Subject: Testmail
Date: Thu, 4 Oct 2012 08:57:47 +0000
Message-Id: <test1.506D4F8B.8010309@bitbone.de>
Content-Type: text/plain; charset=UTF-8
Content-Transfer-Encoding: 7bit

Testmail, die hoffentlich einen Disclaimer bekommt."""

        expected = """From: Tester <test-nodisclaimer@example.org>
To: dontcare@example.org, Name ist wichtig <important@example.org>
Subject: Testmail
Date: Thu, 4 Oct 2012 08:57:47 +0000
Message-Id: <test1.506D4F8B.8010309@bitbone.de>
Content-Type: text/plain; charset=UTF-8
Content-Transfer-Encoding: 7bit
X-Maildisclaimers: processed

Testmail, die hoffentlich einen Disclaimer bekommt."""
        result = cf.rewrite_message(data)
        self.assertEquals(result, expected)

    def test_process(self):
        cf = ContentFilter(None, None)
        cf.license = License(key_file="bitbone/tests/bitbone/utils/samples/valid.key", gpg_home="bitbone/tests/bitbone/utils/samples/gnupg")
        results = []
        def deliver(*args):
            results.append(args)

        cf.deliver = deliver

        # now simulate an incoming message;
        # normally this is handled by the real PureProxy class
        data = """From: Tester <test@example.org>
To: dontcare@example.org, Name ist wichtig <important@example.org>
Subject: Testmail
Date: Thu, 4 Oct 2012 08:57:47 +0000
Message-Id: <test1.506D4F8B.8010309@bitbone.de>
Content-Type: text/plain; charset=UTF-8
Content-Transfer-Encoding: 7bit

Testmail, die hoffentlich einen Disclaimer bekommt."""

        expected = """From: Tester <test@example.org>
To: dontcare@example.org, Name ist wichtig <important@example.org>
Subject: Testmail
Date: Thu, 4 Oct 2012 08:57:47 +0000
Message-Id: <test1.506D4F8B.8010309@bitbone.de>
Content-Type: text/plain; charset=UTF-8
X-Maildisclaimers: processed
Content-Transfer-Encoding: 7bit

Testmail, die hoffentlich einen Disclaimer bekommt.
-- 
Ein Disclaimer"""
        result = cf.rewrite_message(data)
        self.assertEquals(result, expected)

    def test_process_custom_attributes(self):
        cf = ContentFilter(None, None,
            enable_template_ldap=True)
        cf.license = License(key_file="bitbone/tests/bitbone/utils/samples/valid.key", gpg_home="bitbone/tests/bitbone/utils/samples/gnupg")
        results = []
        def deliver(*args):
            results.append(args)

        cf.deliver = deliver

        # now simulate an incoming message;
        # normally this is handled by the real PureProxy class
        data = """From: Tester <test@example.org>
To: dontcare@example.org, Name ist wichtig <important@example.org>
Subject: Testmail
Date: Thu, 4 Oct 2012 08:57:47 +0000
Message-Id: <test1.506D4F8B.8010309@bitbone.de>
Content-Type: text/plain; charset=UTF-8
Content-Transfer-Encoding: 7bit

Testmail, die hoffentlich einen Disclaimer bekommt."""

        expected = """From: Tester <test@example.org>
To: dontcare@example.org, Name ist wichtig <important@example.org>
Subject: Testmail
Date: Thu, 4 Oct 2012 08:57:47 +0000
Message-Id: <test1.506D4F8B.8010309@bitbone.de>
Content-Type: text/plain; charset=UTF-8
X-Maildisclaimers: processed
Content-Transfer-Encoding: 7bit

Testmail, die hoffentlich einen Disclaimer bekommt.
-- 
Ein Disclaimer{'ldap:comment': 'xxx'}"""
        result = cf.rewrite_message(data)
        self.assertEquals(result, expected)

    def test_process_no_disclaimer_type(self):
        cf = ContentFilter(None, None)
        results = []
        def deliver(*args):
            results.append(args)

        cf.deliver = deliver
        # now simulate an incoming message;
        data = """From: Tester <test@example.org>
To: foo@example.org
Subject: Testmail
Date: Thu, 4 Oct 2012 08:57:47 +0000
Content-Type: no/disclaimers; x=1
Message-Id: <test1.506D4F8B.8010309@bitbone.de>

Testmail, die hoffentlich keinen Disclaimer bekommt."""
        result = cf.rewrite_message(data)
        self.assertEquals(result, data)

    def test_process_no_disclaimer_target(self):
        cf = ContentFilter(None, None)
        results = []
        def deliver(*args):
            results.append(args)

        cf.deliver = deliver
        # now simulate an incoming message;
        data = """From: Tester <test@example.org>
To: no-disclaimers@example.org
Subject: Testmail
Date: Thu, 4 Oct 2012 08:57:47 +0000
Content-Type: text/plain; charset=UTF-8
Message-Id: <test1.506D4F8B.8010309@bitbone.de>

Testmail, die hoffentlich keinen Disclaimer bekommt."""
        result = cf.rewrite_message(data)
        self.assertEquals(result, data)

    def test_process_no_disclaimer_source(self):
        cf = ContentFilter(None, None)
        cf.license = License(key_file="bitbone/tests/bitbone/utils/samples/valid.key", gpg_home="bitbone/tests/bitbone/utils/samples/gnupg")
        results = []
        def deliver(*args):
            results.append(args)

        cf.deliver = deliver
        # now simulate an incoming message;
        data = """From: Tester <no-disclaimer-source-address@example.org>
To: someone@example.org
Subject: Testmail
Date: Thu, 4 Oct 2012 08:57:47 +0000
Content-Type: text/plain; charset=UTF-8
Message-Id: <test1.506D4F8B.8010309@bitbone.de>

Testmail, die hoffentlich keinen Disclaimer bekommt."""
        result = cf.rewrite_message(data)
        self.assertEquals(result, data)

    def test_process_no_disclaimer_targets(self):
        cf = ContentFilter(None, None)
        cf.license = License(key_file="bitbone/tests/bitbone/utils/samples/valid.key", gpg_home="bitbone/tests/bitbone/utils/samples/gnupg")
        results = []
        def deliver(*args):
            results.append(args)

        cf.deliver = deliver
        # now simulate an incoming message;
        data = """From: Tester <test@example.org>
To: no-disclaimers@example.org, someoneelse@example.org
Subject: Testmail
Date: Thu, 4 Oct 2012 08:57:47 +0000
Content-Type: text/plain; charset=UTF-8
Message-Id: <test1.506D4F8B.8010309@bitbone.de>
Content-Transfer-Encoding: 7bit

Testmail, die hoffentlich einen Disclaimer bekommt."""

        expected = """From: Tester <test@example.org>
To: no-disclaimers@example.org, someoneelse@example.org
Subject: Testmail
Date: Thu, 4 Oct 2012 08:57:47 +0000
Content-Type: text/plain; charset=UTF-8
Message-Id: <test1.506D4F8B.8010309@bitbone.de>
X-Maildisclaimers: processed
Content-Transfer-Encoding: 7bit

Testmail, die hoffentlich einen Disclaimer bekommt.
-- 
Ein Disclaimer"""
        result = cf.rewrite_message(data)
        self.assertEquals(result, expected)

    def test_process_no_disclaimer_target_cc(self):
        cf = ContentFilter(None, None)
        cf.license = License(key_file="bitbone/tests/bitbone/utils/samples/valid.key", gpg_home="bitbone/tests/bitbone/utils/samples/gnupg")
        results = []
        def deliver(*args):
            results.append(args)

        cf.deliver = deliver
        # now simulate an incoming message;
        data = """From: Tester <test@example.org>
To: no-disclaimers@example.org
Cc: someoneelse@example.org
Subject: Testmail
Date: Thu, 4 Oct 2012 08:57:47 +0000
Content-Type: text/plain; charset=UTF-8
Message-Id: <test1.506D4F8B.8010309@bitbone.de>
Content-Transfer-Encoding: 7bit

Testmail, die hoffentlich einen Disclaimer bekommt."""

        expected = """From: Tester <test@example.org>
To: no-disclaimers@example.org
Cc: someoneelse@example.org
Subject: Testmail
Date: Thu, 4 Oct 2012 08:57:47 +0000
Content-Type: text/plain; charset=UTF-8
Message-Id: <test1.506D4F8B.8010309@bitbone.de>
X-Maildisclaimers: processed
Content-Transfer-Encoding: 7bit

Testmail, die hoffentlich einen Disclaimer bekommt.
-- 
Ein Disclaimer"""
        result = cf.rewrite_message(data)
        self.assertEquals(result, expected)

    def test_process_special_embedders(self):
        cf = ContentFilter(None, None)
        cf.license = License(key_file="bitbone/tests/bitbone/utils/samples/valid.key", gpg_home="bitbone/tests/bitbone/utils/samples/gnupg")
        results = []
        def deliver(*args):
            results.append(args)

        cf.deliver = deliver
        # now simulate an incoming message;
        input = """From: Tester <test@example.org>
To: no-disclaimers@example.org
Cc: someoneelse@example.org
Subject: Testmail
Date: Thu, 4 Oct 2012 08:57:47 +0000
Content-Type: text/plain; charset=UTF-8
Message-Id: <test1.506D4F8B.8010309@bitbone.de>
Content-Transfer-Encoding: 7bit

Testmail, die hoffentlich einen Disclaimer bekommt.
|
Nach dem Disclaimer"""

        expected = """From: Tester <test@example.org>
To: no-disclaimers@example.org
Cc: someoneelse@example.org
Subject: Testmail
Date: Thu, 4 Oct 2012 08:57:47 +0000
Content-Type: text/plain; charset=UTF-8
Message-Id: <test1.506D4F8B.8010309@bitbone.de>
X-Maildisclaimers: processed
Content-Transfer-Encoding: 7bit

Testmail, die hoffentlich einen Disclaimer bekommt.

-- 
Ein Disclaimer
Nach dem Disclaimer"""
        result = cf.rewrite_message(input)
        self.assertEquals(result, expected)

    def test_process_special_embedders_fallback(self):
        cf = ContentFilter(None, None)
        cf.license = License(key_file="bitbone/tests/bitbone/utils/samples/valid.key", gpg_home="bitbone/tests/bitbone/utils/samples/gnupg")
        results = []
        def deliver(*args):
            results.append(args)

        cf.deliver = deliver
        # now simulate an incoming message;
        input = """From: Tester <test@example.org>
To: no-disclaimers@example.org
Cc: someoneelse@example.org
Subject: Testmail
Date: Thu, 4 Oct 2012 08:57:47 +0000
Content-Type: text/plain; charset=UTF-8
Message-Id: <test1.506D4F8B.8010309@bitbone.de>
Content-Transfer-Encoding: 7bit

Testmail, die hoffentlich einen Disclaimer bekommt.
|
Nach dem Disclaimer"""

        expected = """From: Tester <test@example.org>
To: no-disclaimers@example.org
Cc: someoneelse@example.org
Subject: Testmail
Date: Thu, 4 Oct 2012 08:57:47 +0000
Content-Type: text/plain; charset=UTF-8
Message-Id: <test1.506D4F8B.8010309@bitbone.de>
X-Maildisclaimers: processed
Content-Transfer-Encoding: 7bit

Testmail, die hoffentlich einen Disclaimer bekommt.

-- 
Ein Disclaimer
Nach dem Disclaimer"""
        result = cf.rewrite_message(input)
        self.assertEquals(result, expected)

    def test_process_no_disclaimer_header(self):
        cf = ContentFilter(None, None)
        results = []
        def deliver(*args):
            results.append(args)

        cf.deliver = deliver
        # now simulate an incoming message;
        data = """From: Tester <test@example.org>
To: no-disclaimers@example.org
Cc: someoneelse@example.org
Subject: Testmail
Date: Thu, 4 Oct 2012 08:57:47 +0000
Content-Type: text/plain; charset=UTF-8
Message-Id: <test1.506D4F8B.8010309@bitbone.de>
Content-Transfer-Encoding: 7bit
X-Zarafa-Rule-Action: redirect

Testmail, die hoffentlich keinen Disclaimer bekommt."""
        result = cf.rewrite_message(data)
        self.assertEquals(result, data)

    def test_process_bad_rcpt(self):
        cf = ContentFilter(None, None)
        results = []
        def deliver(*args):
            results.append(args)

        cf.deliver = deliver
        # now simulate an incoming message;
        # normally this is handled by the real PureProxy class
        data = """From: test@example.com
To: 
Subject: Testmail
Date: Thu, 4 Oct 2012 08:57:47 +0000
Content-Type: text/plain; charset=UTF-8
Message-Id: <test1.506D4F8B.8010309@bitbone.de>

Testmail, die hoffentlich keinen Disclaimer bekommt."""
        result = cf.rewrite_message(data)
        self.assertEquals(result, data)

    def test_process_bad_sender(self):
        cf = ContentFilter(None, None)
        results = []
        def deliver(*args):
            results.append(args)

        cf.deliver = deliver
        # now simulate an incoming message;
        # normally this is handled by the real PureProxy class
        data = """From:
To: dontcarex@example.org
Subject: Testmail
Date: Thu, 4 Oct 2012 08:57:47 +0000
Content-Type: text/plain; charset=UTF-8
Message-Id: <test1.506D4F8B.8010309@bitbone.de>

Testmail, die hoffentlich keinen Disclaimer bekommt."""
        result = cf.rewrite_message(data)
        self.assertEquals(result, data)

    def test_process_nonlocal(self):
        cf = ContentFilter(None, None)
        results = []
        def deliver(*args):
            results.append(args)

        cf.deliver = deliver
        # now simulate an incoming message;
        # normally this is handled by the real PureProxy class
        data = """From: Tester <test@example.com>
To: dontcare@example.org, Name ist wichtig <important@example.org>
Subject: Testmail
Date: Thu, 4 Oct 2012 08:57:47 +0000
Content-Type: text/plain; charset=UTF-8
Message-Id: <test1.506D4F8B.8010309@bitbone.de>

Testmail, die hoffentlich keinen Disclaimer bekommt."""
        result = cf.rewrite_message(data)
        self.assertEquals(result, data)

    def test_is_local_address(self):
        cf = ContentFilter(None, None)
        self.assertEquals(cf.is_local_address('someone@example.org'), True)
        self.assertEquals(cf.is_local_address('someone@examle.com'), False)
        self.assertEquals(cf.is_local_address('someone'), True)

    def test_is_local_address_case(self):
        cf = ContentFilter(None, None)
        self.assertEquals(cf.is_local_address('someone@EXAMPLE.org'), True)

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
