# -*- coding: utf-8 -*-
import os
import unittest
import shutil
import email
from tempfile import mkdtemp
from ....disclaimers import storage

class Base(unittest.TestCase):
    def setUp(self):
        self.tmpdir = mkdtemp()
        self.d_dir = os.path.join(self.tmpdir,
            'opt', 'bitbone', 'etc', 'maildisclaimers')
        os.makedirs(self.d_dir)
        self.d = storage.Store(data_dir=self.d_dir)
        self.test_msg = email.message_from_string("""From: test@example.org
To: receiver@example.com
Subject: Test

Test""")

    def tearDown(self):
        if self.tmpdir != '/' and '..' not in self.tmpdir:
            shutil.rmtree(self.tmpdir)


class TestStore(Base):
    def test_count(self):
        self.assertEquals(self.d.count(), 0)

    def test_count_fail(self):
        self.d.data_dir = "/dev/null/invalid"
        self.assertEquals(self.d.count(), 0)

    def test_get_none(self):
        self.assertEquals(
            self.d.get('txt', self.test_msg, {
                'system:uid': 1002,
                'system:gids': [2003, 2004]
            }),
            '')

    def test_get_uid_name(self):
        self.assertEquals(
            self.d.get('txt', self.test_msg, {
                'system:uid': u'max',
                'system:gids': [2003, 2004]
            }),
            '')

    def test_get_gid_name(self):
        self.assertEquals(
            self.d.get('txt', self.test_msg, {
                'system:uid': 1002,
                'system:gids': ['1-2-3-4', u'1-2-3-5']
            }),
            '')

    def test_get_invalid_type(self):
        self.assertEquals(
            self.d.get('x', 1002, [2003, 2004], {}),
            '')

    def test_get_global(self):
        f = open(os.path.join(self.d_dir, 'global.txt'), 'wb')
        f.write("Globaler Disclaimer")
        f.close()
        self.assertEquals(
            self.d.get('txt', self.test_msg, {
            'system:uid': 1002,
            'system:gids': [2003, 2004],
            }),
            '\nGlobaler Disclaimer')

    def test_get_global_ldap(self):
        f = open(os.path.join(self.d_dir, 'global.txt'), 'wb')
        f.write("Globaler Disclaimer, ${LDAP:givenName}")
        f.close()
        def get_dynamic_attributes(keys):
            self.assertEquals(keys, set([u"LDAP:givenName"]))
            return {'ldap:givenname': 'Max Mustermann'}
        self.assertEquals(
            self.d.get('txt', self.test_msg, {
                'system:uid': 1002,
                'system:gids': [2003, 2004],
            },
            get_dynamic_attributes=get_dynamic_attributes),
            '\nGlobaler Disclaimer, Max Mustermann')

    def test_get_group(self):
        f = open(os.path.join(self.d_dir, 'global.txt'), 'wb')
        f.write("Globaler Disclaimer")
        f.close()
        f = open(os.path.join(self.d_dir, 'group-2003.txt'), 'wb')
        f.write("Abteilung Einkauf")
        f.close()
        f = open(os.path.join(self.d_dir, 'group-2004.txt'), 'wb')
        f.write("Falsche Abteilung")
        f.close()
        self.assertEquals(
            self.d.get('txt', self.test_msg, {
                'system:uid': 1002,
                'system:gids': [2003, 2004],
            }),
            '\nAbteilung Einkauf')

    def test_get_user(self):
        f = open(os.path.join(self.d_dir, 'global.txt'), 'wb')
        f.write("Globaler Disclaimer")
        f.close()
        f = open(os.path.join(self.d_dir, 'group-2003.txt'), 'wb')
        f.write("Abteilung Einkauf")
        f.close()
        f = open(os.path.join(self.d_dir, 'group-2004.txt'), 'wb')
        f.write("Falsche Abteilung")
        f.close()
        f = open(os.path.join(self.d_dir, 'user-1001.txt'), 'wb')
        f.write("Falscher User")
        f.close()
        f = open(os.path.join(self.d_dir, 'user-1002.txt'), 'wb')
        f.write("Max Mustermann")
        f.close()
        self.assertEquals(
            self.d.get('txt', self.test_msg,
                {'system:uid': 1002, 'system:gids': [2003, 2004]}),
            '\nMax Mustermann')

    def test_get_umlaut_company(self):
        # Note: the bug which this test checks for is only
        # visible when no LC_* environment variables are set,
        # so use env - run-tests.py to verify
        f = open(os.path.join(self.d_dir, 'company-bitbone Würzburg.txt'), 'wb')
        f.write("Ok")
        f.close()
        self.assertEquals(
            self.d.get('txt', self.test_msg, {
                'system:uid': 1002,
                'system:gids': [2003],
                'company': u'bitbone Würzburg'
            }),
            '\nOk')

    def test_get_sender_domain(self):
        f = open(os.path.join(self.d_dir, 'sdomain-bitbone.de.txt'), 'wb')
        f.write("Bitbone-Disclaimer")
        f.close()
        self.assertEquals(
            self.d.get('txt', self.test_msg, {
                'system:uid': 1002,
                'system:gids': [2003, 2004],
                'from': 'test@bitbone.de'
            }),
            '\nBitbone-Disclaimer')

    def test_get_sender_domain_no_domain(self):
        f = open(os.path.join(self.d_dir, 'sdomain-localhost.txt'), 'wb')
        f.write("asd")
        f.close()
        self.assertEquals(
            self.d.get('txt', self.test_msg, {
                'system:uid': 1002,
                'system:gids': [2003, 2004],
                'from': 'test'
            }),
            '\nasd')

    def test_get_formatting_html(self):
        f = open(os.path.join(self.d_dir, 'global.html'), 'wb')
        f.write("-- <br>\r\n${FULLNAME}<br>\r\n${DEPARTMENT}<br>\r\n")
        f.write("${COMPANY}<br>\r\n")
        f.write("Telefon: ${PHONE} (Mo-Fr, 8-17 Uhr)<br>\r\n")
        f.write("Fax: ${FAX}/3<br>\r\n")
        f.write("Musterstr. 333<br>\r\n90909 Musterhausen")
        f.close()
        self.assertEquals(
            self.d.get('html', self.test_msg, {
                'system:uid': 1002,
                'system:gids': [2003, 2004],
                'fullname': u'Mäx <Möritz> Müstermänn',
                'phone': '+49 3333 3333-33',
                'fax': '+49 3333 3333-44',
                'department': 'Einkauf & Sonstiges',
                'company': 'bitbone'
            }),
            u'-- <br>\r\nM&#228;x &lt;M&#246;ritz&gt; '
            'M&#252;sterm&#228;nn<br>\r\nEinkauf &amp; Sonstiges'
            '<br>\r\n'
            'bitbone<br>\r\n'
            'Telefon: +49 3333 3333-33 (Mo-Fr, 8-17 Uhr)<br>\r\n'
            'Fax: +49 3333 3333-44/3<br>\r\nMusterstr. 333<br>\r\n'
            '90909 Musterhausen')

    def test_get_formatting_html_missing_info(self):
        f = open(os.path.join(self.d_dir, 'global.html'), 'wb')
        f.write("-- <br>\r\n${FULLNAME}<br />\r\n${DEPARTMENT}<br>\r\n")
        f.write("${FULLNAME}<br/>\r\n${DEPARTMENT}<br>\r\n")
        f.write("${COMPANY}<br>\r\n")
        f.write("Telefon: ${PHONE}<br>\r\nFax: ${FAX}<br>\r\n")
        f.write("Musterstr. 333<br>\r\n90909 Musterhausen")
        f.close()
        self.assertEquals(
            self.d.get('html', self.test_msg, {
                'system:uid': 1002,
                'system:gids': [2003, 2004],
                'fullname': u'',
                'phone': '',
                'fax': '',
                'department': '',
                'company': 'bitbone'
            }),
            u'-- <br>\r\nbitbone<br>\r\nTelefon: <br>\r\n'
            'Fax: <br>\r\nMusterstr. 333<br>\r\n'
            '90909 Musterhausen')

    def test_get_formatting_txt(self):
        f = open(os.path.join(self.d_dir, 'global.txt'), 'wb')
        f.write("-- \r\n${FULLNAME}\r\n${DEPARTMENT}\r\n${COMPANY}\r\n")
        f.write("Telefon: ${PHONE} (Mo - Fr, 8-17 Uhr)\r\n")
        f.write("Fax: ${FAX}/3\r\n")
        f.write("Musterstr. 333\r\n90909 Musterhausen")
        f.close()
        self.assertEquals(
            self.d.get('txt', self.test_msg, {
                'system:uid': 1002,
                'system:gids': [2003, 2004],
                'fullname': 'Max Moritz Mustermann',
                'phone': '+49 3333 3333-33',
                'fax': '+49 3333 3333-44',
                'department': 'Einkauf',
                'company': 'bitbone'
            }),
            u'\n-- \r\nMax Moritz Mustermann\r\nEinkauf\r\n'
            'bitbone\r\nTelefon: +49 3333 3333-33 (Mo - Fr, 8-17 Uhr)\r\n'
            'Fax: +49 3333 3333-44/3\r\nMusterstr. 333\r\n'
            '90909 Musterhausen')

    def test_get_formatting_txt_missing_info(self):
        """
        Verifies that trailing line breaks are properly stripped if
        a template keyword is empty
        """
        f = open(os.path.join(self.d_dir, 'global.txt'), 'wb')
        f.write("-- \r\n${FULLNAME}\r\n${DEPARTMENT}\r\n${COMPANY}\r\n")
        f.write("Telefon: \r\nFax: \r\n")
        f.write("Musterstr. 333\r\n90909 Musterhausen")
        f.close()
        self.assertEquals(
            self.d.get('txt', self.test_msg, {
                'system:uid': 1002,
                'system:gids': [2003, 2004],
                'company': 'bitbone'}),
            u'\n-- \r\nbitbone\r\n'
            'Telefon: \r\nFax: \r\n'
            'Musterstr. 333\r\n'
            '90909 Musterhausen')

    def test_get_formatting_txt_invalid_keyword(self):
        f = open(os.path.join(self.d_dir, 'global.txt'), 'wb')
        f.write("-- \r\n${INVALID}")
        f.close()
        self.assertEquals(
            self.d.get('txt', self.test_msg, {
                'system:uid': 1002,
                'system:gids': [2003, 2004]
            }),
            u'\n-- \r\n')

    def test_get_image_path(self):
        self.assertEquals(self.d.get_image_path(),
            os.path.join(self.d_dir, "images"))

    def test_get_image_path_file(self):
        self.assertEquals(self.d.get_image_path("test.gif"),
            os.path.join(self.d_dir, "images", "test.gif"))

    def test_get_image_path_file_evil(self):
        self.assertEquals(self.d.get_image_path("../test.gif"),
            os.path.join(self.d_dir, "images", "test.gif"))
        self.assertEquals(self.d.get_image_path("/test.gif"),
            os.path.join(self.d_dir, "images", "test.gif"))

class TestSpecialEmbedders(Base):
    def setUp(self):
        Base.setUp(self)
        f = open(os.path.join(self.d_dir, 'special-embedders.db'),
            'wb')
        f.write("txt /something/mail /somepattern/\r\n")
        f.write("html /somethingx/mail /somepatternx/\r\n")
        f.close()

    def test(self):
        self.assertEquals(self.d.get_embedder_patterns(),
            ["txt /something/mail /somepattern/",
            "html /somethingx/mail /somepatternx/"])

    def test_nofile(self):
        os.unlink(os.path.join(self.d_dir, 'special-embedders.db'))
        self.assertEquals(self.d.get_embedder_patterns(), [])

    def test_comment(self):
        f = open(os.path.join(self.d_dir, 'special-embedders.db'),
            'ab')
        f.write("#txt /something/mail /somepattern/\r\n")
        f.close()
        self.assertEquals(self.d.get_embedder_patterns(),
            ["txt /something/mail /somepattern/",
            "html /somethingx/mail /somepatternx/"])

    def test_empty_line(self):
        f = open(os.path.join(self.d_dir, 'special-embedders.db'),
            'ab')
        f.write("\r\n")
        f.close()
        self.assertEquals(self.d.get_embedder_patterns(),
            ["txt /something/mail /somepattern/",
            "html /somethingx/mail /somepatternx/"])



class TestIgnoreTarget(Base):
    def setUp(self):
        Base.setUp(self)
        f = open(os.path.join(self.d_dir, 'no-disclaimer-targets.db'),
            'wb')
        f.write("no-disclaimer-user@example.org\r\n")
        f.write("no-disclaimer-user-CI@example.org\r\n")
        f.write("*@no-disclaimer-domain.example.org\r\n")
        f.write("user?@*.user-multidomain-[abc].example.org\r\n")
        f.write("*.multidomain.example.org\n")
        f.write("/^someone@example\.(org|net|com)$/\n")
        f.write("/^ANOTHERONE@example\.(org|net|com)$/\n")
        f.write("/^someone@example\.(comx)\r\n")
        f.write("\r\n") #empty line for file structuring
        f.write("\n") #empty line for file structuring
        f.write("#commented@example.org\r\n")
        f.write("commented-but-working@example.org#test\r\n")
        f.write("commented-but-working2@example.org #test\r\n")
        f.close()

    def test_nofile(self):
        os.remove(os.path.join(self.d_dir,
            'no-disclaimer-targets.db'))
        self.assertEquals(self.d.disclaimers_disabled_for_target(
            'normal-user@example.org'), False)

    def test_unlisted(self):
        self.assertEquals(self.d.disclaimers_disabled_for_target(
            'normal-user@example.org'), False)

    def test_comment(self):
        self.assertEquals(self.d.disclaimers_disabled_for_target(
            'commented@example.org'), False)

        self.assertEquals(self.d.disclaimers_disabled_for_target(
            '#commented@example.org'), False)

    def test_comment_inline(self):
        self.assertEquals(self.d.disclaimers_disabled_for_target(
            'commented-but-working@example.org'), True)

        self.assertEquals(self.d.disclaimers_disabled_for_target(
            'commented-but-working2@example.org'), True)

    def test_plaintext(self):
        self.assertEquals(self.d.disclaimers_disabled_for_target(
            'no-disclaimer-user@example.org'), True)

    def test_plaintext_domainonly(self):
        self.assertEquals(self.d.disclaimers_disabled_for_target(
            'someone@no-disclaimer-domain.example.org'), True)

    def test_plaintext_case_insensitivity_value(self):
        self.assertEquals(self.d.disclaimers_disabled_for_target(
            'no-disclaimer-user@EXAMPLE.org'), True)

    def test_plaintext_case_insensitivity_pattern(self):
        self.assertEquals(self.d.disclaimers_disabled_for_target(
            'no-disclaimer-user-ci@EXAMPLE.ORG'), True)

    def test_glob(self):
        self.assertEquals(self.d.disclaimers_disabled_for_target(
            'user?@someone.user-multidomain-a.example.org'), True)

    def test_glob_domainonly(self):
        self.assertEquals(self.d.disclaimers_disabled_for_target(
            'someone@example.multidomain.example.org'), True)

    def test_regexp(self):
        self.assertEquals(self.d.disclaimers_disabled_for_target(
            'someone@example.com'), True)

    def test_regexp_case_insensitivity_pattern(self):
        self.assertEquals(self.d.disclaimers_disabled_for_target(
            'someone@example.com'), True)

    def test_regexp_missing_end(self):
        self.assertEquals(self.d.disclaimers_disabled_for_target(
            'someone@example.comx'), True)

    def test_multiple_mixed(self):
        self.assertEquals(self.d.disclaimers_disabled_for_targets(
            ['someone@example.org', 'someone@external.example.org']),
            False)

    def test_multiple_all_disabled(self):
        self.assertEquals(self.d.disclaimers_disabled_for_targets(
            ['someone@example.org', 'someone@example.com']),
            True)

    def test_multiple_all_enabled(self):
        self.assertEquals(self.d.disclaimers_disabled_for_targets(
            ['someone@external.example.org', 'some@external.example.com']),
            False)

    def test_empty_list(self):
        # see Store.disclaimers_disabled_for_targets for reasoning
        self.assertEquals(self.d.disclaimers_disabled_for_targets([]),
            False)


class TestIgnoreContentType(Base):
    def setUp(self):
        Base.setUp(self)
        f = open(os.path.join(self.d_dir, 'no-disclaimer-types.db'),
            'wb')
        f.write("multipart/encrypted\r\n")
        f.write("x-encrypted/*\r\n")
        f.write("/multipart\/(signed|encrypted)\r\n")
        f.write("/^slashtest/unescaped$\r\n")
        f.write("/^slashtest\\/escaped$\r\n")
        f.close()

    def test_pass(self):
        self.assertEquals(self.d.disclaimers_disabled_for_type(
            'multipart/mixed'), False)

    def test_plain(self):
        self.assertEquals(self.d.disclaimers_disabled_for_type(
            'multipart/encrypted'), True)

    def test_glob(self):
        self.assertEquals(self.d.disclaimers_disabled_for_type(
            'x-encrypted/something'), True)

    def test_regexp(self):
        self.assertEquals(self.d.disclaimers_disabled_for_type(
            'multipart/signed'), True)

    def test_regexp_slash(self):
        self.assertEquals(self.d.disclaimers_disabled_for_type(
            'slashtest/unescaped'), True)
        self.assertEquals(self.d.disclaimers_disabled_for_type(
            'slashtest/escaped'), True)

    def test_bad_regexp(self):
        f = open(os.path.join(self.d_dir, 'no-disclaimer-types.db'),
            'wb')
        f.write("/(\r\n") # fails, unbalanced parenthesis
        f.close()
        self.assertEquals(self.d.disclaimers_disabled_for_type(
            'doesntmatter'), False)


class TestIgnoreHeader(Base):
    def setUp(self):
        Base.setUp(self)
        f = open(os.path.join(self.d_dir, 'no-disclaimer-headers.db'),
            'wb')
        f.write("/^X-Zarafa-Rule-Action: (redirect|forward)$\r\n")
        f.write("X-No-Disclaimer: 1\r\n")
        f.close()

    def test_pass(self):
        self.assertEquals(self.d.disclaimers_disabled_for_headers(
            ['From: Foobar', 'To: Someone', 'Subject: Something']),
            False)

    def test_static(self):
        self.assertEquals(self.d.disclaimers_disabled_for_headers(
            ['From: Foobar', 'To: Someone', 'Subject: Something',
            'X-No-DisclaiMER: 0']),
            False)
        self.assertEquals(self.d.disclaimers_disabled_for_headers(
            ['From: Foobar', 'To: Someone', 'Subject: Something',
            'X-No-DisclaiMER: 1']),
            True)

    def test_regexp(self):
        self.assertEquals(self.d.disclaimers_disabled_for_headers(
            ['From: Foobar', 'To: Someone', 'Subject: Something',
            'X-Zarafa-Rule-Action: redirect']),
            True)
        self.assertEquals(self.d.disclaimers_disabled_for_headers(
            ['From: Foobar', 'Subject: Something',
            'X-Zarafa-Rule-Action: forward', 'To: Someone']),
            True)

        self.assertEquals(self.d.disclaimers_disabled_for_headers(
            ['From: Foobar', 'Subject: Something',
            'X-Zarafa-Rule-Action: unknown', 'To: Someone']),
            False)

class TestIgnoreSender(Base):
    def setUp(self):
        Base.setUp(self)
        f = open(os.path.join(self.d_dir, 'no-disclaimer-senders.db'),
            'wb')
        f.write("+\r\n") # test parsing
        f.write("+ \r\n") # test parsing
        f.write("nodisclaimer-source@example.org\r\n")
        f.write("+whitelisted@local\r\n")
        f.write("+ whitelisted2@local\r\n")
        f.write("+/^whitelistedre@local$/\r\n")
        f.write("+ /^whitelistedre2@local$/\r\n")
        f.write("*@local\r\n")
        f.close()

    def test_pass(self):
        self.assertEquals(self.d.disclaimers_disabled_for_sender(
            'someone@example.org'), False)

    def test_plain(self):
        self.assertEquals(self.d.disclaimers_disabled_for_sender(
            'nodisclaimer-source@example.org'), True)

    def test_glob(self):
        self.assertEquals(self.d.disclaimers_disabled_for_sender(
            'someone@local'), True)

    def test_whitelist(self):
        self.assertEquals(self.d.disclaimers_disabled_for_sender(
            'whitelisted@local'), False)

        self.assertEquals(self.d.disclaimers_disabled_for_sender(
            'nonwhitelisted@local'), True)

    def test_whitelist_space(self):
        self.assertEquals(self.d.disclaimers_disabled_for_sender(
            'whitelisted2@local'), False)

        self.assertEquals(self.d.disclaimers_disabled_for_sender(
            'nonwhitelisted@local'), True)

    def test_whitelist_regexp(self):
        self.assertEquals(self.d.disclaimers_disabled_for_sender(
            'whitelistedre@local'), False)
        self.assertEquals(self.d.disclaimers_disabled_for_sender(
            'nonwhitelisted@local'), True)

    def test_whitelist_regexp_with_space(self):
        self.assertEquals(self.d.disclaimers_disabled_for_sender(
            'whitelistedre2@local'), False)
        self.assertEquals(self.d.disclaimers_disabled_for_sender(
            'nonwhitelisted@local'), True)


class TestCustomRulesBackend(Base):
    def test(self):
        rules_path = os.path.join(self.d_dir, "rules.py")
        fh = open(rules_path, "w")
        fh.write("""
def find_disclaimer(type, mail, sender_info):
    return str([type, mail.get("From")])
""")
        fh.close()
        b = storage.CustomRulesBackend(path=rules_path)
        msg = email.message_from_string("""From: x@example.org
To: y@example.org
Subject: Test

xyz""")
        info = {}
        self.assertEquals(b.find_disclaimer("html", msg, info),
            "['html', 'x@example.org']")

    def test_func_validation(self):
        rules_path = os.path.join(self.d_dir, "rules.py")
        fh = open(rules_path, "w")
        fh.write("""
def x(type, mail, sender_info):
    return str([type, mail.get("From")])
""")
        fh.close()
        self.assertRaises(storage.CustomRulesError,
            storage.CustomRulesBackend, path=rules_path)

    def test_missing(self):
        rules_path = os.path.join(self.d_dir, "rules.py")
        self.assertRaises(storage.CustomRulesError,
            storage.CustomRulesBackend, path=rules_path)

    def test_failure(self):
        rules_path = os.path.join(self.d_dir, "rules.py")
        fh = open(rules_path, "w")
        fh.write("""
def find_disclaimer(type, mail, sender_info):
    0/0
""")
        fh.close()
        b = storage.CustomRulesBackend(path=rules_path)
        msg = email.message_from_string("""From: x@example.org
To: y@example.org
Subject: Test

xyz""")
        info = {}
        self.assertRaises(storage.CustomRulesError,
            b.find_disclaimer, "html", msg, info)




# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
