import unittest
import socket
from .common import AsynchatMock
from ...utils import replace_global_symbols
from ...mock import CallTracingMock

class MyAsynchatMock(AsynchatMock):
    def connect(self, *args, **kwargs):
        self._trace_call('connect', *args, **kwargs)

    def handle_connect(self, *args, **kwargs):
        self._trace_call('handle_connect', *args, **kwargs)

    def close(self, *args, **kwargs):
        self._trace_call('close', *args, **kwargs)


with replace_global_symbols({
    'asynchat.async_chat': MyAsynchatMock}, ['bitbone.smtp.client']):
    from ....smtp.client import (Client, UnexpectedServerReply,
        CommandNotSentError)

class User(CallTracingMock):
    pass

class TestClient(unittest.TestCase):
    def test(self):
        u = User()
        self.assertEquals(u.mock_get_next_call(), ('__init__',))
        c = Client()
        self.assertEquals(c.mock_get_next_call(), ('__init__',))
        c.connect('localhost', 25).callback(u.connected)
        self.assertEquals(c.mock_get_next_call(),
            ('set_terminator', '\r\n'))
        self.assertEquals(c.mock_get_next_call(), ('create_socket', 2, 1))
        self.assertEquals(c.mock_get_next_call(),
            ('connect', ('localhost', 25)))
        c.handle_connect()
        self.assertEquals(c.mock_get_next_call(),
            ('handle_connect',))
        c.collect_incoming_data('220 Hello')
        c.found_terminator()
        self.assertEquals(u.mock_get_next_call(), ('connected', '220',
            'Hello'))
        c.docmd('helo', 'x').callback(u.helo_callback)
        self.assertEquals(c.mock_get_next_call(),
            ('push', 'helo x\r\n'))
        c.collect_incoming_data('250 server')
        c.found_terminator()
        self.assertEquals(u.mock_get_next_call(), ('helo_callback',
            '250', 'server'))
        c.docmd('mail', 'from:a').callback(u.rcpt_callback)
        self.assertEquals(c.mock_get_next_call(),
            ('push', 'mail from:a\r\n'))
        c.collect_incoming_data('250 mail ok')
        c.found_terminator()
        self.assertEquals(u.mock_get_next_call(), ('rcpt_callback',
            '250', 'mail ok'))
        c.docmd('rcpt', 'to:x').callback(u.rcpt_callback)
        self.assertEquals(c.mock_get_next_call(),
            ('push', 'rcpt to:x\r\n'))
        c.collect_incoming_data('250 rcpt ok')
        c.found_terminator()
        self.assertEquals(u.mock_get_next_call(), ('rcpt_callback',
            '250', 'rcpt ok'))
        c.docmd('data').callback(u.data_callback)
        self.assertEquals(c.mock_get_next_call(),
            ('push', 'data\r\n'))
        c.collect_incoming_data('354 start sending')
        c.found_terminator()
        self.assertEquals(u.mock_get_next_call(), ('data_callback',
            '354', 'start sending'))
        c.push('Testmail\r\n.\r\n').callback(u.send_callback)
        self.assertEquals(c.mock_get_next_call(),
            ('push', 'Testmail\r\n.\r\n'))
        c.collect_incoming_data('250 ok, delivered')
        c.found_terminator()
        self.assertEquals(u.mock_get_next_call(), ('send_callback',
            '250', 'ok, delivered'))
        c.close()
        self.assertEquals(c.mock_get_next_call(),
            ('close', ))

    def test_esmtp_features(self):
        u = User()
        c = Client()
        c.connect('localhost', 25).callback(u.connected)
        c.handle_connect()
        c.collect_incoming_data('220 Hello')
        c.found_terminator()
        c.docmd('ehlo', 'x').callback(u.helo_callback)
        c.collect_incoming_data('250-server.hostname')
        c.found_terminator()
        c.collect_incoming_data('250-FEATURE1')
        c.found_terminator()
        c.collect_incoming_data('250-FEATURE2')
        c.found_terminator()
        c.collect_incoming_data('250 FEATURE3')
        c.found_terminator()
        self.assertEquals(u.mock_get_next_call(),
            ('__init__',))
        self.assertEquals(u.mock_get_next_call(),
            ('connected', '220', 'Hello'))
        self.assertEquals(u.mock_get_next_call(),
            ('helo_callback', '250',
                ['server.hostname', 'FEATURE1', 'FEATURE2', 'FEATURE3']))
        self.assertEquals(c.esmtp_features,
            set(['FEATURE1', 'FEATURE2', 'FEATURE3']))

    def test_chaining(self):
        u = User()
        c = Client()
        c.connect('localhost', 25).callback(u.connected)
        c.docmd('helo x')

    def test_unexpected_server_reply(self):
        c = Client()
        c.connect('localhost', 25)
        c.collect_incoming_data('100 One')
        c.found_terminator()
        c.collect_incoming_data('200 Two')
        self.assertRaises(UnexpectedServerReply,
            c.found_terminator)

    def test_ehlo_or_helo_esmtp(self):
        u = User()
        self.assertEquals(u.mock_get_next_call(),
            ('__init__',))
        c = Client()
        self.assertEquals(c.mock_get_next_call(),
            ('__init__',))
        c.connect('localhost', 25)
        self.assertEquals(c.mock_get_next_call(),
            ('set_terminator', '\r\n'))
        self.assertEquals(c.mock_get_next_call()[0],
            'create_socket')
        self.assertEquals(c.mock_get_next_call()[0],
            'connect')
        c.handle_connect()
        self.assertEquals(c.mock_get_next_call()[0],
            'handle_connect')
        c.collect_incoming_data('220 Hello')
        c.found_terminator()
        c.ehlo_or_helo_if_needed().callback(u.ehlo_helo_callback)
        self.assertEquals(c.mock_get_next_call(),
            ('push', 'ehlo %s\r\n' % socket.getfqdn()))
        c.collect_incoming_data('250 server.hostname')
        c.found_terminator()
        self.assertEquals(u.mock_get_next_call(),
            ('ehlo_helo_callback', '250', 'server.hostname'))
        self.assertEquals(c.esmtp, True)

    def test_ehlo_or_helo_duplicate(self):
        u = User()
        self.assertEquals(u.mock_get_next_call(),
            ('__init__',))
        c = Client()
        self.assertEquals(c.mock_get_next_call(),
            ('__init__',))
        c.connect('localhost', 25)
        self.assertEquals(c.mock_get_next_call(),
            ('set_terminator', '\r\n'))
        self.assertEquals(c.mock_get_next_call()[0],
            'create_socket')
        self.assertEquals(c.mock_get_next_call()[0],
            'connect')
        c.handle_connect()
        self.assertEquals(c.mock_get_next_call()[0],
            'handle_connect')
        c.collect_incoming_data('220 Hello')
        c.found_terminator()
        c.ehlo_or_helo_if_needed().callback(u.ehlo_helo_callback)
        self.assertEquals(c.mock_get_next_call(),
            ('push', 'ehlo %s\r\n' % socket.getfqdn()))
        c.collect_incoming_data('250 server.hostname')
        c.found_terminator()
        self.assertEquals(u.mock_get_next_call(),
            ('ehlo_helo_callback', '250', 'server.hostname'))
        c.ehlo_or_helo_if_needed()
        c.something_else()
        self.assertEquals(c.mock_get_next_call(),
            ('something_else',))

    def test_ehlo_or_helo_noesmtp(self):
        u = User()
        self.assertEquals(u.mock_get_next_call(),
            ('__init__',))
        c = Client()
        self.assertEquals(c.mock_get_next_call(),
            ('__init__',))
        c.connect('localhost', 25)
        self.assertEquals(c.mock_get_next_call(),
            ('set_terminator', '\r\n',))
        self.assertEquals(c.mock_get_next_call()[0],
            'create_socket')
        self.assertEquals(c.mock_get_next_call()[0],
            'connect')
        c.handle_connect()
        self.assertEquals(c.mock_get_next_call()[0],
            'handle_connect')
        c.collect_incoming_data('220 Hello')
        c.found_terminator()
        self.assertEquals(c.helo_ehlo_done, False)
        c.ehlo_or_helo_if_needed().callback(u.helo_done)
        self.assertEquals(c.mock_get_next_call(),
            ('push', 'ehlo %s\r\n' % socket.getfqdn()))
        c.collect_incoming_data('555 command not recognized')
        c.found_terminator()
        self.assertEquals(c.esmtp, False)
        self.assertEquals(c.mock_get_next_call(),
            ('push', 'helo %s\r\n' % socket.getfqdn()))
        self.assertEquals(c.helo_ehlo_done, False)
        c.collect_incoming_data('250 \o/')
        c.found_terminator()
        self.assertEquals(c.helo_ehlo_done, True)
        self.assertEquals(u.mock_get_next_call(),
            ('helo_done', '250', '\o/'))

    def test_pipelining(self):
        u = User()
        c = Client()
        c.connect('localhost', 25)
        c.handle_connect()
        c.collect_incoming_data('220 Hello')
        c.found_terminator()
        c.docmd('ehlo', 'x')
        c.collect_incoming_data('250-server.hostname')
        c.found_terminator()
        c.collect_incoming_data('250 PIPELINING')
        c.found_terminator()
        c.docmd('test1', 'x')
        c.docmd('test2', 'x')
        self.assertEquals(c.mock_get_next_call(),
            ('__init__',))
        self.assertEquals(c.mock_get_next_call(),
            ('set_terminator', '\r\n'))
        self.assertEquals(c.mock_get_next_call()[0],
            'create_socket')
        self.assertEquals(c.mock_get_next_call(),
            ('connect', ('localhost', 25)))
        self.assertEquals(c.mock_get_next_call(),
            ('handle_connect', ))
        self.assertEquals(c.mock_get_next_call(),
            ('push', 'ehlo x\r\n'))
        self.assertEquals(c.mock_get_next_call(),
            ('push', 'test1 x\r\n'))
        self.assertEquals(c.mock_get_next_call(),
            ('push', 'test2 x\r\n'))

    def test_no_pipelining(self):
        u = User()
        c = Client()
        c.connect('localhost', 25)
        c.handle_connect()
        c.collect_incoming_data('220 Hello')
        c.found_terminator()
        c.docmd('ehlo', 'x')
        c.docmd('test1', 'x')
        c.docmd('test2', 'x')
        self.assertEquals(c.mock_get_next_call(),
            ('__init__',))
        self.assertEquals(c.mock_get_next_call(),
            ('set_terminator', '\r\n'))
        self.assertEquals(c.mock_get_next_call()[0],
            'create_socket')
        self.assertEquals(c.mock_get_next_call(),
            ('connect', ('localhost', 25)))
        self.assertEquals(c.mock_get_next_call(),
            ('handle_connect', ))
        self.assertEquals(c.mock_get_next_call(),
            ('push', 'ehlo x\r\n'))
        c.test_foo()
        self.assertEquals(c.mock_get_next_call(),
            ('test_foo',))
        c.collect_incoming_data('250 server.hostname #ehlo')
        c.found_terminator()
        self.assertEquals(c.mock_get_next_call(),
            ('push', 'test1 x\r\n'))
        c.test_bar()
        self.assertEquals(c.mock_get_next_call(),
            ('test_bar',))
        c.collect_incoming_data('220 Ok #test1')
        c.found_terminator()
        self.assertEquals(c.mock_get_next_call(),
            ('push', 'test2 x\r\n'))
        c.collect_incoming_data('220 Ok #test2')
        c.found_terminator()
        c.test_end()
        self.assertEquals(c.mock_get_next_call(),
            ('test_end',))

    def test_want_result_for(self):
        u = User()
        c = Client()
        c.connect('localhost', 25)
        c.handle_connect()

        self.assertEquals(u.mock_get_next_call(), ('__init__',))

        # connect is a special case, as it's a network function and not
        # a protocol component
        c.want_reply_for(':CONNECT').callback(u.connect_reply)
        c.collect_incoming_data('220 Hello')
        c.found_terminator()
        self.assertEquals(u.mock_get_next_call(),
            ('connect_reply', '220', 'Hello'))

        # test normal case
        c.docmd('EHLO', 'x')
        c.want_reply_for('EHLO').callback(u.ehlo_reply)
        c.collect_incoming_data('250 example.org')
        c.found_terminator()
        self.assertEquals(u.mock_get_next_call(),
            ('ehlo_reply', '250', 'example.org'))

        # test case insensitivity
        c.docmd('ehlo', 'x')
        c.want_reply_for('EHLO').callback(u.ehlo_reply)
        c.collect_incoming_data('250 2.example.org')
        c.found_terminator()
        self.assertEquals(u.mock_get_next_call(),
            ('ehlo_reply', '250', '2.example.org'))

        # test invalid request
        self.assertRaises(CommandNotSentError,
            c.want_reply_for, 'NOT-YET-SENT-COMMAND')

        # test DOT, another special case, as it is no single-line command
        c.docmd('data')
        c.push('blabla\r\n.\r\n')
        c.want_reply_for('.').callback(u.dot_reply)
        c.collect_incoming_data('354 Give me data...')
        c.found_terminator()
        c.collect_incoming_data('250 Ok')
        c.found_terminator()
        self.assertEquals(u.mock_get_next_call(),
            ('dot_reply', '250', 'Ok'))

        # close is a special case, as it's a network function and not
        # a protocol component
        c.want_reply_for(':CLOSE').callback(u.close_callback)
        c.handle_close()
        self.assertEquals(u.mock_get_next_call(),
            ('close_callback',))

    def test_esmtp_feature_case_variants(self):
        u = User()
        c = Client()
        c.connect('localhost', 25).callback(u.connected)
        c.handle_connect()
        c.collect_incoming_data('220 Hello')
        c.found_terminator()
        c.docmd('ehlo', 'x')
        c.collect_incoming_data('250-server.hostname')
        c.found_terminator()
        c.collect_incoming_data('250-FEaTURE1')
        c.found_terminator()
        c.collect_incoming_data('250-feature2')
        c.found_terminator()
        c.collect_incoming_data('250 XFORWARD bla')
        c.found_terminator()
        self.assertEquals(c.esmtp_features,
            set(['FEaTURE1', 'feature2', 'XFORWARD bla']))
        self.failUnless('FEATURE1' in c.esmtp_features)
        self.failUnless('FEATURE2' in c.esmtp_features)
        self.failUnless('FEaTURE2' in c.esmtp_features)
        self.failUnless('XFORWARD' in c.esmtp_features)

    def test_esmtp_no_features(self):
        u = User()
        c = Client()
        c.connect('localhost', 25).callback(u.connected)
        c.handle_connect()
        c.collect_incoming_data('220 Hello')
        c.found_terminator()
        c.docmd('ehlo', 'x')
        c.collect_incoming_data('250 server.hostname')
        c.found_terminator()
        self.assertEquals(c.esmtp_features,
            set([]))

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
