import socket
import unittest
from ...mock import Mock, CallTracingMock
from ...utils import replace_global_symbols
from .common import ESMTPTest, AsyncoreDispatcherMock, AsynchatMock

class CallLaterMock(CallTracingMock):
    instances = [None]

    def __init__(self, callback, delay=0):
        CallTracingMock.__init__(self)
        self.callback = callback
        self.instances[0] = self

    def trigger(self):
        self.callback()


with replace_global_symbols({
    'asyncore.dispatcher': AsyncoreDispatcherMock,
    'asynchat.async_chat': AsynchatMock,
    'bitbone.timer.CallLater': CallLaterMock}, ['bitbone.smtp.esmtpd']):
    from ....smtp.esmtpd import ESMTPServer, ESMTPChannel

class HandlerMock(Mock):
    def process_message(self, sender, recipients, data):
        if '!RAISE' in data:
            raise Exception("Test error condition")
        self.sender = sender
        self.recipients = recipients
        self.data = data
        if '!CUSTOMCODE' in data:
            return ('251', 'Custom SMTP message')


class TestESMTPServer(unittest.TestCase):
    def test_init(self):
        e = ESMTPServer()

    def test_accept(self):
        def handler(sock, address, message_handler):
            handler_data.append((sock, address, message_handler))
            class x(object): pass
            return x()
        e = ESMTPServer(handler=handler)
        handler_data = []
        e.accept = lambda: ('socket', ('127.0.0.1', 3333))
        e.handle_accept()
        self.assertEquals(handler_data[0][0], 'socket')
        self.assertEquals(handler_data[0][1], ('127.0.0.1', 3333))
        self.failUnless(callable(handler_data[0][2]))

    def test_make_listener(self):
        e = ESMTPServer()
        e.make_listener(('127.0.0.1', 3333))
        self.assertEquals(e.mock_get_next_call(),
                ('__init__',))
        self.assertEquals(e.mock_get_next_call(),
                ('create_socket', socket.AF_INET, socket.SOCK_STREAM))
        self.assertEquals(e.mock_get_next_call(),
                ('set_reuse_addr',))
        self.assertEquals(e.mock_get_next_call(),
                ('bind', ('127.0.0.1', 3333)))
        self.assertEquals(e.mock_get_next_call(),
                ('listen', 5))

    def test_listen_backlog(self):
        e = ESMTPServer()
        e.make_listener(('127.0.0.1', 3333), 0)
        e.mock_skip_calls(4)
        self.assertEquals(e.mock_get_next_call(), ('listen', 0))

    def test_make_listener_fail(self):
        e = ESMTPServer()
        self.assertRaises(Exception, e.make_listener, ('127.0.0.1', 0))
        e.mock_skip_calls(4)
        self.assertEquals(e.mock_get_next_call(),
                ('close',))


class TestESMTPChannel(ESMTPTest):
    def setUp(self):
        self.handler = HandlerMock()
        self.e = ESMTPChannel(4, 5,
            message_handler=self.handler.process_message)

    def test_init(self):
        e = ESMTPChannel(1, 2, 3)
        self.assertEquals(e.remote_addr, 2)

    def test_parse_helo(self):
        self.assertEquals(self.e.mock_get_next_call(), ('__init__', 4))
        self.assertEquals(self.e.mock_get_next_call(),
            ('set_terminator', '\r\n'))
        self.e.mock_skip_calls(1)
        self.assertEquals(self.e.client_greeting, None)
        self.e.collect_incoming_data('HELO foobar')
        self.e.found_terminator()
        call = self.e.mock_get_next_call()
        self.assertEquals(call[0], 'push')
        self.assertRegexpMatches(call[1], r'^250 (\S+)\r\n$')
        self.assertEquals(self.e.client_greeting, 'foobar')

    def test_parse_empty_data(self):
        self.consume()
        self.e.found_terminator()
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '500 5.5.2 Error: bad syntax\r\n'))
        self.assertRaises(IndexError, self.e.mock_get_next_call)

    def test_parse_buffer(self):
        self.consume()
        self.assertEquals(self.e.esmtp, None)
        self.assertEquals(self.sendc('HELO x')[1][:4], '250 ')
        self.assertEquals(self.e.esmtp, False)
        self.assertEquals(self.sendc('TEST')[1][:4], '502 ')

    def test_helo_no_arg(self):
        self.consume()
        self.assertEquals(self.e.client_greeting, None)
        self.assertEquals(self.sendc('HELO'),
            ('push', '501 Syntax: HELO hostname\r\n'))

    def test_ehlo_no_arg(self):
        self.consume()
        self.assertEquals(self.e.client_greeting, None)
        self.assertEquals(self.sendc('EHLO'),
            ('push', '501 Syntax: EHLO hostname\r\n'))
        self.assertEquals(self.e.client_greeting, None)

    def test_ehlo(self):
        self.consume()
        self.assertEquals(self.e.client_greeting, None)
        self.assertEquals(self.e.esmtp, None)
        res = self.sendc('EHLO foobar', consume=4)
        self.assertEquals(res[0][1][:4], '250-')
        self.assertEquals(res[1:4], [
            ('push', '250-8BITMIME\r\n'),
            ('push', '250-ENHANCEDSTATUSCODES\r\n'),
            ('push', '250 PIPELINING\r\n'),
        ])
        self.assertEquals(self.e.client_greeting, "foobar")
        self.assertEquals(self.e.esmtp, True)

    def test_banner(self):
        self.e.mock_skip_calls(2)
        call = self.e.mock_get_next_call()
        self.assertEquals(call[0], 'push')
        self.assertRegexpMatches(call[1],
            r'^220 (\S+) ESMTP Python/bitbone\.esmtpd\r\n$')

    def test_unknown_command(self):
        self.consume()
        self.assertEquals(self.sendc('unknown-command foobar'),
            ('push', '502 5.5.2 Error: command "unknown-command" '
                'not recognized\r\n'))

    def test_unknown_command_with_catchall_method(self):
        result = []
        def catchall(*args):
            result.append(args)
            return '250', 'Ok'
        self.e.smtp_ = catchall
        self.consume()
        self.assertEquals(self.sendc('unknown- command foobar'),
            ('push', '250 Ok\r\n'))

    def test_mail_from(self):
        self.consume()
        self.sendc("helo x")
        self.assertEquals(self.e.mail_from, None)
        self.assertEquals(self.sendc('mail from:someone@example.org'),
            ('push', '250 2.1.0 Ok\r\n'))
        self.assertEquals(self.e.mail_from, 'someone@example.org')

    def test_mail_from_params(self):
        self.consume()
        self.sendc("helo x")
        self.assertEquals(self.e.mail_params, [])
        self.sendc('mail from:someone@example.org BODY=7bit')
        self.assertEquals(self.e.mail_from, 'someone@example.org')
        self.assertEquals(self.e.mail_params, [('BODY=7bit')])

    def test_mail_from_duplicate(self):
        self.consume()
        self.sendc('helo x')
        self.sendc('mail from:bla@example.org')
        self.assertEquals(self.e.mail_from, 'bla@example.org')
        self.assertEquals(self.sendc('mail from:xyz'),
            ('push', '203 5.5.1 Error: nested MAIL command\r\n'))
        self.assertEquals(self.e.mail_from, 'bla@example.org')

    def test_mail_from_case_variations(self):
        self.consume()
        self.sendc("helo x")
        self.sendc('MAIL FROM:bla@example.org')
        self.assertEquals(self.e.mail_from, "bla@example.org")
        self.sendc('rset')
        self.sendc('mail FROM:bla@example.org')
        self.assertEquals(self.e.mail_from, "bla@example.org")
        self.sendc('rset')
        self.sendc('MAIL from:bla@example.org')
        self.assertEquals(self.e.mail_from, "bla@example.org")
        self.sendc('rset')
        self.sendc('MaIL FRoM:bla@example.org')
        self.assertEquals(self.e.mail_from, "bla@example.org")
        self.sendc('rset')

    def test_mail_from_case_preserve(self):
        self.consume()
        self.sendc('helo x')
        self.sendc('mail from:Mustermann.Max@example.org')
        self.assertEquals(self.e.mail_from, "Mustermann.Max@example.org")

    def test_mail_from_whitespace(self):
        self.consume()
        self.sendc('helo x')
        self.sendc('mail from:   whitespace@example.org  ')
        self.assertEquals(self.e.mail_from, 'whitespace@example.org')

    def test_mail_from_angle_brackets(self):
        self.consume()
        self.sendc('helo x')
        self.sendc('mail from:  <abcde@example.org>  ')
        self.assertEquals(self.e.mail_from, 'abcde@example.org')

    def test_mail_from_no_helo(self):
        self.consume()
        self.assertEquals(self.sendc('mail from: abc'),
            ('push', '503 5.5.1 Error: send HELO/EHLO first\r\n'))

    def test_mail_no_from(self):
        self.consume()
        self.sendc('helo x')
        self.assertEquals(self.sendc('mail xyz:foo@example.org'),
            ('push', '501 5.5.4 Syntax: MAIL FROM:<address>\r\n'))
        self.assertEquals(self.e.mail_from, None)

    def test_rset(self):
        self.consume()
        self.sendc('helo x')
        self.sendc('mail from: abc')
        self.sendc('rcpt to: def')
        self.assertEquals(self.e.mail_from, 'abc')
        self.assertEquals(self.e.recipients, ['def'])
        self.assertEquals(self.sendc('rset x'),
            ('push', '501 5.5.4 Syntax: RSET\r\n'))
        self.assertEquals(self.e.mail_from, 'abc')
        self.assertEquals(self.sendc('rset', consume=2), [
            ('set_terminator', '\r\n'),
            ('push', '250 2.0.0 Ok\r\n')])
        self.assertEquals(self.e.mail_from, None)
        self.assertEquals(self.e.mail_params, [])
        self.assertEquals(self.e.recipients, [])

    def test_quit(self):
        self.consume()
        self.assertEquals(self.sendc('quit', consume=2), [
            ('push', '221 2.0.0 Bye\r\n'),
            ('close_when_done',)])

    def test_rcpt_to(self):
        self.consume()
        self.sendc('helo x')
        self.sendc('mail from: asd')
        self.assertEquals(self.e.recipients, [])
        self.assertEquals(self.sendc('rcpt to: asd@example.org'),
            ('push', '250 2.1.5 Ok\r\n'))
        self.assertEquals(self.sendc('rcpt to: xyz@example.org'),
            ('push', '250 2.1.5 Ok\r\n'))
        self.assertEquals(self.e.recipients,
            ['asd@example.org', 'xyz@example.org'])

    def test_rcpt_no_to(self):
        self.consume()
        self.sendc('helo x')
        self.sendc('mail from: asd')
        self.assertEquals(self.e.recipients, [])
        self.assertEquals(self.sendc('rcpt xx: asd@example.org'),
            ('push', '501 5.5.4 Syntax: RCPT TO:<address>\r\n'))
        self.assertEquals(self.e.recipients,
            [])

    def test_rcpt_to_no_mail(self):
        self.consume()
        self.assertEquals(self.sendc('rcpt to: asd'),
            ('push', '503 5.5.1 Error: need MAIL command\r\n'))
        self.sendc('helo x')
        self.assertEquals(self.sendc('rcpt to: asd'),
            ('push', '503 5.5.1 Error: need MAIL command\r\n'))

    def test_data_no_rcpt_to(self):
        self.consume()
        self.assertEquals(self.sendc('data'),
            ('push', '503 5.5.1 Error: need RCPT command\r\n'))

    def test_data_arg(self):
        self.consume()
        self.sendc('helo x')
        self.sendc('mail from: asd')
        self.sendc('rcpt to: asd')
        self.assertEquals(self.sendc('data x'),
            ('push', '501 5.5.4 Syntax: DATA\r\n'))

    def test_data_command(self):
        self.consume()
        self.sendc('helo x')
        self.sendc('mail from: asd')
        self.sendc('rcpt to: asd')
        self.assertEquals(self.sendc('data', consume=2), [
            ('set_terminator', '\r\n.\r\n'),
            ('push', '354 End data with <CR><LF>.<CR><LF>\r\n')])

    def test_data_decoding(self):
        self.consume()
        self.sendc('helo x')
        self.sendc('mail from: asd')
        self.sendc('rcpt to: foobar')
        self.sendc('rcpt to: x@example.org')
        self.sendc('data', consume=2)
        self.sendc(
            'From: asd\r\nTo: foobar\r\nSubject: Test\r\n\r\n'
            'Content\r\n..\r\nblabla\r\n...')
        self.failUnless(self.handler.data.endswith(
            'From: asd\r\nTo: foobar\r\nSubject: Test\r\n\r\n'
            'Content\r\n.\r\nblabla\r\n..'))

    def test_data_delivery(self):
        self.consume()
        self.sendc('helo x')
        self.sendc('mail from: asd')
        self.sendc('rcpt to: foobar')
        self.sendc('rcpt to: x@example.org')
        self.sendc('data', consume=2)
        self.assertEquals(self.sendc(
            'From: asd\r\nTo: foobar\r\nSubject: Test\r\n\r\n'
            'Content..blabla', consume=2), [
                ('set_terminator', '\r\n'),
                ('push', '250 2.0.0 Ok\r\n')])
        self.assertEquals(self.handler.sender, 'asd')
        self.assertEquals(self.handler.recipients,
                ['foobar', 'x@example.org'])
        self.failUnless(self.handler.data.startswith('Received: from'))
        self.failUnless(self.handler.data.endswith(
            'From: asd\r\nTo: foobar\r\nSubject: Test\r\n\r\n'
            'Content..blabla'))

    def test_delivery_exception(self):
        self.consume()
        self.sendc('helo x')
        self.sendc('mail from: asd')
        self.sendc('rcpt to: foobar')
        self.sendc('rcpt to: x@example.org')
        self.sendc('data', consume=2)
        self.assertEquals(self.sendc(
            'From: asd\r\nTo: foobar\r\nSubject: Test\r\n\r\n'
            'Content..blabla !RAISE'),
            ('push', '451 4.0.0 Unknown message handler error\r\n'))

    def test_delivery_custom_code(self):
        self.consume()
        self.sendc('helo x')
        self.sendc('mail from: asd')
        self.sendc('rcpt to: foobar')
        self.sendc('rcpt to: x@example.org')
        self.sendc('data', consume=2)
        self.assertEquals(self.sendc(
            'From: asd\r\nTo: foobar\r\nSubject: Test\r\n\r\n'
            'Content..blabla !CUSTOMCODE', consume=2), [
                ('set_terminator', '\r\n'),
                ('push', '251 Custom SMTP message\r\n')])

    def test_state_after_data_with_empty_from(self):
        self.consume()
        self.sendc('helo localhost')
        self.sendc('mail from:<>')
        self.sendc('rcpt to:foobar')
        self.sendc('data', consume=2)
        self.assertEquals(self.sendc(
            'From: asd\r\nTo: foobar\r\nSubject: Test\r\n\r\n'
            'Content..blabla', consume=2), [
            ('set_terminator', '\r\n'),
            ('push', '250 2.0.0 Ok\r\n')])
        self.assertEquals(self.handler.sender, '')
        self.assertEquals(self.handler.recipients,
                ['foobar'])
        self.failUnless(self.handler.data.endswith(
            'From: asd\r\nTo: foobar\r\nSubject: Test\r\n\r\n'
            'Content..blabla'))

        self.assertEquals(self.sendc('noop x'),
            ('push', '250 2.0.0 Ok\r\n'))

    def test_state_after_data(self):
        self.consume()
        self.sendc('helo x')
        self.sendc('mail from: asd')
        self.sendc('rcpt to: foobar')
        self.sendc('rcpt to: x@example.org')
        self.sendc('data', consume=2)
        self.assertEquals(self.sendc(
            'From: asd\r\nTo: foobar\r\nSubject: Test\r\n\r\n'
            'Content..blabla', consume=2), [
            ('set_terminator', '\r\n'),
            ('push', '250 2.0.0 Ok\r\n')])
        self.assertEquals(self.handler.sender, 'asd')
        self.assertEquals(self.handler.recipients,
                ['foobar', 'x@example.org'])
        self.failUnless(self.handler.data.endswith(
            'From: asd\r\nTo: foobar\r\nSubject: Test\r\n\r\n'
            'Content..blabla'))

        # second mail
        self.sendc('mail from: asd2')
        self.sendc('rcpt to: foobar2')
        self.sendc('data', consume=2)
        self.assertEquals(self.sendc(
            'From: asd2\r\nTo: foobar2\r\nSubject: Test2\r\n\r\n'
            'Content..blabla2', consume=2), [
                ('set_terminator', '\r\n'),
                ('push', '250 2.0.0 Ok\r\n')])
        self.assertEquals(self.handler.sender, 'asd2')
        self.assertEquals(self.handler.recipients,
                ['foobar2'])
        self.failUnless(self.handler.data.endswith(
            'From: asd2\r\nTo: foobar2\r\nSubject: Test2\r\n\r\n'
            'Content..blabla2'))

    def test_noop(self):
        self.consume()
        self.assertEquals(self.sendc('noop x'),
            ('push', '250 2.0.0 Ok\r\n'))
        self.assertEquals(self.sendc('noop '),
            ('push', '250 2.0.0 Ok\r\n'))
        self.assertEquals(self.sendc('noop'),
            ('push', '250 2.0.0 Ok\r\n'))

    def test_push_multiline(self):
        self.consume()
        self.e.push('100', 'test\ntest2')
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '100-test\r\n'))
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '100 test2\r\n'))

    def test_prepend_received_from_line(self):
        self.e.client_greeting = 'clienthelo'
        self.e.remote_addr = ('127.0.0.1', 0)
        self.e.server_software_name = 'bitbone.esmtpd'
        self.e.fqdn = 'myfqdn'
        self.e.esmtp = True
        self.assertRegexpMatches(self.e.prepend_received_from_line(''),
            r'^Received: from \[127\.0\.0\.1\] \(helo=clienthelo\)\r\n\t'
            r'by myfqdn \(bitbone\.esmtpd\) with ESMTP id '
            r'PASSTHROUGH;\r\n\t'
            r'\S{3}, \d{2} \S{3} \d{4,} \d{2}:\d{2}:\d{2} [\-\+]\d{4}\r\n'
        )

    def test_handle_close(self):
        self.consume()
        self.e.handle_close()
        self.assertEquals(self.e.mock_get_next_call(),
            ('handle_close',))

    def test_handle_closed_twice(self):
        self.consume()
        self.e.handle_close()
        self.assertEquals(self.e.mock_get_next_call(),
            ('handle_close',))
        self.e.handle_close()

    def test_disconnect_callback_cleanup(self):
        ctm = CallLaterMock.instances[0]
        self.assertEquals(ctm.mock_get_next_call(), ('__init__',))
        self.e.handle_close()
        self.assertEquals(ctm.mock_get_next_call(), ('destroy',))

    def test_client_timeout(self):
        self.consume()
        ctm = CallLaterMock.instances[0]
        self.assertEquals(ctm.mock_get_next_call(), ('__init__',))
        self.e.collect_incoming_data('helo x')
        self.assertEquals(ctm.mock_get_next_call(), ('reset',))
        ctm.trigger()
        res = self.e.mock_get_next_call()
        self.assertEquals(res[0], 'push')
        self.failUnless(res[1].startswith('421 4.4.2 '))
        self.failUnless(res[1].endswith(' Error: timeout exceeded\r\n'))
        self.assertEquals(self.e.mock_get_next_call(),
            ('handle_close',))

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
