import unittest
from functools import partial
from ...mock import CallTracingMock
from ...utils import replace_global_symbols
from .common import (ESMTPTest, AsynchatMock,
    AsyncoreDispatcherMock, asynchat_instances)
from ....defer import Deferred
from ....smtp.client import CommandNotSentError

with replace_global_symbols({
    'asyncore.dispatcher': AsyncoreDispatcherMock,
    'asynchat.async_chat': AsynchatMock},
    ['bitbone.smtp.esmtpd', 'bitbone.smtp.esmtpd_proxy']):
    from ....smtp.esmtpd_proxy import ESMTPProxy, ESMTPProxyChannel

class SMTPMock(CallTracingMock):
    def __init__(self, *args):
        CallTracingMock.__init__(self, *args)
        self._deferreds = []
        self.helo_ehlo_done = True
        self.esmtp_features = set([
            'SOME-UNSUPPORTED-FEATURE'])
        self._fail_want_reply_for_EHLO = False

    def invoke_deferreds(self, ignore=(':CLOSE',)):
        deferreds = []
        for command, callback in self._deferreds:
            if command in ignore:
                deferreds.append((command, callback))
                continue
            callback()
        self._deferreds = deferreds

    def _make_reply(self, command, *args):
        d = Deferred()
        self._deferreds.append((command, partial(d, command, *args)))
        return d

    def push(self, *args):
        self._trace_call('push', *args)
        return self._make_reply('250', '2.1.0 Ok <gw:.>')

    def docmd(self, cmd, args=""):
        self._trace_call('docmd', cmd, args)
        if cmd == 'mail':
            return self._make_reply(250, "2.1.0 Ok <gw:mail>")
        elif cmd == 'rcpt':
            return self._make_reply(250, "2.1.0 Ok <gw:rcpt>")
        elif cmd == 'data':
            if args is not "":
                raise RuntimeError(
                    "Got no empty string for DATA") #pragma: no cover
            return self._make_reply(354, "Expecting data <gw:data>")
        elif cmd == 'xforward':
            return self._make_reply('250', 'Ok <gw:xforward>')

    def want_reply_for(self, command):
        self._trace_call('want_reply_for', command)
        args = (250, 'Hi')
        if self._fail_want_reply_for_EHLO and command == 'EHLO':
            raise CommandNotSentError(command)
        if command == ':CLOSE':
            args = [command]
        return self._make_reply(*args)

    def smtp_XFORWARD(self, arg):
        self._trace_call('smtp_XFORWARD', arg)
        return self._make_reply(250, 'Ok')


class TestESMTPProxy(unittest.TestCase):
    def test(self):
        def factory(*args, **kwargs):
            smtp = SMTPMock(('127.0.0.1', 10025), *args, **kwargs)
            self.smtp = smtp
            return smtp
        p = ESMTPProxy(factory)
        p.accept = lambda: ('socket', ('127.0.0.1', 3333))
        p.handle_accept()
        self.assertEquals(self.smtp.mock_get_next_call(),
            ('__init__', ('127.0.0.1', 10025)))
        self.assertEquals(self.smtp.mock_get_next_call(),
            ('want_reply_for', ':CLOSE'))
        self.assertEquals(self.smtp.mock_get_next_call(),
            ('ehlo_or_helo_if_needed',))

    def test_rewriter(self):
        def factory(*args, **kwargs):
            smtp = SMTPMock(('127.0.0.1', 10025), *args, **kwargs)
            self.smtp = smtp
            return smtp
        res = []
        def rewriter(data):
            res.append(data)
            return data

        oldlen = len(asynchat_instances)
        p = ESMTPProxy(factory)
        p.rewrite_message = rewriter
        p.accept = lambda: ('socket', ('127.0.0.1', 3333))
        p.handle_accept()
        self.assertEquals(self.smtp.mock_get_next_call(),
            ('__init__', ('127.0.0.1', 10025)))
        self.assertEquals(self.smtp.mock_get_next_call(),
            ('want_reply_for', ':CLOSE'))
        self.assertEquals(self.smtp.mock_get_next_call(),
            ('ehlo_or_helo_if_needed',))

        self.assertEquals(len(asynchat_instances) - oldlen, 1)
        channel = asynchat_instances.pop()
        def send(data):
            channel.collect_incoming_data(data)
            channel.found_terminator()
        send("ehlo x")
        send("mail from:<x>")
        send("rcpt to:<y>")
        send("data")
        send("some test")
        res = res[0]
        self.failUnless(res.startswith('Received: '))
        self.failUnless('\r\n' in res)
        self.assertEquals(res.split('\r\n')[-1], 'some test')



class TestESMTPProxyChannel(ESMTPTest):
    def setUp(self):
        self.s = SMTPMock()
        self.e = ESMTPProxyChannel(
            None, #sock
            None, #client address
            smtp_handler=self.s)

    def test_successful_session(self):
        self.consume()
        ret = self.sendc('ehlo asd', consume=2)
        self.assertEquals(ret[0][1][:4], '250-')
        self.assertEquals(ret[1:], [
            ('push', '250 8BITMIME\r\n'),
        ])
        self.send('mail from:<sender>')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '250 2.1.0 Ok <gw:mail>\r\n'))
        self.send('rcpt to:<recipient1>')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '250 2.1.0 Ok <gw:rcpt>\r\n'))
        self.send('data')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call(),
            ('set_terminator', '\r\n.\r\n'))
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '354 Expecting data <gw:data>\r\n'))
        self.send('Some test mail')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call(),
            ('set_terminator', '\r\n'))
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '250 2.1.0 Ok <gw:.>\r\n'))
        self.send('quit')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '221 2.0.0 Bye\r\n'))
        self.assertEquals(self.e.mock_get_next_call(),
            ('close_when_done',))

        self.assertEquals(self.s.mock_get_next_call(),
            ('__init__',))
        self.assertEquals(self.s.mock_get_next_call(),
            ('want_reply_for', ':CLOSE'))
        self.assertEquals(self.s.mock_get_next_call(),
            ('ehlo_or_helo_if_needed',))
        self.assertEquals(self.s.mock_get_next_call(),
            ('docmd', 'mail', 'from:<sender>'))
        self.assertEquals(self.s.mock_get_next_call(),
            ('docmd', 'rcpt', 'to:<recipient1>'))
        self.assertEquals(self.s.mock_get_next_call(),
            ('docmd', 'data', ''))
        res = self.s.mock_get_next_call()
        self.assertEquals(res[0], 'push')
        res = res[1]
        self.failUnless(res.startswith('Received: '))
        self.failUnless('\r\n' in res)
        exp = 'Some test mail\r\n.\r\n'
        self.assertEquals(res[-len(exp):], exp)
        self.e.handle_close()
        self.assertEquals(self.s.mock_get_next_call(),
            ('docmd', 'quit', ''))
        self.assertEquals(self.s.mock_get_next_call(),
            ('close_when_done',))

    def test_successful_session_with_dot(self):
        self.consume()
        ret = self.sendc('ehlo asd', consume=2)
        self.assertEquals(ret[0][1][:4], '250-')
        self.assertEquals(ret[1:], [
            ('push', '250 8BITMIME\r\n'),
        ])
        self.send('mail from:<sender>')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '250 2.1.0 Ok <gw:mail>\r\n'))
        self.send('rcpt to:<recipient1>')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '250 2.1.0 Ok <gw:rcpt>\r\n'))
        self.send('data')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call(),
            ('set_terminator', '\r\n.\r\n'))
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '354 Expecting data <gw:data>\r\n'))
        self.send('Some test mail\r'
            '..With some dots\n'
            '..all\r\n'
            '..over\n'
            '..\r\n'
            'the place.')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call(),
            ('set_terminator', '\r\n'))
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '250 2.1.0 Ok <gw:.>\r\n'))
        self.send('quit')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '221 2.0.0 Bye\r\n'))
        self.assertEquals(self.e.mock_get_next_call(),
            ('close_when_done',))

        self.assertEquals(self.s.mock_get_next_call(),
            ('__init__',))
        self.assertEquals(self.s.mock_get_next_call(),
            ('want_reply_for', ':CLOSE'))
        self.assertEquals(self.s.mock_get_next_call(),
            ('ehlo_or_helo_if_needed',))
        self.assertEquals(self.s.mock_get_next_call(),
            ('docmd', 'mail', 'from:<sender>'))
        self.assertEquals(self.s.mock_get_next_call(),
            ('docmd', 'rcpt', 'to:<recipient1>'))
        self.assertEquals(self.s.mock_get_next_call(),
            ('docmd', 'data', ''))
        res = self.s.mock_get_next_call()
        self.assertEquals(res[0], 'push')
        res = res[1]
        self.failUnless(res.startswith('Received: '))
        self.failUnless('\r\n' in res)
        exp = (
            'Some test mail\r'
            '..With some dots\n'
            '..all\r\n'
            '..over\n'
            '..\r\n'
            'the place.\r\n.\r\n')
        self.assertEquals(res[-len(exp):], exp)
        self.e.handle_close()
        self.assertEquals(self.s.mock_get_next_call(),
            ('docmd', 'quit', ''))
        self.assertEquals(self.s.mock_get_next_call(),
            ('close_when_done',))

    def test_rewriter(self):
        class RewriterMock(object):
            def rewrite_message(self, data):
                return data + '\r\n-- xyz'

        rewriter = RewriterMock()
        self.s = SMTPMock()
        self.e = ESMTPProxyChannel(
            None, #sock
            None, #client address
            smtp_handler=self.s,
            rewriter=rewriter)

        self.consume()
        ret = self.sendc('ehlo asd', consume=2)
        self.assertEquals(ret[0][1][:4], '250-')
        self.assertEquals(ret[1:], [
            ('push', '250 8BITMIME\r\n'),
        ])
        self.send('mail from:<sender>')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '250 2.1.0 Ok <gw:mail>\r\n'))
        self.send('rcpt to:<recipient1>')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '250 2.1.0 Ok <gw:rcpt>\r\n'))
        self.send('data')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call(),
            ('set_terminator', '\r\n.\r\n'))
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '354 Expecting data <gw:data>\r\n'))
        self.send('Some test mail\r\nIt sometimes includes dots '
            'at the beginning of a line...\r\n..\r\n')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call(),
            ('set_terminator', '\r\n'))
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '250 2.1.0 Ok <gw:.>\r\n'))
        self.send('quit')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '221 2.0.0 Bye\r\n'))
        self.assertEquals(self.e.mock_get_next_call(),
            ('close_when_done',))

        self.assertEquals(self.s.mock_get_next_call(),
            ('__init__',))
        self.assertEquals(self.s.mock_get_next_call(),
            ('want_reply_for', ':CLOSE'))
        self.assertEquals(self.s.mock_get_next_call(),
            ('ehlo_or_helo_if_needed',))
        self.assertEquals(self.s.mock_get_next_call(),
            ('docmd', 'mail', 'from:<sender>'))
        self.assertEquals(self.s.mock_get_next_call(),
            ('docmd', 'rcpt', 'to:<recipient1>'))
        self.assertEquals(self.s.mock_get_next_call(),
            ('docmd', 'data', ''))
        res = self.s.mock_get_next_call()
        self.assertEquals(res[0], 'push')
        res = res[1]
        self.failUnless(res.startswith('Received: '))
        self.failUnless('\r\n' in res)
        exp = ('Some test mail\r\nIt sometimes includes dots'
                ' at the beginning of a line...\r\n..\r\n'
                '\r\n-- xyz\r\n.\r\n')
        self.assertEquals(res[-len(exp):], exp)
        self.e.handle_close()
        self.assertEquals(self.s.mock_get_next_call(),
            ('docmd', 'quit', ''))
        self.assertEquals(self.s.mock_get_next_call(),
            ('close_when_done',))

    def test_rewriter_exception(self):
        class RewriterMock(object):
            def rewrite_message(self, data):
                0/0

        rewriter = RewriterMock()
        self.s = SMTPMock()
        self.e = ESMTPProxyChannel(
            None, #sock
            None, #client address
            smtp_handler=self.s,
            rewriter=rewriter)

        self.consume()
        ret = self.sendc('ehlo asd', consume=2)
        self.assertEquals(ret[0][1][:4], '250-')
        self.assertEquals(ret[1:], [
            ('push', '250 8BITMIME\r\n'),
        ])
        self.send('mail from:<sender>')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '250 2.1.0 Ok <gw:mail>\r\n'))
        self.send('rcpt to:<recipient1>')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '250 2.1.0 Ok <gw:rcpt>\r\n'))
        self.send('data')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call(),
            ('set_terminator', '\r\n.\r\n'))
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '354 Expecting data <gw:data>\r\n'))
        self.send('Some test mail\r\nIt sometimes includes dots '
            'at the beginning of a line...\r\n..\r\n')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call(),
            ('set_terminator', '\r\n'))
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '451 4.4.0 Requested action aborted: local '
                'error in processing\r\n'))
        self.send('quit')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '221 2.0.0 Bye\r\n'))
        self.assertEquals(self.e.mock_get_next_call(),
            ('close_when_done',))

        self.assertEquals(self.s.mock_get_next_call(),
            ('__init__',))
        self.assertEquals(self.s.mock_get_next_call(),
            ('want_reply_for', ':CLOSE'))
        self.assertEquals(self.s.mock_get_next_call(),
            ('ehlo_or_helo_if_needed',))
        self.assertEquals(self.s.mock_get_next_call(),
            ('docmd', 'mail', 'from:<sender>'))
        self.assertEquals(self.s.mock_get_next_call(),
            ('docmd', 'rcpt', 'to:<recipient1>'))
        self.assertEquals(self.s.mock_get_next_call(),
            ('docmd', 'data', ''))
        self.e.handle_close()
        self.assertEquals(self.s.mock_get_next_call(),
            ('docmd', 'quit', ''))
        self.assertEquals(self.s.mock_get_next_call(),
            ('close_when_done',))

    def test_ehlo_upstream_early(self):
        self.consume()
        self.s.helo_ehlo_done = True
        self.s.esmtp_features.add('PIPELINING')
        ret = self.sendc('ehlo asd', consume=3)
        self.assertEquals(ret[0][1][:4], '250-')
        self.assertEquals(ret[1:], [
            ('push', '250-PIPELINING\r\n'),
            ('push', '250 8BITMIME\r\n'),
        ])

    def test_dead_upstream(self):
        self.consume()
        self.send('helo want-close')
        self.assertEquals(self.s.mock_get_next_call(), ('__init__',))
        self.assertEquals(self.s.mock_get_next_call(),
            ('want_reply_for', ':CLOSE'))
        self.assertEquals(self.s.mock_get_next_call(),
            ('ehlo_or_helo_if_needed',))
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call()[1][:4], '250 ')
        self.s.invoke_deferreds(ignore=())
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '421 4.4.0 Service not available, closing '
                'transmission channel (upstream failure)\r\n'))
        self.assertEquals(self.e.mock_get_next_call(),
            ('handle_close',))

    def test_ehlo_upstream_late(self):
        self.consume()
        self.s.helo_ehlo_done = False
        self.send('ehlo asd')
        self.assertEquals(self.s.mock_get_next_call(), ('__init__',))
        self.assertEquals(self.s.mock_get_next_call(),
            ('want_reply_for', ':CLOSE'))
        self.assertEquals(self.s.mock_get_next_call(),
            ('ehlo_or_helo_if_needed',))
        self.assertEquals(self.s.mock_get_next_call(),
            ('want_reply_for', 'EHLO'))
        self.s.esmtp_features.add('PIPELINING')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call()[1][:4], '250-')
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '250-PIPELINING\r\n'))
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '250 8BITMIME\r\n'))

    def test_ehlo_upstream_EHLO_done_HELO_not(self):
        self.consume()
        self.s.helo_ehlo_done = False
        self.s._fail_want_reply_for_EHLO = True
        self.send('ehlo asd')
        self.assertEquals(self.s.mock_get_next_call(), ('__init__',))
        self.assertEquals(self.s.mock_get_next_call(),
            ('want_reply_for', ':CLOSE'))
        self.assertEquals(self.s.mock_get_next_call(),
            ('ehlo_or_helo_if_needed',))
        self.assertEquals(self.s.mock_get_next_call(),
            ('want_reply_for', 'EHLO'))
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call()[1][:4], '250-')
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '250 8BITMIME\r\n'))

    def test_xforward(self):
        self.consume()
        self.s.helo_ehlo_done = True
        self.s.esmtp_features.add('XFORWARD NAME IP')
        ret = self.sendc('ehlo asd', consume=3)
        self.assertEquals(ret[0][1][:4], '250-')
        self.assertEquals(ret[1:], [
            ('push', '250-8BITMIME\r\n'),
            ('push', '250 XFORWARD NAME IP\r\n'),
        ])
        self.send('xforward ip=127.0.0.1')
        self.s.invoke_deferreds()
        self.assertEquals(self.e.mock_get_next_call(),
            ('push', '250 Ok <gw:xforward>\r\n'))

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
