import unittest
from ...timer import TimerManager, timer_manager, CallLater
from ..mock import CallTracingMock

class TimeSimulator(object):
    def __init__(self, start=0):
        self._time = start

    def time(self):
        return self._time

    def sleep(self, delay):
        self._time += delay


class TestTimerManager(unittest.TestCase):
    def setUp(self):
        self.ts = TimeSimulator(start=10)
        self.t = TimerManager(self.ts.time)
        self.ctm = CallTracingMock()

    def test(self):
        self.t.at(20, self.ctm.callback)
        self.t.run()
        self.assertRaises(Exception, self.ctm.mock_get_next_call())
        self.assertEquals(self.t.time_to_next_event(), 10)
        self.ts.sleep(10)
        self.t.run()
        self.assertEquals(self.ctm.mock_get_next_call(), ('callback',))
        self.assertEquals(self.t.time_to_next_event(), None)

    def test_global(self):
        self.failUnless(isinstance(timer_manager, TimerManager))

    def test_multiple_same_time(self):
        self.t.at(20, self.ctm.callback1)
        self.t.at(20, self.ctm.callback2)
        self.t.run()
        self.assertRaises(Exception, self.ctm.mock_get_next_call())
        self.assertEquals(self.t.time_to_next_event(), 10)
        self.ts.sleep(10)
        self.t.run()
        self.assertEquals(self.ctm.mock_get_next_call(), ('callback1',))
        self.assertEquals(self.ctm.mock_get_next_call(), ('callback2',))
        self.assertEquals(self.t.time_to_next_event(), None)

    def test_second_then_first(self):
        self.t.at(20, self.ctm.callback1)
        self.t.at(15, self.ctm.callback2)
        self.t.run()
        self.assertRaises(Exception, self.ctm.mock_get_next_call())
        self.assertEquals(self.t.time_to_next_event(), 5)
        self.ts.sleep(10)
        self.t.run()
        self.assertEquals(self.ctm.mock_get_next_call(), ('callback2',))
        self.assertEquals(self.ctm.mock_get_next_call(), ('callback1',))
        self.assertEquals(self.t.time_to_next_event(), None)

    def test_cancel(self):
        id = self.t.at(20, self.ctm.callback1)
        self.assertEquals(self.t.time_to_next_event(), 10)
        self.t.cancel(id)
        self.assertEquals(self.t.time_to_next_event(), None)

    def test_cancel_2(self):
        id = self.t.at(20, self.ctm.callback1)
        self.assertEquals(self.t.time_to_next_event(), 10)
        self.t.cancel(id)
        self.assertEquals(self.t.time_to_next_event(), None)

    def test_cancel_3(self):
        id = self.t.at(20, self.ctm.callback1)
        self.assertEquals(self.t.time_to_next_event(), 10)
        self.t.cancel(id)
        self.assertEquals(self.t.time_to_next_event(), None)
        self.t.run()

    def test_cancel_4(self):
        class Test(object):
            def callback(self):
                self.t.cancel(self.id)
        test = Test()
        test.t = self.t
        id = self.t.at(20, test.callback)
        test.id = id
        self.assertEquals(self.t.time_to_next_event(), 10)
        self.ts.sleep(10)
        self.t.run()
        self.assertEquals(self.t.time_to_next_event(), None)

    def test_time_to_next_event_negative(self):
        # time_to_next_event() should not return negative values --
        # it should return 0 if the next timer is already due;
        # this is because select() & co don't support negative timeouts
        id = self.t.at(20, self.ctm.callback)
        self.ts.sleep(30)
        self.assertEquals(self.t.time_to_next_event(), 0)


class TestCallLater(unittest.TestCase):
    def setUp(self):
        self.ctm = CallTracingMock()
        self.ts = TimeSimulator(start=40)
        self.t = TimerManager(self.ts.time)

    def test(self):
        c = CallLater(self.ctm.callback, 3, manager=self.t)
        self.ts.sleep(4)
        self.t.run()
        self.assertEquals(self.ctm.mock_get_next_call(), ('__init__',))
        self.assertEquals(self.ctm.mock_get_next_call(), ('callback',))

    def test_cancel(self):
        c = CallLater(self.ctm.callback, 3, manager=self.t)
        self.ts.sleep(4)
        c.cancel()
        self.t.run()
        self.ctm.sep()
        self.assertEquals(self.ctm.mock_get_next_call(), ('__init__',))
        self.assertEquals(self.ctm.mock_get_next_call(), ('sep',))
        self.assertEquals(c.manager, None)
        self.assertEquals(c.callback, None)

    def test_reset(self):
        c = CallLater(self.ctm.callback, 3, manager=self.t)
        self.ts.sleep(2)
        self.t.run()
        self.ctm.sep1()
        self.ts.sleep(1)
        c.reset()
        self.ts.sleep(2)
        self.t.run()
        self.ctm.sep2()
        self.ts.sleep(1)
        self.t.run()
        self.assertEquals(self.ctm.mock_get_next_call(), ('__init__',))
        self.assertEquals(self.ctm.mock_get_next_call(), ('sep1',))
        self.assertEquals(self.ctm.mock_get_next_call(), ('sep2',))
        self.assertEquals(self.ctm.mock_get_next_call(), ('callback',))

    def test_destroy_callback_still_there(self):
        c = CallLater(self.ctm.callback, 3, manager=self.t)
        self.assertEquals(self.t.time_to_next_event(), 3)
        c.destroy()
        self.assertEquals(self.t.time_to_next_event(), None)

    def test_destroy_callback_already_gone(self):
        c = CallLater(self.ctm.callback, 3, manager=self.t)
        self.assertEquals(self.t.time_to_next_event(), 3)
        self.ts.sleep(3)
        self.t.run()
        c.destroy()
        self.assertEquals(self.t.time_to_next_event(), None)

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
