import sys

class replace_global_symbols(object):
    """
    Dynamically replaces entries of the global module table.
    Useful for testing.
    """

    """
    Class variable which holds references to all modules which have been
    removed from or imported to sys.modules.
    This ensures, that modules are not garbage-collected, which could
    happen if a module imports a name from another module, but not
    the module itself.
    Typical error would be that module globals are all None.
    """
    _refs = set()

    def __init__(self, replacements, dependencies=None):
        """
        Constructor

        @param dict replacements {str name => object attribute}
        @param list dependencies which should be reloaded afterwards
        """
        self.replacements = replacements
        self.dependencies = dependencies or []
        self._replaced_global_modules = {}

    def __enter__(self):
        # remove all old instances of all depending modules from the
        # global module store
        for dep in self.dependencies:
            self.remove_global_module(dep)

        for name, symbol in self.replacements.items():
            self.replace_global_symbol(name, symbol)

        for name in self.dependencies:
            self.import_module(name)

    def remove_global_module(self, name):
        """
        Removes a module from the global sys.modules list,
        independent of the import technique used.
        (import pkg.module ends up somewhere else than
        from pkg import module or from pkg.module import Class)

        It also keeps a reference to the old module to avoid it being
        garbage-collected, as symbols from the module might still be
        in use by other users.

        @param str name pkg.module
        """
        comps = name.split('.')
        for idx, comp in enumerate(comps):
            # this loop iterates over pkg, pkg.subpkg, pkg.subsubpkg...
            path = '.'.join(comps[:idx+1])
            if path not in sys.modules:
                continue
            root = sys.modules[path]
            found = True
            for comp2 in comps[idx+1:-1]:
                # this loop removes sys.modules[pkg].subpkg.subsubpkg,
                # sys.modules[pkg.subpkg].subsubpkg etc.
                if not hasattr(root, comp2):
                    found = False
                    break
                root = getattr(root, comp2)
            if hasattr(root, comps[-1]):
                self._refs.add(getattr(root, comps[-1]))
                delattr(root, comps[-1])

        if name in sys.modules:
            # and finally remove sys.modules[pkg.subpkg.subsubpkg]
            self._refs.add(sys.modules[name])
            del sys.modules[name]

    def import_module(self, name):
        """
        Imports a module, adds it to the internal reference tracking
        mechanism and returns the module instance.

        @param str name pkg.mod
        @return module
        """
        mod = __import__(name)
        self._refs.add(mod)
        return mod

    def replace_global_symbol(self, name, symbol):
        """
        Replaces a symbol from a global module.

        This method makes sure that previous users of the affected
        global module are not influenced in any way, in other words, they
        should stay working as they are.

        @param str name pkg.mod.Symbol
        @param object replacement symbol
        """
        comps = name.split('.')
        mod_name = '.'.join(comps[:-1])

        if mod_name not in self._replaced_global_modules:
            # first, kick out old instances of this module so that we don't
            # modify behaviour of users, which imported this module before us
            self.remove_global_module(mod_name)

            # now re-import this module...
            root = self.import_module(mod_name)

            # and add our patches to it...
            for comp in comps[1:-1]:
                if not hasattr(root, comp):
                    return False
                root = getattr(root, comp)

            self._replaced_global_modules[mod_name] = root

        setattr(self._replaced_global_modules[mod_name], comps[-1], symbol)

    def restore_global_symbol(self, symbol):
        """
        Restores a global symbol to its previous state by removing
        the modified version (while still keeping a reference to it)
        and re-importing the symbol using the usual import mechanism.
        """
        comps = symbol.split('.')
        mod_name = '.'.join(comps[:-1])
        self.remove_global_module(mod_name)
        self.import_module(mod_name)

    def __exit__(self, exc_type, exc_value, exc_tb):
        for name in self.replacements:
            self.restore_global_symbol(name)

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
