from time import time
from heapq import heappop, heappush

class TimerManager(object):
    """
    A simple Timer Manager which supports registering callbacks
    for specific points in time, running these and integrating
    them into some kind of main loop.
    """
    def __init__(self, get_current_time=time):
        """
        Constructor

        @param callback get_current_time: usually time.time
        """
        self.get_current_time = get_current_time
        self._next_timers = []
        self._timers = {}

    def at(self, time, callback):
        """
        Registers a callback, i.e. makes sure that "callback" is called
        at "time"

        @param int time
        @param callback callback
        """
        if time not in self._timers:
            heappush(self._next_timers, time)
        list = self._timers.setdefault(time, [])
        list.append(callback)
        return (time, callback)

    def cancel(self, id):
        """
        Cancels a previously registered event.

        @param object id: event id returned by .at()
        """
        time, callback = id
        self._timers[time].remove(callback)
        if not self._timers[time]:
            del self._timers[time]
            try:
                self._next_timers[self._next_timers.index(time)] = None
            except ValueError:
                # this case may happen if a callback attempts to
                # cancel itself; in this case, the callback's triggering
                # timer has already been popped from the heap
                pass

    def run(self):
        """
        Executes all calbacks which are due.
        """
        now = self.get_current_time()
        while self._next_timers:

            if self._next_timers[0] is None:
                # handle cancelled events (=> discard & ignore)
                heappop(self._next_timers)
                continue

            if self._next_timers[0] > now:
                # don't run yet
                break

            time = heappop(self._next_timers)
            for callback in self._timers[time]:
                callback()

            if time in self._timers:
                # might have been deleted by a callback
                del self._timers[time]

    def time_to_next_event(self):
        """
        Returns the time left until the next .run() invocation should
        happen so that the next callback is called in time.

        @return int seconds
        """
        if not self._next_timers or self._next_timers[0] is None:
            return None
        return max(0, self._next_timers[0] - self.get_current_time())


# a global, singleton timer manager
timer_manager = TimerManager()


class CallLater(object):
    """
    A simple object-oriented interface to manage delayed calls.
    """
    def __init__(self, callback, delay=0, manager=timer_manager):
        """
        Constructor.

        @param callable callback
        @param int delay: in seconds
        @param TimerManager manager: defaults to the global timer_manager
        """
        self.manager = manager
        self.callback = callback
        self.delay = delay
        self.setup()

    def setup(self):
        """
        Registers the required event at the TimerManager.
        """
        self.timer_id = self.manager.at(
            self.manager.get_current_time() + self.delay,
            self.callback)

    def cancel(self):
        """
        Cancels this event.

        Once this method is called, the associated callback will never
        be called by us again, unless re-registered.
        """
        self._cancel()
        self.destroy()

    def _cancel(self):
        """
        Internal helper method, which does the actual cancellation.
        """
        self.manager.cancel(self.timer_id)

    def destroy(self):
        """
        Cleanup method which removes unneeded references and avoids
        cyclic references.
        """
        try:
            self._cancel()
        except KeyError:
            pass
        self.manager = None
        self.callback = None

    def reset(self):
        """
        Resets this delayed call.

        You can use this to implement timeouts, e.g.:

          - Client connects:
            self.timeout_call = CallLater(self.handle_timeout, 25)
          - Client sends some data, so reset the timeout:
            self.timeout_call.reset()
          - (same scenario might happen several times)
          - Client doesn't send any data in 25 seconds, so the event
            triggers and self.handle_timeout is called
        """
        self._cancel()
        self.setup()

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
