from ctypes import CDLL, c_int, c_uint, c_char_p, POINTER, pointer
from ctypes.util import find_library

libc = CDLL(find_library("c"))
libc.getgrouplist.restype = c_int
# ideally, we would also populate .argtypes here, but we cannot
# as the argtypes vary depending on the number of expected results

# pre-define some types which ctypes does not provide for us and/or
# define aliases
c_gid_t = c_uint
c_int_p = POINTER(c_int)

class GroupListLookupError(Exception): pass

def getgrouplist(username, group):
    """
    Wrapper for the libc function getgrouplist()

    ... which returns a list of all supplementary group ids for the
    given username.

    The parameter group must be set to a gid, usually the primary gid is
    passed using this variable. It will be returned as part of the
    result set.

    >=Python-3.3 has this in the standard library in the posix module.

    man getgrouplist(3)
    http://man7.org/linux/man-pages/man3/getgrouplist.3.html

    @param str username
    @param int group: primary group id
    @param int num_groups: maximum number of returned groups
    @return list: [100, 200, ...]
    """
    ngroups = c_int(1024)
    for count in (0, 1):
        # We call the library function once and expect a maximum number
        # of groups of 1024. If this is exceeded, the library function
        # returns -1, but gives us the number of groups to
        # expect.
        # We then run the library function with this new upper bound again.
        # As such, two calls might be needed, which is the reason for this
        # loop.

        # dynamically create the array type and the related pointer;
        # this has to happen dynamically in each iteration, as
        # ngroups.value will change between loop iterations.
        c_gid_t_array = c_gid_t * ngroups.value
        c_gid_t_array_ptr = POINTER(c_gid_t_array)
        # ... and update the function signature (from python's view)
        libc.getgrouplist.argtypes = (
            c_char_p, # const char *user
            c_gid_t, # gid_t group
            c_gid_t_array_ptr, # gid_t *groups
            c_int_p # int *ngroups
        )

        # our result list...
        groups = c_gid_t_array()

        # and the actual library call
        num_returned_groups = libc.getgrouplist(
            username,
            group,
            pointer(groups),
            pointer(ngroups))

        if num_returned_groups >= 0:
            # there are less than ngroups groups, as such we got the
            # full list and are done now
            break
        # otherwise, the loop is repeated with the changed ngroups
        # value (this gets changed in place by the library call)

        if count >= 1:
            # sanity check:
            # we should never reach this point; if we do, something
            # went wrong...
            raise GroupListLookupError("getgrouplist() returned -1")

    # the array has to be sliced since we may have over-estimated our
    # groups array length; if we didn't slice, we would see multiple
    # occurences of the element "0", which shouldn't be there
    return groups[0:num_returned_groups]

if __name__ == '__main__':
    import sys
    print getgrouplist(sys.argv[1], int(sys.argv[2]))

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
