import logging
import os
import re
import time
import math
import GnuPGInterface
from .paths import find_file


logger = logging.getLogger(__name__)

class License(object):
    LICENSE_WARNING = (
        "Your license is about to expire in %s days. "
        "Please contact sales@bitbone.de if you wish to extend your license."
        "\n\n"
        "Please note that after your license expires, your messages will "
        "be just passed through and no disclaimer will be attached.")

    LICENSE_EXPIRED = (
        "Your license is expired. No disclaimers will be attached and the "
        "mails will be just passed through.")

    def __init__(self, key_file="/etc/maildisclaimer/license/base", gpg_home="/usr/share/maildisclaimer/.gnupg"):
        self.key_refresh_time = 0
        self.key_file = os.path.abspath(key_file)
        self.gpg_home = gpg_home
        self._key = {}

        self.version = self.get_version()
        self.install_time = self.get_install_time()

    def is_valid(self):
        key = self.read_key()

        if not bool(key):
            logger.info("Empty key")
            return False

        if key["Version"] != self.version:
            logger.info("Invalid version; key: %s - installed: %s" % (key["Version"], self.version))
            return False

        if key["Type"] == "trial":
            if not "Days" in key:
                logger.info("Trial key has no valid days")
                return False

            if time.time() > self.install_time + (self.seconds_from_days(int(key["Days"]))):
                logger.info("Trial key is expired: install_date %s, days: %s" % (self.install_time, key["Days"]))
                return False

        return True

    def get_install_time(self):
        filename = find_file(["/etc/maildisclaimer/systemid", "./systemid"])
        if not filename:
            return 0
        return os.stat(filename).st_mtime

    def get_version(self):
        try:
            version = find_file(["/usr/share/maildisclaimer/VERSION", "./VERSION"])
            f = open(version, "r")
            version = f.read()
            f.close()

            match = re.match("^(\d+)\..*", version)
            return match.group(1)
        except:
            return ""

    def read_key(self, force=False):
        if not force and self.key_refresh_time > 0 and self.key_refresh_time >= time.time():
            return self._key

        self.key_refresh_time = time.time() + (3600*24)

        try:
            self._key = self._parse_gpg_output()
            return self._key

        except Exception, msg:
            # Verification failed somehow (gpg signature failed, file not found, etc)
            logger.error("License validation failed due to underlying problem: %s", msg)

        self._key = {}
        return {}

    def seconds_from_days(self, d):
        return d*3600*24

    def _parse_gpg_output(self):
        gnupg = GnuPGInterface.GnuPG()
        gnupg.options.homedir = self.gpg_home

        p1 = gnupg.run(['--decrypt', self.key_file],
            create_fhs=['stdout', 'stderr', 'status'])
        p1.handles["stderr"].close() #discard
        status = p1.handles["status"].readlines()
        p1.handles["status"].close()

        key = p1.handles["stdout"].readlines()
        p1.handles["stdout"].close()
        p1.wait()

        matched = False

        for line in status:
            if re.match(r'^\[GNUPG:\]\s+VALIDSIG\s+38AD9D3AB0BEACFC1F5F6CDE851FE118A52446ED\s', line):
                matched = True
                break

        if not matched:
            logger.debug("gpg status: %r", status)
            raise KeyError("GPG could not confirm that the license file "
                "is signed by the proper key.")

        ret = {}
        for line in key:
            (key, value) = re.split(r"\s*:\s*", line.strip())
            ret[key] = value

        return ret


    def report(self):
        key = self.read_key()

        if len(key) == 0:
            return self.LICENSE_EXPIRED

        if key["Type"] == "trial":

            if self.install_time + (self.seconds_from_days(int(key["Days"]))) - time.time() < 0:
                return self.LICENSE_EXPIRED
            elif self.install_time + (self.seconds_from_days(int(key["Days"]))) - time.time() <= self.seconds_from_days(5):
                return self.LICENSE_WARNING % int( math.ceil( (self.install_time + (self.seconds_from_days(int(key["Days"]))) - time.time()) / 3600 / 24 ) )
            else:
                return ""

        if self.is_valid():
            return ""

    def replace_license(self, content):
        fd = open(self.key_file, "w")
        fd.write(content)
        fd.close()

        self.key_refresh_time = 0

        return self.read_key()

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
