#! /usr/bin/python
import os
import logging
import asyncore
import bitbone.performance
from copy import copy
from functools import partial
from bitbone.timer import timer_manager
from bitbone.disclaimers.storage import CustomRulesBackend, try_names, domain_of_address
from bitbone.disclaimers.smtp import ContentFilter
from bitbone.disclaimers.email import extract_targets_from_header
from optparse import OptionParser, Option
from bitbone.config import MDConfigReader

logger = logging.getLogger(__name__)

def check_bool(option, opt, value):
    try:
        if value.lower() in [ "yes", "true", "1" ]:
            return True
        else:
            return False
    except ValueError:
        raise OptionValueError(
            "option %s: invalid bool value: %r" % (opt, value))

class BoolSupportOption(Option):
    TYPES = Option.TYPES + ("bool",)
    TYPE_CHECKER = copy(Option.TYPE_CHECKER)
    TYPE_CHECKER["bool"] = check_bool
    TYPED_ACTIONS = Option.TYPED_ACTIONS + ("store_true",)

    def takes_value(self):
        if self.type == "bool":
            return False
        return Option.takes_value(self)

class Main(object):

    def make_optparser(self):
        self.parser = OptionParser(option_class=BoolSupportOption)

        self.parser.add_option("-d", "--debug", dest="debug",
            action="store_true",
            type="bool",
            default=False, help="show debug messages on the console")

        self.parser.add_option("--email-dump-path", dest="email_dump_path",
            metavar='DIRECTORY',
            default=None, help="if set, problematic mails will be "
                "written to the specified directory")

        self.parser.add_option("-q", "--quiet", dest="quiet",
            action="store_true",
            type="bool",
            default=False, help="don't output anything")

        self.parser.add_option("--syslog", dest="syslog",
            action="store_true",
            type="bool",
            default=False, help="log to syslog")

        self.parser.add_option("-g", "--gateway", dest="gateway",
            help="upstream mail server "
                "(usually 127.0.0.1:10025 or something")

        self.parser.add_option("-b", "--bind", dest="bind",
            help="local listening socket, usually 127.0.0.1:10026")

        self.parser.add_option('--local-domain-config',
            help="path to a config file, which contains virtual_mailbox_"
                "domains = domain1, domain2 (usually /etc/postfix/main.cf)",
            default="/etc/postfix/main.cf")

        self.parser.add_option('--ldap-config',
            help="path to a config file, which contains the LDAP "
                "login credentials and field mappings "
                "(usually /etc/maildisclaimer/ldap.cfg)",
            default='/etc/maildisclaimer/ldap.cfg')

        self.parser.add_option('--disclaimer-dir',
            help="path to the directory which will contain "
                "all disclaimer-related files (data + configuration) "
                "(usually /opt/bitbone/config/maildisclaimers)",
            default='/opt/bitbone/config/maildisclaimers')

        self.parser.add_option("--always-embed-html",
            dest="always_embed_html",
            type="bool",
            action="store_true", default=False,
            help="always (also) includes the html "
                "disclaimer, even for text/plain messages")

        self.parser.add_option("--prefer-direct-embedding",
            dest="prefer_direct_embedding",
            type="bool",
            action="store_true", default=False,
            help="try hard to avoid MIME embedding")

        self.parser.add_option("--enable-template-ldap",
            dest="enable_template_ldap",
            type="bool",
            action="store_true", default=False,
            help="permit arbitrary ${ldap:...} syntax")

        self.parser.add_option("--account-backend",
            dest="account_backend",
            default="passwd",
            help="select a different account backend [passwd|ldap]")

        self.parser.add_option("--avoid-quoted-printable",
            dest="avoid_quoted_printable",
            type="bool",
            action="store_true", default=False,
            help="avoid quoted-printable encoding even if the input was quopri-encoded")

        self.parser.add_option("--custom-rules",
            dest="custom_rules",
            default=None,
            help="use the given python file with custom rules for disclaimer selection")

        self.read_config()

        # use default from config file
        self.parser.set_defaults(**self.config.data)

    def read_config(self):

        try:
            self.config = MDConfigReader()

            default = "/etc/maildisclaimer/server.cfg"
            cfg = os.getenv("MAILDISCLAIMER_CFG", default)

            self.config.read(cfg)

        except IOError, msg:
            if default != cfg:
                self.parser.error("Could not read configuration: %s" % str(msg))
            pass

    def parse_arguments(self, args=None):
        options, args = self.parser.parse_args(args)

        self.options = options
        self.args = args

    def make_content_filter(self):
        logger.info("Validating configuration...")

        # don't initialize any objects since we are going to stop anyway
        if self.args and self.args[0] == "stop":
            return

        if (not self.options.ldap_config or
                not os.path.isfile(self.options.ldap_config)):
            self.parser.error("'%s' does not exist "
                "(fix --ldap-config)" % self.options.ldap_config)

        if (not self.options.local_domain_config or
                not os.path.isfile(self.options.local_domain_config)):
            self.parser.error("'%s' does not exist "
                "(fix --local-domain-config)" %
                self.options.local_domain_config)

        if not self.options.gateway or self.options.gateway.count(':') != 1:
            self.parser.error("--gateway must be given and point "
                "to the upstream SMTP server")
        gateway = self.options.gateway.split(':', 1)
        gateway = (gateway[0], int(gateway[1]))

        if not self.options.bind or self.options.bind.count(':') != 1:
            self.parser.error("--bind must be given and point to the local "
                    "listening socket")
        bind = self.options.bind.split(':', 1)
        bind = (bind[0], int(bind[1]))

        if self.options.debug:
            loglevel = logging.DEBUG
        elif self.options.quiet:
            loglevel = logging.CRITICAL
        else:
            loglevel = logging.INFO

        if self.options.syslog:
            formatter = logging.Formatter(
                'bitbone-maildisclaimer: '
                '%(name)s.%(funcName)s %(levelname)s %(message)s')
            from bitbone.syslog import FaultTolerantSysLogHandler
            from logging.handlers import SysLogHandler
            handler = FaultTolerantSysLogHandler(address="/dev/log",
                facility=SysLogHandler.LOG_MAIL)
            handler.setFormatter(formatter)
            handler.setLevel(loglevel)
            logging.getLogger().addHandler(handler)
            logging.getLogger().setLevel(loglevel)
        else:
            logging.basicConfig(
                format='[%(asctime)s] %(name)s.%(funcName)s %(levelname)s: %(message)s',
                level=loglevel)

        if self.options.account_backend == 'passwd':
            from bitbone.accounts.passwd import AccountLookup \
                as account_lookup
            logger.info("using passwd account backend")
        elif self.options.account_backend == 'ldap':
            from bitbone.accounts.ldap import AccountLookup \
                as account_lookup
            logger.info("using ldap account backend")
        else:
            self.parser.error("invalid account backend")

        custom_rules_backend = None
        if self.options.custom_rules:
            try:
                custom_rules_backend = CustomRulesBackend(
                    path=self.options.custom_rules)
            except Exception as exc:
                self.parser.error("unable to set up custom rules (%s)" % exc)

        self.cf = ContentFilter(bind, gateway,
            local_domain_config=self.options.local_domain_config,
            ldap_config=self.options.ldap_config,
            disclaimer_dir=self.options.disclaimer_dir,
            always_embed_html=self.options.always_embed_html,
            prefer_direct_embedding=self.options.prefer_direct_embedding,
            enable_template_ldap=self.options.enable_template_ldap,
            account_lookup=account_lookup,
            email_dump_path=self.options.email_dump_path,
            avoid_quoted_printable=self.options.avoid_quoted_printable,
            custom_rules_backend=custom_rules_backend)

        if custom_rules_backend:
            custom_rules_backend.add_helper('is_local',
                self.cf.is_local_address)
            custom_rules_backend.add_helper('try_names',
                partial(try_names, self.options.disclaimer_dir))
            custom_rules_backend.add_helper('domain',
                domain_of_address)
            custom_rules_backend.add_helper('all_targets',
                extract_targets_from_header)
            custom_rules_backend.load()

    def run(self):
        self.cf.make_listener() #pragma: nocover
        return self.loop() #pragma: nocover

    def loop(self):
        logger.info("Entering mainloop")
        while 1:
            timer_manager.run()
            time_left = timer_manager.time_to_next_event()
            asyncore.loop(timeout=time_left, count=1)

if __name__ == '__main__':
    m = Main()
    m.make_optparser()
    m.parse_arguments()
    m.make_content_filter()
    if m.options.debug:
        from bitbone.debug import setup
        setup()
    m.run()

# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.
#
# Copyright (C) 2015 bitbone AG
